#!/usr/bin/perl -w

# Copyright 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Kevin Ryde, January 2020
#
# Usage: perl Udelta-2.pl
#
# Count the number of words on 2 symbols which avoid the pattern word
#
#     Udelta = ab w bc x ca y ba z ac
#
# of Baker, McNulty, and Taylor.  They show Udelta is 4-avoidable, but
# 3-unavoidable, and hence 2-unavoidable too.  This means there are no
# infinite words on 2 symbols which avoid Udelta, so every long enough word
# of 2 symbols will match Udelta.
#
# $Udelta below is the pattern word.  It is turned into a regexp with
# back-references.  A string which matches the regexp "encounters" Udelta
# and a string which which does not match the regexp "avoids" Udelta.
#
# If a given string encounters Udelta then there's no point trying to extend
# it to something longer, since the substring will always match Udelta.
# So the strategy here is to keep @strings as all the strings of length $len
# which do not match Udelta, and try adding one new symbol (0 or 1) to each
# in turn.
#
# Each new string has the new symbol at the start.  Since the rest does not
# match Udelta, this new symbol must be part of the match, so the regexp is
# anchored "^" for a match at the start.  This anchoring is a good speedup
# on medium to long strings.
#
# Expect a run-time up to about 2 hours at about 6000 bogomips (which is now
# not a particularly fast computer).

use 5.006;
use strict;
use warnings;
use File::Spec;
$|=1;

my $Udelta = 'ab w bc x ca y ba z ac';
my $radix = 2;

print "Udelta $Udelta\n";
my @Udelta = split / */, $Udelta;   # pattern symbols
my %Usymbol_count;
foreach my $s (@Udelta) {
  $Usymbol_count{$s}++;             # how many times each Usymbol occurs
}

my %Usymbol_to_re;                  # back-reference string
my $Uregexp = '';
my $capnum = 0;
foreach my $symbol (@Udelta) {
  if ($Usymbol_to_re{$symbol}) {
    $Uregexp .= $Usymbol_to_re{$symbol};      # existing back-reference
  } elsif ($Usymbol_count{$symbol} <= 1) {
    $Uregexp .= '.+';                         # occurs once, no group
  } else {
    $Uregexp .= '(.+)';                       # occurs >= 2 times, make group
    $capnum++;
    $Usymbol_to_re{$symbol} = "\\$capnum";    # back-reference string
  }
}
print "regexp $Uregexp\n";
$Uregexp = qr/^$Uregexp/o;     # anchored since new symbol will be at start

my $t = time();

my @avoid_counts;              # a(n)
my @strings = ('');
for (my $len = 0; ; $len++) {
  $avoid_counts[$len] = scalar(@strings);
  last unless @strings;
  my @new_strings;
  my $not_count = 0;
  foreach my $string (@strings) {
    foreach my $symbol (0 .. $radix-1) {
      my $new_string = $symbol . $string;
      if ($new_string =~ $Uregexp) {  # encounters Udelta
        $not_count++;
      } else {                        # avoids Udelta
        push @new_strings, $new_string;
      }
    }
  }
  print "len=$len  in ",scalar(@strings),
    " out ",scalar(@new_strings),"  not $not_count\n";
  @strings = @new_strings;
}

my $total = 0;
foreach my $len (0 .. $#avoid_counts) {
  print $avoid_counts[$len],',';
  if ($len==18 || $len==30) { print "\n"; }
  $total += $avoid_counts[$len];
}
print "\ntotal $total\n";

$t = (time() - $t)/60;
printf "runtime %.1f minutes\n", $t;


use Devel::Mallinfo;

my $h = Devel::Mallinfo::mallinfo();
print "mallinfo:\n";
my $field;
foreach my $field (sort keys %$h) {
  printf "  %-10s  %7d\n", $field, $h->{$field};
}

exit 0;
