\\ Copyright 2019, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


default(recover,0);
default(strictargs,1);
read("dragon-defines.gp");

check_count = 0;
check_equal(got,want,name="") =
{
  check_count++;
  if(!(got===want),
     print("wrong "name);
     print("got  "got);
     print("want "want);
     quit(1),
     print("ok "check_count" "name));
}


\\-----------------------------------------------------------------------------
\\ Kempner Anormal Number Systems
\\ 
\\     A. J. Kempner, "Anormal Systems of Numeration", American Mathematical
\\     Monthly, volume 43, number 10,  December 1936, pages 610-617.
\\     http://www.jstor.org/stable/2300532

\\ a = base
\\ a^n <= c < a^(n+1)
\\ digit alpha remainder beta so  c = alpha*a^n + beta
\\ 0 < alpha <= floor(a)      integer
\\ so 0 <= beta < a^n

anormal_digits(c,base,limit=10) = {
  my(l=List([]), p=1);
  while(p*base <= c, p*=base);
  while(c && (p>1 || #l < limit),
    my(alpha=floor(c/p));
    listput(l,alpha);
    c -= alpha*p;
    p /= base);
  Vec(l);
}
check_equal(anormal_digits(13,2), [1,1,0,1]);

\\ Kempner's example 2 in base 3/2
check_equal(anormal_digits(2, 3/2), [1,0, 0,1, 0,0,0,0,0,1]);
{
  my(limit=1000, c=2); \
  check_equal(nearly_equal(fromdigits_any(anormal_digits(c, 3/2, limit),
                                          3/2)*(3/2)^-(limit-2),
                           c),
              1);;
}


\\ Kempner Anormal
\\ c = r/s rational in least terms
\\ base u/v
\\ all prime factors of s must be factors of u.v.(u^l - v^l)
\\ where l = period length
\\
\\ u=3, v=2 those primes in s
\\ other primes p in s, p coprime to u=3,v=2 so u^l-v^l eventually repeat
\\
\\ vector(10,k, 3^k-2^k)
\\ for(k=1,10, print(factor(3^k-2^k)))
find_32_power(p) = {
  my(three=Mod(3,p), two=Mod(2,p));
  for(k=1,1000,
    if(three-two == 0, return(k));
    three *= 3;
    two *= 2);
  error("not found p=",p);
}
\\ for(i=3,50, my(p=prime(i));  print(p" "find_32_power(p)));
\\ vector(15,i,i+=2; find_32_power(prime(i)))
\\ A320383 multiplicative order 3/2 mod prime(n)


\\ A058840 Renyi beta expansion of 1 in base 3/2, sequence gives y(0),y(1),...
\\ A058840 starts  1,1, 0,1,0,0,0,0,0,1
\\ 2 in 3/2 starts 1,0, 0,1,0,0,0,0,0,1
\\                      ^-- 1/2
\\ anormal_digits(2, 3/2, 10)
check_equal(anormal_digits(1, 3/2, 10), [1]);

\\ forstep(c=1/24,2,1/24, print(anormal_digits(c, 3/2, 10)))
\\
check_equal((3/2)^2, 9/4);
check_equal(9/4 - 3/2, 3/4);
\\ 3/2 <= c < 9/4        [3/2, 9/4]    c-3/2 = [0, 3/4]
\\ 1 <= c-3/2 < 3/2
\\
\\ never start 1,1
\\ (3/2)^n <= c < (3/2)^(n+1)  alpha=1
\\ d = c - (3/2)^n
\\ 0 <= d < 1/2*(3/2)^n = 3/4*(3/2)^(n-1)
\\ (3/2)^(n-1) <= d < (3/2)^n  for alpha=1
\\ but d < 3/4*(3/2)^(n-1) is too small
\\ and also never 1,1 anywhere lower
{
  my(v=anormal_digits(2, 3/2, #A058840_samples));
  check_equal(v[2], 0);
  v[2]=1; 
  check_equal(v, A058840_samples);
}

\\ v=anormal_digits(2, 3/2, 100)
\\ v[2]=1
\\ fromdigits_any(v,3/2) / (3/2.0)^(#v-2)
\\ anormal_digits(3, 3/2, 12)
\\ at n=-1 c=1/2
\\ n=0 p=1    1*1 + 1/2 = 5/2
\\ n=1 p=3/2  1*3/2 + 5/2 = 4


\\-----------------------------------------------------------------------------
\\ Base sqrt(5)

floorQuad(x) = my(ret=0); while(x>=1,ret++;x--); ret; \\ pending GP 2.13.x allowing floor() of quads
a_vector(len) = my(v=vector(len),b=quadgen(20),r=1/2); for(i=1,len, r*=b; r-=(v[i]=floorQuad(r))); v;

check_equal(a_vector(30), \
            [1,0,0,1,0,1,1,0,1,1,1,0,1,0,0,0,1,0,1,0,1,0,0,1,0,0,0,1,0,2]);
{
  my(v=a_vector(200),
     got=sum(i=1,#v, v[i]/sqrt(5)^i));
  check_equal(nearly_equal(got,1/2), 1,
             "base sqrt(5) nearly equal 1/2");
}

\\ a - b*sqrt(5) - d >= 0
\\ a - d >= b*sqrt(5)
\\ (a - d)^2 >= 5*b^2
\\ a^2 - 2ad + d^2 >= 5*b^2
\\ d^2 - 2ad >= 5*b^2 - a^2
\\ 2ad - d^2 <= a^2 - 5*b^2
\\ either
\\ 4a - 4 <= a^2 - 5*b^2     (a-2)^2 >= 5*b^2    a-2 >= sqrtint(5*b^2)
\\ 2a - 1 <= a^2 - 5*b^2     (a-1)^2 >= 5*b^2    a-1 >= sqrtint(5*b^2)
\\
\\ a/2 + base*b/2 - d >=0
\\ d <= a/2 + base*b/2
\\ d <= a/2 + sqrtint(base*b^2/4)     if a>0
\\
a_vector_other(len) =
{
  my(v=vector(len),
     base=quadgen(20),
     a=1, b=0,
     r=1/2);
  for(i=1,len,
     print("at i="i" r="r"  "r*1.0);
     [a,b]=[5*b,a];
     r*=base;
     r == (a + base*b)/2 || error();
     my(d = floorQuad(r));
     my(d2 = (a + sign(b)*sqrtint(5*b^2)) \2 );
     if(d != d2,
        print("  d="d"  d2 = "d2"   a="a" b="b));
     print(d <= a/2-ceil(sqrt(5)*b/2));
     v[i]=d;
     r-=d;
     a -= 2*d;
  );
  v;
}
check_equal(a_vector(20), \
            a_vector_other(20));


\\-----------------------------------------------------------------------------
\\ A058840 Code

a_vector(len) = my(v=vector(len),c=2,d=1); for(i=1,len, if(c>=d,c-=d;v[i]=1); c*=3;d<<=1); v;

{
  my(v=OEIS_samples("A058840"));
  print("A058840 samples size ",#v);
  check_equal(a_vector(#v), v,
              "A058840 a_vector()");
}
{
  my(g=OEIS_bfile_gf("A058840"));
  print("A058840 b-file max term ",poldegree(g));
  check_equal(Polrev(a_vector(poldegree(g)+1)),  g);

  my(v=anormal_digits(2,3/2, poldegree(g)+1));
  check_equal(v[2], 0,    \\ 0 in anormal
              "anormal_digits 2 in base 3/2 second term");
  v[2] = 1;  \\ 1 in A058840
  check_equal(Polrev(v),  g);
}



\\-----------------------------------------------------------------------------
\\ A058840 Renyi Beta
\\ 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0,
\\
\\ A058840 x,y sequence descriptions seem not quite right

Renyi_Beta_x_vector(x,r,len) =
{
  (1 < r && r < 2) || error();    \\ base
  (0 <= x && x <= 1) || error();  \\ initial value
  my(ret=vector(len));
  for(i=1,len,
     ret[i]=x;
     if(r*x>1, x=r*x-1,
               x=r*x));
  ret;
}
check_equal(Renyi_Beta_x_vector(1, 3/2, 8), \
            [1, 1/2, 3/4, 1/8, 3/16, 9/32, 27/64, 81/128]);

Renyi_Beta_y_vector(x,r,len) =
{
  my(v=Renyi_Beta_x_vector(x,r,len-1));
  \\ v=v[^1];            \\ y(i) = x(i+1) so y(0)=x(1)
   v=concat(1,v);    \\ y(i+1) = x(i) so y(1)=x(0)
  v=apply(x->x*r, v);
  apply(floor,v);
}
check_equal(Renyi_Beta_y_vector(1, 3/2, 10), \
            [1, 1, 0, 1, 0, 0, 0, 0, 0, 1], "y vector");
check_equal(Renyi_Beta_y_vector(1, 3/2, #A058840_samples), \
            A058840_samples);


Renyi_Beta_digits(x,r,len) = {
  (1 <= r && r <= 2) || error();  \\ base
  my(ret=vector(len));
  ret[1]=1;
  for(i=2,len,
    \\ print("at i="i" x="x"  digits "ret[1..i-1]);
    x *= r;
    x -= (ret[i] = floor(x));
  );
  ret;
}
check_equal(Renyi_Beta_digits(1, 3/2, 10), \
            [1, 1, 0, 1, 0, 0, 0, 0, 0, 1]);
check_equal(Renyi_Beta_digits(1, 3/2, #A058840_samples), \
            A058840_samples);
check_equal(fromdigits_any([1],3/2),  1);
\\ fromdigits_any(Renyi_Beta_digits(1, 3/2, 100),3/2)*(3/2)^-99.0



\\-----------------------------------------------------------------------------



\\-----------------------------------------------------------------------------

\\ starting 0, q-development of x=1

v=[1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0]
subst(Polrev(v),'x,1/(3/2))*1.0  *(3/2)^0
v[2]=0
subst(Polrev(v),'x,1/(3/2))*1.0  *(3/2)^0
0 + 1/(3/2) + 1/(3/2)^3 + 1/(3/2)^9 + 0.0
print();
quit


g=OEIS_bfile_gf("A058840");
subst(g,'x,1/(3/2))*1.0     \\ 2.0
g+O(x^12)
print();

\\ 1 0    2           8
\\ 1,0, 0,1,0,0,0,0,0,1
3/2 + 1/(3/2)^2 + 1/(3/2)^8 + 0.0     \\ 2.0

quit



\\-----------------------------------------------------------------------------
r=1;
{
  for(i=1,12,
     r*=3/2;
     e=floor(r);
     print("e ",e);
     r-=e);
}
quit


\\-----------------------------------------------------------------------------
\\ Renyi

\\ A. Renyi, Representation for Real Numbers and Their Ergodic Properties,
\\ Acta. Math. Acad. Sci. Hung., volume 8, 1957, pages 477-493.
\\ 
\\ Positive function y=f(x) and represent real number x as
\\     x = e0 + f(e1 + f(e2 + f(e3 + ...)))
\\ Digits e = e_n(x)    n=0,1,... 
\\ Remainders r_n(x) = f(e_{n+1} + f(e_{n+2} + f(e_{n+3} + ...)))
\\
\\ e_0(x) = floor(x)                    r_0(x) = frac(x)
\\ e_{n+1}(x) = floor(phi( r_n(x) ))    r_{n+1}(x) = frac(phi( r_n(x) ))
\\ phi inverse func of f, so x=phi(y) for y=f(x)
\\
\\ f(x) = x/q   q=2,3,...
\\ usual fraction digits
\\     inf  e_n
\\ x = sum  ---
\\     n=0  q^n
\\
\\ f(x) = 1/x   continued fraction representation of x

base=2;
base=3/2;
f(x) = x/base;
phi(x) = x*base;
phi(f('x)) == 'x || error();

{
  \\ my(f = (x)->1/x,
  \\    phi = (x)->1/x);
  \\ phi(f('x)) == 'x || error();

  Renyi_e_vector(x,len) = 
  print("Renyi_e_vector() x="x);
  my(v=vector(len),
     e=floor(x),
     r=frac(x));
  for(i=1,#v,
     print("at i="i"  e="e" r="r);
     v[i]=e;
     e = floor(phi(r));
     r = frac(phi(r)));
  v;
}

\\ Renyi_e_vector(1/7,15)
check_equal(2/(3/2), 4/3);
check_equal(f(3/2), 1);
check_equal(f(1), 2/3);
check_equal(f(f(1)), 4/9);
\\ check_equal(f(3/4)+1, 9/8);
check_equal(phi(3/2)-1, 5/4);
check_equal(4/3+3/2, 17/6);
check_equal(4/3 *3/2,  2);    \\ shifted up one place to Kempner 2
x = 1+f(3/2)
x=5/4
x=1+1/8
x=0+3/4
x=1+1/2
x=1+f(3/2)
 Renyi_e_vector(x,12)
\\ check_equal(Renyi_Beta_y_vector(1, 3/2, 10), \
\\             [1, 1, 0, 1, 0, 0, 0, 0, 0, 1]);
quit

\\-----------------------------------------------------------------------------
\\ 100 100 100 repeating

check_equal(vecper_frac_fromdigits([[], [1,0,0]],3/2), 18/19);
check_equal(anormal_digits(18/19, 3/2, 20), \
            [0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1]);

fromdigits_any([1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0], 3/2) * (3/2)^-46.0
f=fromdigits_any(anormal_digits(18/19, 3/2, 100)[^2], 3/2) * (3/2)^-99.0
bestappr(f,1000)


\\-----------------------------------------------------------------------------
\\ 2 canonical 1s to 0s

vector_substrings_set(v,len) = \
  Set(vector(#v-(len-1),i, v[i..i+len-1]));

{
  my(v=anormal_digits(18/19,3/2,3000));
  print(vector_substrings_set(v,1));
  print(vector(14,len, #vector_substrings_set(v,len)));

  v=vector_cumulative(v) * 1.0;
  for(i=1,#v,v[i]/=i);
  v=v[100..#v];
  plothraw([1..#v],v,1);
  print(v[#v]);
}


\\-----------------------------------------------------------------------------
print("end");
quit