#!/usr/bin/perl -w

# Copyright 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use Carp 'croak';
use List::Util 'max','min','sum','pairmap';
use Math::BaseCnv 'cnv';
use FLAT;

use Math::PlanePath::DragonCurve 118; # v.118 for xyxy_to_n_either()
use Math::PlanePath::Base::Digits
  'bit_split_lowtohigh','digit_join_lowtohigh';

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyDragon ':all';
use MyPrintwrap ':all';
use MyFLAT ':all';

our @ipc;
$|=1;


# uncomment this to run the ### lines
# use Smart::Comments;

my $path = Math::PlanePath::DragonCurve->new;
my $any_re = '(0|1)*';

sub FLAT_check_disjoint {
  my ($f1, $f2) = @_;
  my $f1_name = $f1->{'name'} // 'first';
  my $f2_name = $f2->{'name'} // 'second';
  my $intersect = $f1->as_dfa->intersect($f2->as_dfa);
  if (! $intersect->is_empty) {
    my $it = $intersect->new_acyclic_string_generator;
    # my $it = $intersect->new_deepdft_string_generator(10);
    print "common to supposed disjoint $f1_name and $f2_name\n";
    if ($f1->equals($f2)) {
      print "  $f1_name equals $f2_name\n";
    } elsif ($f1->is_subset_of($f2)) {
      print "  $f1_name is subset of $f2_name\n";
    } elsif ($f2->is_subset_of($f1)) {
      print "  $f2_name is subset of $f1_name\n";
    }
    my $count = 0;
    while (my $str = $it->()) {
      if (++$count > 40) {
        print "  ... and more\n";
        last;
      }
      print "  $str\n";
    }
    croak "not disjoint";
  }
  print "$f1_name disjoint $f2_name, ok\n";
}

# $fa should be a FLAT::DFA.
# Return a new FLAT::DFA which matches strings which will go to accepting
# from any starting state of $fa.
sub FLAT_accepting_from_anywhere {
  my ($fa) = @_;
  my $intersection = $fa;
  foreach my $starting ($fa->get_states) {
    my $fstarting = $fa->MyFLAT::as_nfa;
    $fstarting->unset_starting($fstarting->get_starting);
    $fstarting->set_starting($starting);
    $intersection = $intersection->intersect($fstarting->as_dfa);
  }
  $intersection->{'name'} = join(' ',
                                 $fa->{'name'} // (),
                                 'accept from anywhere');
  return $intersection;
}

sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  ### xyxy_is_traversed_either(): "$x1,$y1 to $x2,$y2"
  my $arms = $path->arms_count;
  foreach my $n1 ($path->xy_to_n_list($x1,$y1)) {
    next if defined $n_limit && $n1 > $n_limit;
    foreach my $n2 ($path->xy_to_n_list($x2,$y2)) {
      next if defined $n_limit && $n2 > $n_limit;
      if (abs($n1-$n2) == $arms) {
        ### yes: "$n1 to $n2"
        return 1;
      }
    }
  }
  return 0;
}

# Return a FLAT::DFA which matches strings of anything in alphabet.
#   alphabet    => arrayref of symbols (strings)
#   alphabet_fa => FLAT::FA to get the alphabet of symbols
#   min_length  => integer>=0
sub FLAT_any {
  my %options = @_;
  my $min_length = $options{'min_length'} || 0;
  my $alphabet   = $options{'alphabet'}
    // do {
      my $fa = $options{'alphabet_fa'};
      (defined $fa ? [ $fa->alphabet ] : undef)
    }
    // croak "No alphabet specified";

  require FLAT::DFA;
  my $fa = FLAT::DFA->new;
  my @states = $fa->add_states($min_length+1);
  $fa->set_starting($states[0]);
  foreach my $i (0 .. $#states-1) {
    $fa->add_transition($states[$i], $states[$i+1], @$alphabet);
  }
  $fa->add_transition($states[-1], $states[-1], @$alphabet);
  $fa->set_accepting($states[-1]);
  return $fa;
}

# Return a FLAT::DFA derived from $fa.
# Among the strings matched by $fa, any string with a proper substring also
# matched by $fa is removed.
sub FLAT_remove_own_superstrings {
  my ($fa) = @_;
  my $any = FLAT_any (min_length=>1, alphabet_fa=>$fa);
  my $new_fa = $fa->difference($any->MyFLAT::concat($fa));
  $new_fa = $new_fa->difference($fa->MyFLAT::concat($any));
  if (defined (my $name = $fa->{'name'})) {
    $new_fa->{'name'} = "$name sans superstrings";
  }
  return $new_fa;
}


#------------------------------------------------------------------------------

sub path_bitstr_to_conf {
  my ($str) = @_;
  my $n_limit = 1 << length($str);
  my $n = cnv($str,2,10);
  my $side;
  my ($x1,$y1, $x2,$y2);
  if ($n & 1) {
    ($x1,$y1) = $path->n_to_xy($n+1);
    ($x2,$y2) = $path->n_to_xy($n);
    $side = 1;
  } else {
    ($x1,$y1) = $path->n_to_xy($n);
    ($x2,$y2) = $path->n_to_xy($n+1);
    $side = 0;
  }
  my $dx = $x2 - $x1;
  my $dy = $y2 - $y1;
  my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
  my @segs;
  foreach my $s (@$surround_segs_all) {
    my ($ox,$oy, $dir) = @$s;
    my $px1 = $x1 + $ox*$dx + $oy*$lx;
    my $py1 = $y1 + $ox*$dy + $oy*$ly;
    my ($pdx,$pdy) = ($dx,$dy);
    foreach (1 .. $dir) { ($pdx,$pdy) = xy_rotate_plus90($pdx,$pdy); }
    my $px2 = $px1 + $pdx;
    my $py2 = $py1 + $pdy;
    if (path_xyxy_is_traversed_either($path, $px1,$py1, $px2,$py2, $n_limit)) {
      push @segs, $s;
    }
  }
  return { segs => \@segs, side => $side };
}

# Return true if $fa at $state is accepting and also any number of following
# 0s are accepted there too.
sub FLAT_state_accepts_zeros {
  my ($fa, $state) = @_;
  my %seen;
  until ($seen{$state}++) {
    $fa->is_accepting($state) or return 0;
    my @next = $fa->successors($state,'0');
    @next==1 or return 0;
    $state = $next[0];
  }
  return 1;
}

#==============================================================================


{
  # fRpred terminating fractions
  my $f = fRpred_flat();
  foreach my $len (1 .. 5) {
    # $f = $f->difference($f->MyFLAT::concat(
    #                     FLAT::Regex->new(('0'x$len).'1')->as_dfa))
    #   ->as_dfa;

    $f = $f->intersect($f->MyFLAT::concat(FLAT::Regex->new('0'x$len)->as_dfa))
      ->as_dfa;
    $f = MyFLAT::prefix_accepting($f);
    $f = $f->MyFLAT::minimize;
  }

  # my $f = FLAT::Regex->new('(0|1)* 0*')->as_dfa;
  # $f = MyFLAT::prefix_accepting($f);
  # $f = $f->intersect($fRpred_surround_flat);

  # FLAT_show_breadth($f,5);
  # $f = $f->MyFLAT::reverse;
  $f = $f->MyFLAT::minimize;
  MyFLAT::view($f);

  FLAT_check_is_equal ($f, SpredRight_k_flat());
  exit 0;
}
{
  # fRpred terminating fractions xxx0000...

  my $f = fRpred_flat();
  for (my $more = 1; $more--; ) {
    foreach my $state ($f->get_accepting) {
      if (! FLAT_state_accepts_zeros($f,$state)) {
        $f->unset_accepting($state);
        $more = 1;
      }
    }
  }

  # $f = $f->MyFLAT::minimize;
  MyFLAT::view($f);
  my $s = SpredLeft_k_flat()->union(fraction_flat(0,1,2))->as_dfa;
  $s = $s->MyFLAT::minimize;
  MyFLAT::view($s);
  FLAT_check_is_equal ($f, $s);
  exit 0;
}
{
  # view of fBpred
  FLAT_print_tikz(MyFLAT::separate_sinks(fNonRpred_flat()));
  # MyFLAT::view(MyFLAT::separate_sinks(NonRpred_flat()));
  exit;

  fBpred_surround_init();
  my $r = $fNonRpred_surround_flat->MyFLAT::minimize;
  $r = MyFLAT::separate_sinks($r);
  FLAT_print_tikz($r);
  MyFLAT::view($r);
  my $l = $fNonLpred_surround_flat->MyFLAT::minimize;
  $l = MyFLAT::separate_sinks($l);
  MyFLAT::view($l);
  # my $b = $fNonBpred_surround_flat->MyFLAT::minimize;
  # MyFLAT::view($b);
  exit 0;
}
{
  # fBpred_cases_extra memoizing
  fBpred_cases_extra('left',0,11);
  fBpred_cases_extra('left',0,11);
  exit 0;
}

{
  # fRpred minimized state table prints
  fBpred_surround_init();
  foreach my $f ($fNonRpred_surround_flat,
                 $fNonLpred_surround_flat) {
    my $name = $f->{'name'};
    $name =~ s/ .*//;
    my $f = $f->as_dfa;
    $f = $f->MyFLAT::minimize;

    FLAT_print_perl_table($f,'state_table');
    FLAT_print_perl_accepting($f,'accepting');
    print "my \$starting = ",$f->get_starting,";\n";

    # FLAT_print_gp_inline_table($f,"${name}_state_table");
    # FLAT_print_gp_inline_accepting($f,'accepting');
  }
  exit 0;
}
{
  my $f;
  # $f = fraction_flat(2, 3, 2);
  # $f = fraction_flat(11, 15, 2);
  $f = fraction_flat(123, 1000, 10);
  MyFLAT::view($f);
  # $f = $f->as_min_dfa;
  # MyFLAT::view($f);
  exit 0;
}
{
  # strings common between extras
  my @extras_R_array = fRpred_cases_extra_list();
  # $extra_plus1_right_flat,
  # $extra_minus11_right_flat,
  # $extra_minus7_right_flat,
  FLAT_show_breadth($extras_R_array[0]->intersect($extras_R_array[1]),7);
  exit 0;
}
{
  my $fRpred_by_cases = Rpred_flat()->union(fRpred_cases_extra_list())->as_dfa;
  fBpred_surround_init();
  FLAT_check_is_equal ($fRpred_by_cases,
                       $fRpred_surround_flat);
}
{
  my $extra_plus11_left_flat
    = do {
      my $f = SpredLeft_k_flat()->intersect(bits_N_even_flat())
        ->MyFLAT::concat(fraction_flat(11,15, 2));
      $f = MyFLAT::prefix_accepting($f);
      $f->{'name'} = 'extra_plus11_left';
      $f;
    };
  FLAT_check_is_equal(fBpred_cases_extra('left',0,11),
                      $extra_plus11_left_flat);
  exit 0;
}
{
  # string which reaches accepting from anywhere

  # NonRpred
  # 10101
  # 10110
  # 10111
  # 11000
  # 10100101
  # 10100110
  # 10100111
  # 11001000
  # 10100100101
  # 10100100110
  # 10100100111
  # 11001001000

  # my $f = DpredLeft_k_flat();
  my $f;
  # $f = Lpred_k_flat();
  # $f = LpredInf_flat();
  # $f = SpredLeft_k_flat();
  # $f = SpredRight_k_flat();
  # $f = Spred_k_flat();
  # $f = SpredRight_runs_flat();
  # $f = DpredLeft_runs_flat();
  # $f = DpredLeft_runs_flat()->union(DpredRight_runs_flat());
  $f = NonRpred_flat();
  # MyFLAT::view(FLAT_any(min_length=>1, alphabet_fa=>$f));
  # MyFLAT::view($f);
  my $int = FLAT_accepting_from_anywhere($f);
  $int = $int->MyFLAT::minimize;
  $int = FLAT_remove_own_superstrings($int);
  $int = $int->MyFLAT::minimize;

  $int = $int->difference($int
                          ->MyFLAT::concat(FLAT::Regex->new('(0|1) (0|1)*'))
                          ->as_dfa);
  $int = $int->difference(FLAT::Regex->new('(0|1) (0|1)*')
                          ->MyFLAT::concat($int)->as_dfa);

  $int = $int->MyFLAT::minimize;
  MyFLAT::view($int);

  FLAT_show_breadth($int, 12);
  FLAT_show_acyclics($int);
  exit 0;
}

{
  # is_finite() not right when cycle not includes an accepting state

  my $f = FLAT::Regex->new('a* b');
  $f = $f->as_nfa;
  # $f = $f->as_min_dfa;
  # MyFLAT::view($f);
  print $f->is_infinite ? "yes\n" : "no\n";
  exit 0;
}

{
  $triple_15ths_union_segs->print;
  exit 0;
}


{
  # eventually enclosed configurations intersections

  fBpred_surround_init();
  foreach my $f ($fNonRpred_surround_flat,
                 $fNonLpred_surround_flat,
                 $fNonBpred_surround_flat) {
    my @intersection;
    my $count = 0;
    foreach my $state (0 .. $#fBpred_surround_state_to_conf) {
      next unless $f->is_accepting($state);
      $count++;
      my $conf = $fBpred_surround_state_to_conf[$state];
      $intersection[$conf->{'side'}] //= $conf;
      $intersection[$conf->{'side'}] = $intersection[$conf->{'side'}]->intersection($conf);
    }
    print "$f->{name} count $count accepting, intersection is\n";
    $intersection[0]->print;
    $intersection[1]->print;
  }
  exit 0;
}

{
  # initial prefixes of f bits recognised as non-boundary
  fBpred_surround_init();

  my $len = 8;
  FLAT_show_breadth(NonBpred_k_flat(),$len);
  FLAT_show_breadth($fNonBpred_surround_flat,$len);

  # FLAT_show_breadth($fNonBpred_surround_flat,$len);

  # my $f = bits_of_length_flat(6);
  # my $int = NonBpred_k_flat()->intersect($f);
  # if ($int->is_finite) {
  #   print "finite\n";
  # }
  # FLAT_show_acyclics($int);
  #
  # $int = $fNonBpred_surround_flat->intersect($f);
  # if ($int->is_finite) {
  #   print "finite\n";
  # }
  # $int = $int->trim_sinks;
  # $int = $int->as_min_dfa;
  # MyFLAT::view($int);
  # # FLAT_show_acyclics($int);

  exit 0;
}

{
  # print state tables fBpred surround for dragon.tex

  fBpred_surround_init();
  FLAT_print_gp_inline_table($fNonRpred_surround_flat,
                             'fBpred_surround_state_table');
  FLAT_print_gp_inline_accepting($fNonRpred_surround_flat,
                                'fBpred_surround_state_R_enclosed');
  FLAT_print_gp_inline_accepting($fNonLpred_surround_flat,
                                'fBpred_surround_state_L_enclosed');
  FLAT_print_gp_inline_accepting($fNonBpred_surround_flat,
                                'fBpred_surround_state_B_enclosed');
  exit 0;
}



{
  # view

  my $f = FLAT::Regex->new("( $DpredRight_k_flat_fre
                            | $DpredLeft_k_flat_fre )")->as_dfa;
  $f = Dpred_k_flat();
  $f = SpredLeft_k_flat();
  $f = DpredLeft_k_flat();
  $f = DpredLeft_runs_flat();
  $f = Spred_k_flat();
  $f = $f->MyFLAT::minimize;
  MyFLAT::view($f);
  # $f = $f->MyFLAT::reverse;
  # $f = $f->MyFLAT::minimize;
  # MyFLAT::view($f);
  exit 0;
}

{
  # 10101 reaches accepting in DpredLeft, but doesn't remain there
  my $f;
  $f = FLAT::Regex->new('(0|1)* 10101')->as_dfa;
  $f = FLAT::Regex->new('(0|1)* 11001')->as_dfa;
  FLAT_check_is_subset($f, DpredLeft_k_flat());
  FLAT_check_disjoint($f, SpredLeft_k_flat());
  exit 0;
}
{
  # Bpred and cases vs surround
  fBpred_surround_init();

  my $s = $fBpred_surround_flat;
  my $c = fBpred_cases_flat();

  $s = $s->MyFLAT::minimize;
  $c = $c->MyFLAT::minimize;

  print "fBpred_cases num states    ", $c->num_states,"\n";
  print "fBpred_surround num states ", $s->num_states,"\n";

  FLAT_check_is_equal($s, $c);

  $s = MyFLAT::prefix_accepting($s);
  my $suffix = FLAT::Regex->new('(0|1)* 0111 0111 0111 0111 0111')->as_dfa;
  $suffix = $suffix->MyFLAT::minimize;
  my $diff = $s->intersect($suffix);
  $diff->{'name'} = 'diff';
  FLAT_show_acyclics($diff);
  MyFLAT::view($diff);
  FLAT_show_breadth($diff, 10);
  exit;

  FLAT_check_is_equal(fBpred_cases_flat()->complement, fNonBpred_cases_flat());

  exit 0;

  {
    my $f = SpredLeft_k_flat()
      ->MyFLAT::concat(FLAT::Regex->new('0'),    # 2n
                       # FLAT::Regex->new('(0001)'),
                       # fraction_flat(1,15, 2),   # +1/15
                      );
    $f->{'name'} = 'try';
    #    $f = MyFLAT::prefix_accepting($f);
    FLAT_check_is_subset($f, $fBpred_surround_flat);
    # FLAT_check_is_subset($extra_plus1_left_flat, $fBpred_surround_flat);
  }

  # FLAT_show_breadth($fBpred_surround_flat, 6);

  #    9<----8
  #    ^
  #    |
  # 01111011  N=123
  # 011110111  N=247
  # 0111101110  N=494
  # 01111011101  N=989
  # 0111011  N=59
  # 01110111  N=119
  # 011101110  N=238
  # 0111011101  N=477
  # 00111011  N=59
  # 001110111  N=119
  # 0011101110  N=238
  # 00111011101  N=477
  #
  my $str = '011110100101';
  $str = '0010111100001001';
  $str = '101011';
  $str = '10001011101';
  foreach my $extra (fRpred_cases_extra_list(), fLpred_cases_extra_list()) {
    if ($extra->contains($str)) {
      print "$extra->{name} contains $str\n";
    } else {
      print "$extra->{name} not\n";
    }
  }
  print "str path "; conf_print(path_bitstr_to_conf($str));
  # MyFLAT::view($fNonBpred_surround_flat);

  my $extras = NonBpred_k_flat()->difference($fNonBpred_surround_flat);
  FLAT_show_breadth($extras, 6);

  # 10101
  # extra_plus11_left contains 101 01
  # extra_minus1_right contains 10101 [1]
  # extra_minus11_right contains 10101 [1]
  #
  # 010101
  #
  # 101010
  # extra_plus11_left contains 101010
  # extra_minus1_right contains 101010
  # extra_minus11_right contains 101010
  #
  # 101011
  # extra_minus1_right contains 101011
  # extra_minus11_right contains 101011
  #

  exit 0;
}
{
  # cases vs surround on 15ths

  # k=3
  # 001011001  s=1 c=0  k=3 num=21=1+6/15 len=9(of 19)
  # 0010110011  s=1 c=0  k=3 num=21=1+6/15 len=10(of 19)
  # 010100110  s=1 c=0  k=3 num=39=2+9/15 len=9(of 19)
  # 0101001100  s=1 c=0  k=3 num=39=2+9/15 len=10(of 19)
  # 01011001  s=1 c=0  k=3 num=42=2+12/15 len=8(of 19)
  # 010110011  s=1 c=0  k=3 num=42=2+12/15 len=9(of 19)
  # 10100110  s=1 c=0  k=3 num=78=5+3/15 len=8(of 19)
  # 101001100  s=1 c=0  k=3 num=78=5+3/15 len=9(of 19)
  # 1010110  s=1 c=0  k=3 num=81=5+6/15 len=7(of 19)
  # 10101100  s=1 c=0  k=3 num=81=5+6/15 len=8(of 19)
  # 1011001  s=1 c=0  k=3 num=84=5+9/15 len=7(of 19)
  # 10110011  s=1 c=0  k=3 num=84=5+9/15 len=8(of 19)
  # 110100110  s=1 c=0  k=3 num=99=6+9/15 len=9(of 19)
  # 1101001100  s=1 c=0  k=3 num=99=6+9/15 len=10(of 19)

  fBpred_surround_init();
  my $s = $fBpred_surround_flat;
  my $c = fBpred_cases_flat();
  $s = MyFLAT::prefix_accepting($s);
  $c = MyFLAT::prefix_accepting($c);

  my $fraclen = 16;
  foreach my $k (0 .. 8) {
    print "k=$k\n";
    my $den = 15 * 2**$k;
    foreach my $num (0 .. $den-1) {
      my $nn = int(($num << $fraclen)/15);
      my $str = sprintf '%0*b', $k+$fraclen, $nn;
      my $str_len = length($str);
      foreach my $len (0 .. length($str)) {
        my $substr = substr($str,0,$len);
        my $sp = $s->contains($substr) ? 1 : 0;
        my $cp = $c->contains($substr) ? 1 : 0;
        if ($sp != $cp) {
          my $q = int($num/15);
          my $r = $num % 15;
          print "$substr  s=$sp c=$cp  k=$k num=$num=$q+$r/15 len=$len(of $str_len)\n";
        }
      }
    }
  }
  exit 0;
}

{
  # +7/15

  # length("10101110111011101110111")
  # length("101")
  # (5+7/15)/8 == 41/60

  foreach my $str ('10101110111011101110111',
                   '1010101110111011101110111',
                   '101000011110001110111011101110111',
                   '10100001111001010101110111011101110111',
                   '101000011110010100101110111011101110111',
                   '10100001111001010010101110111011101110111',
                   '101000011110010001110111011101110111',
                   '101000011110011010101110111011101110111',
                  ) {
    foreach my $extra (fRpred_cases_extra_list(), fLpred_cases_extra_list()) {
      print $extra->contains($str) ? 1 : 0;
    }
    print "\n";
  }
  exit 0;
}


{
  # 14 bits fNonBpred fully surrounded
  # 13 bits fNonBpred_eventually surrounded
  # 10 bits NonBpred surrounded              11011 00111 00100 11101
  fBpred_surround_init();
  my $str = join('',map{Paper($_)}1..20);
  print "$str\n";
  my $f;
  $f = $fNonBpred_surround_flat;
  $f = NonBpred_k_flat();
  FLAT_show_transitions($f, $str);
  foreach my $len (8 .. 15) {
    print "len=$len ",$f->contains(substr($str,0,$len)) ? "yes\n" : "no\n";
  }
  exit 0;
}

{
  # extra bit patterns
  # FLAT_show_breadth($extra_plus1_left_flat, 3);
  # my $int = NonBpred_k_flat()->intersect($extra_plus1_right_flat);
  # FLAT_show_breadth($int, 12);
  exit 0;
}

{
  # Bpred from surrounds
  # MyFLAT::view(Bpred_k_flat());

  fBpred_surround_init();

  # {
  #   my $str = '010110';
  #   my $n = cnv($str,2,10);
  #   my $f1 = Bpred_k_flat();
  #   my $f2 = $Bpred_by_surround_flat;
  #   print "path N=$n right boundary ",$path->_UNDOCUMENTED__n_segment_is_right_boundary($n)?1:0, "\n";
  #   my $width = max(length($f1->{'name'}), length($f2->{'name'}));
  #   printf "%-*s contains %d\n", $width, $f1->{'name'}, $f1->contains($str)?1:0;
  #   printf "%-*s contains %d\n", $width, $f2->{'name'}, $f2->contains($str)?1:0;
  #   printf "%-*s ", $width, $f1->{'name'}; FLAT_show_transitions($f1, $str);
  #   printf "%-*s ", $width, $f2->{'name'}; FLAT_show_transitions($f2, $str);
  #   print "Rpred "; FLAT_show_transitions(Rpred_flat(), $str);
  #   print "Lpred "; FLAT_show_transitions(Lpred_k_flat(), $str);
  #   print "Bpred "; FLAT_show_transitions(Bpred_k_flat(), $str);
  #   printf "Rpred contains %d\n", Rpred_flat()->contains($str)?1:0;
  #   printf "Lpred contains %d\n", Lpred_k_flat()->contains($str)?1:0;
  #   printf "Bpred contains %d\n", Bpred_k_flat()->contains($str)?1:0;
  #   # print "path "; conf_print(path_bitstr_to_conf('0'));
  #   # conf_print($state_to_conf[2]);
  #   # conf_print($state_to_conf[6]);
  #   # print "path "; conf_print(path_bitstr_to_conf('110'));
  #   # conf_print($state_to_conf[11]);
  #   # print "path "; conf_print(path_bitstr_to_conf('1101'));
  #   # conf_print($state_to_conf[21]);
  #   # print "path "; conf_print(path_bitstr_to_conf('11011'));
  #   # conf_print($state_to_conf[39]);
  #   # conf_print($state_to_conf[58]);
  #   print "str path "; conf_print(path_bitstr_to_conf($str));
  # }

  {
    my $f = $fBpred_surround_flat;
    print "minimize $f->{'name'} ...";
    $f = $f->as_min_dfa;
    my $num_states = $f->num_states;
    print "done, num states $num_states\n";
  }
  {
    my $f = $fNonBpred_surround_flat;
    print "minimize $f->{'name'} ...";
    $f = $f->as_min_dfa;
    my $num_states = $f->num_states;
    print "done, num states $num_states\n";
  }

  # FLAT_check_is_equal($Bpred_by_surround_flat,
  #                     $NonBpred_by_surround_flat->complement);
  # FLAT_check_is_equal(NonBpred_k_flat(), $NonBpred_by_surround_flat);
  # FLAT_check_is_equal(Bpred_k_flat(),    $Bpred_by_surround_flat);

  FLAT_check_is_subset(MyFLAT::eventually_accepting($fNonRpred_surround_flat),
                       MyFLAT::eventually_accepting(NonRpred_flat()));
  FLAT_check_is_subset(MyFLAT::eventually_accepting($fNonLpred_surround_flat),
                       MyFLAT::eventually_accepting(NonLpred_k_flat()));
  FLAT_check_is_subset(MyFLAT::eventually_accepting($fNonBpred_surround_flat),
                       MyFLAT::eventually_accepting(NonBpred_k_flat()));


  FLAT_check_is_equal(NonBpred_k_flat(), NonRpred_flat()->intersect(NonLpred_k_flat()));
  FLAT_check_is_equal($fNonBpred_surround_flat,
                      $fNonRpred_surround_flat->intersect($fNonLpred_surround_flat));

  FLAT_check_accepting_remain_so(NonRpred_flat());
  FLAT_check_accepting_remain_so(NonLpred_k_flat());
  FLAT_check_accepting_remain_so(NonBpred_k_flat());
  # FLAT_check_accepting_remain_so($NonRpred_by_surround_flat);
  # FLAT_check_accepting_remain_so($NonLpred_by_surround_flat);
  # FLAT_check_accepting_remain_so($NonBpred_by_surround_flat);
  FLAT_check_accepting_remain_so($fNonRpred_surround_flat);
  FLAT_check_accepting_remain_so($fNonLpred_surround_flat);
  FLAT_check_accepting_remain_so($fNonBpred_surround_flat);

  # print $fNonRpred_surround_flat->as_summary;
  # FLAT_check_is_equal(Rpred_flat(), $Rpred_by_surround_flat);
  # FLAT_check_is_equal(Lpred_k_flat(), $Lpred_by_surround_flat);
  # FLAT_check_is_equal(NonRpred_flat(), $NonRpred_by_surround_flat);
  # FLAT_check_is_equal(NonLpred_k_flat(), $NonLpred_by_surround_flat);

  # fNonB subset of NonB since two squares of NonB are not full for fNonB
  # fNonR subset of NonR since one square of NonR not fully for fNonR
  FLAT_check_is_subset($fNonRpred_surround_flat, NonRpred_flat());
  FLAT_check_is_subset($fNonLpred_surround_flat, NonLpred_k_flat());
  FLAT_check_is_subset($fNonBpred_surround_flat, NonBpred_k_flat());

  # B subset of fB since some enclosed B are still ok for fB
  # R subset of fR since some enclosed R are still ok for fR
  FLAT_check_is_subset(Bpred_k_flat(), $fBpred_surround_flat);
  FLAT_check_is_subset(Rpred_flat(), $fRpred_surround_flat);
  FLAT_check_is_subset(Lpred_k_flat(), $fLpred_surround_flat);

  # {
  #   my $str = '01010011';
  #   $str = '01111010011';
  #   $str = '11011';
  #   my $n = cnv($str,2,10);
  #   print "path N=$n right boundary ",$path->_UNDOCUMENTED__n_segment_is_right_boundary($n)?1:0, "\n";
  #   print "NonRpred_flat contains ",NonRpred_flat()->contains($str)?1:0, "\n";
  #   print "fNonRpred_surround_flat contains ",$fNonRpred_surround_flat->contains($str)?1:0, "\n";
  #   FLAT_show_transitions(NonRpred_flat(), $str);
  #   FLAT_show_transitions($fNonRpred_surround_flat, $str);
  #   conf_print($state_to_conf[21]);
  #   print "path "; conf_print(path_bitstr_to_conf($str));
  # }

  # {
  #   # print $Rpred_by_surround_flat->as_summary;
  #   # 1* (1)-> 5* (0)-> 12* (0)-> 22* (0)-> 37
  #   # FLAT_show_transitions($Rpred_by_surround_flat, '1000');
  #   conf_print($state_to_conf[0]);
  #   conf_print($state_to_conf[3]);
  #   conf_print($state_to_conf[8]);
  #   conf_print($state_to_conf[17]);
  #   conf_print($state_to_conf[30]);
  #   # exit;
  # }

  # {
  #   my $str = '11010010011';
  #   $str = '1011';
  #   $str = '01010011';
  #   my $n = cnv($str,2,10);
  #   print "path N=$n right boundary ",$path->_UNDOCUMENTED__n_segment_is_right_boundary($n)?1:0, "\n";
  #   print "Rpred_flat contains ",Rpred_flat()->contains($str)?1:0, "\n";
  #   print "fRpred_surround_flat contains ",$fRpred_surround_flat->contains($str)?1:0, "\n";
  #   FLAT_show_transitions($fRpred_surround_flat, $str);
  # }

  exit 0;
}



{
  # Dpred and Spred vs path

  # FLAT_show_breadth(Lpred_k_LtoH_flat()->MyFLAT::reverse, 6);

  #      Spred(2n) concat 1/15
  #    xxxxxxxxxxxx0 0001 0001 0001 ...
  # for -1/15 need low 1 of Spred
  # Spred(2(n-1)) concat 29/15
  #
  # my $DpredInf_runs_right_re
  #   = "([] | $any_re 0) (100* | 011*) (1 (100* | 011*))*";  # with t=1


  my $f;
  # $f = NonLpred_k_LtoH_flat();
  # $f = NonBlobPredInf_flat();
  # $f = NonRpred_flatregex();
  # $f = NonRpred_flat();
  # $f = Rpred_flat();
  # $f = Bpred_k_flat();
  # $f = zero_digits_flat(2);
  $f = SpredLeft_k_flat();
  $f = SpredRight_k_flat();
  $f = DpredLeft_k_flat();
  $f = DpredRight_k_flat();
  $f = fBpred_cases_extras_flat();
  $f = fFifthsPred_flat();
  # $f = fNonBpred_surround_flat();
  $f = fBpred_cases_flat();
  $f = $fNonRpred_surround_flat;

  # $f = $f->complement;
  # $f = $f->as_dfa;
  # $f = $f->as_nfa;
  # $f = $f->trim_sinks;
  print "minimize\n"; $f = $f->as_min_dfa; print "done\n";
  print "prune\n"; $f->prune; print "done\n";
  # $f = $f->as_regex;

  FLAT_show_breadth($f, 4, 'hightolow');

  # if ($f->can('as_summary')) { print $f->as_summary; }
  MyFLAT::view($f);
  # FLAT_show_acyclics($f);
  # FLAT_show_deep($f);
  exit 0;

}

{
  my $banana_re = 'b(an)*';
  my $banana_re_flat  = FLAT::Regex->new($banana_re);
  exit 0;
}
{
  my $f = DpredLeft_k_flat();
  my $c = FLAT::Regex->new('0')->as_dfa->MyFLAT::concat($f);
  FLAT_check_is_subset($c, $f);
  print $c->is_subset_of($f) ? 1 : 0;
  exit 0;
}


