/* Copyright 2021 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Record high run lengths occurring in wparity sequence.
   (Cf. Run lengths sequence is A083279.)

   (Might be a modest speedup to be had with some assembler "add with carry"
   rather than the C level idiom new_t < orig_t.  Calculating "3*w" with lea
   instruction not touching the carry flag could help, and applying the
   "/2"s by new destination when reach a whole word worth.)


   Run length record indices:
   1 1
   2 3
   3 4
   4 22
   5 26
   6 88
   7 174
   8 328
   9 1040
   10 8478
   11 20329
   12 24587
   13 130257
   14 616378
   15 1011132
   16 2271349

*/

#define NDEBUG 1
#include <assert.h>
#include <gmp.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/times.h>

#if NDEBUG
#define DEBUG(expr)
#else
#define DEBUG(expr)  expr
#endif

#define W_LIMIT  10*1000*1000  /* words */

typedef unsigned  WORD;  /* an unsigned type */
#define WORD_BITLEN  (sizeof(unsigned)*CHAR_BIT)
#define MASK  ((~(WORD)0) >> 1)

#define WANT_W_VALUES               0
#define WANT_ALL_RUNLENS            0  /* A083279 */
#define WANT_RECORD_RUNLENS         0
#define WANT_RECORD_RUNLEN_INDICES  1

#ifdef __GNUC__
#define LIKELY(cond)    __builtin_expect((cond) != 0, 1)
#define UNLIKELY(cond)  __builtin_expect((cond) != 0, 0)
#else
#define LIKELY(cond)    (cond)
#define UNLIKELY(cond)  (cond)
#endif

static WORD w[W_LIMIT];
WORD *w_above;
unsigned long steps;
unsigned w_nails = 0;
unsigned output_n = 1;

double
times_user (void)
{
  struct tms t;
  times(&t);
  return (t.tms_utime / (double)sysconf(_SC_CLK_TCK));
}

int
next_w (void)
{
  steps++;
#if 0
  {
    WORD c = w[0] & 1;
    WORD *p = w;
    do {
      WORD t = *p;
      WORD new_c = t >> (WORD_BITLEN-1);
      WORD orig_t = t;
      t += (t<<1) + c;
      new_c += (t < orig_t);
      *p++ = t;
      c = new_c;
    } while (LIKELY( p<w_above ));
    if (UNLIKELY( c )) {
      DEBUG(fprintf(stderr, "\nextend w_above=%u  %X %X\n", w_above-w, w[0], w[1]));
      *w_above++ = c;
    }
    assert((w[0]&1)==0);
    p = w;
    WORD t = w[0];
    do {
      __builtin_prefetch(p+1);
      WORD next_t = p[1];
      *p++ = (t>>1) | (next_t << (WORD_BITLEN-1));
      t = next_t;
    } while (LIKELY( p<w_above ));

    return (w_above - w <= W_LIMIT - 3);
  }
#endif

#if 1
  {
    WORD t = w[0];
    WORD c = t&1;
    WORD *p = w;
    do {
      assert (c==0 || c==1);
      WORD next_t = p[1];
      WORD orig_t = t;
      t += c + (t>>1) + (next_t << (WORD_BITLEN-1));
      c = t < orig_t;
      *p++ = t;
      t = next_t;
    } while (LIKELY( p < w_above ));

    if (UNLIKELY( c )) {
      *w_above++ = c;
      DEBUG(fprintf(stderr, "\nextend w_above=%u  %X %X\n", w_above-w, w[0], w[1]));
      if (UNLIKELY( w_above >= w + W_LIMIT - 3 )) {
        return(0);
      }
    }
    return(1);
  }
#endif
}

void
w_print (void)
{
  const WORD *p;
  printf("LtoH ");
  for(p=w; p<w_above; p++) {
    WORD t = *p;
    int i;
    for(i=0; i<WORD_BITLEN-w_nails; i++) {
      printf("%d", t&1);
      t>>=1;
    }
    if (t) {
      printf("[");
      while (t) {
        printf("%d", t&1);
        t>>=1;
      }
      printf("]");
    }
    printf(",");
  }
  printf("\n");
}

int
by_full (void)
{
  double tstart = times_user();
  unsigned runlen = 0;
  unsigned long run_number = 1;
  unsigned max_runlen = 0;
  WORD prev_wparity = 1;
  memset(w,'\0',sizeof(w));
  w[0] = 1;
  w_above = w+1;
  w_nails = 0;
  steps = 0;
  for (;;) {
    DEBUG(printf("steps=%lu: ",steps); w_print());

    if (! next_w()) {
      break;
    }

    WORD wparity = w[0] & 1;
    if (wparity == prev_wparity) {
      runlen++;
    } else {
      if (WANT_ALL_RUNLENS) {
        printf ("%u %u\n", output_n++, runlen);
      }
      if (UNLIKELY( runlen > max_runlen )) {
        if (WANT_RECORD_RUNLENS) {
          printf ("%u %u\n", output_n++, runlen);
        }
        if (WANT_RECORD_RUNLEN_INDICES) {
          printf ("%u %lu\n", output_n++, run_number);
        }
        max_runlen = runlen;
      }
      prev_wparity = wparity;
      runlen = 1;
      run_number++;
    }
  }
  fprintf(stderr, "\n");
  fprintf(stderr, "stop at w_len=%u steps=%lu full time %.1f seconds\n",
          w_above-w, steps, times_user()-tstart);
  DEBUG(fprintf(stderr, "  "); w_print());
  return (0);
}

int
by_nails (void)
{
  struct tms t; times(&t);
  unsigned runlen = 0;
  unsigned long run_number = 1;
  unsigned max_runlen = 0;
  WORD prev_wparity = 1;

  memset(w,'\0',sizeof(w));
  w[0] = 1;
  w_above = w+1;
  w_nails = 1;
  for (; steps < 5200000000000;) {
    DEBUG(w_print());
    steps++;

    /* w <- ceil(3/2 * w) */
    WORD t = w[0];
    WORD c = t&1;
    WORD *p;
    for (p=w; p<w_above; p++) {
      WORD next_t = p[1];
      WORD lo = t>>1;
      WORD hi = (next_t & 1) << (WORD_BITLEN-2);
      t += c;
      t += lo;
      t += hi;
      WORD new_t = t & MASK;
      c = t >> (WORD_BITLEN-1);
      t = next_t;
      p[0] = new_t;
    }
    if (UNLIKELY( c )) {
      *w_above++ = 1;
      DEBUG(fprintf(stderr, "\nextend w_above = %u\n", w_above-w));
      if (UNLIKELY( w_above >= w + W_LIMIT - 2 )) {
        fprintf(stderr, "\nstop at w_len=%u\n", w_above-w);
        break;
      }
    }

    WORD wparity = w[0] & 1;
    if (wparity == prev_wparity) {
      runlen++;
    } else {
      if (WANT_ALL_RUNLENS) {
        printf ("%5lu %2u\n", run_number, runlen);
      }
      if (UNLIKELY( runlen > max_runlen )) {
        if (WANT_RECORD_RUNLENS) {
          printf ("%d,", runlen);
        }
        if (WANT_RECORD_RUNLEN_INDICES) {
          printf ("%lu,", run_number);
        }
        max_runlen = runlen;
      }
      prev_wparity = wparity;
      runlen = 1;
      run_number++;
    }
  }
  struct tms t2;
  times(&t2);
  fprintf(stderr, "\nsteps %lu elapsed time %.1f seconds\n",
          steps,
          (t2.tms_utime - t.tms_utime)/(double)sysconf(_SC_CLK_TCK));

  DEBUG(w_print());
  return (0);
}

void
propagate_nails (void)
{
  WORD c = 0;
  WORD *p;
  for (p=w; p<w_above; p++) {
    WORD t = *p;
    WORD new_c = t >> (WORD_BITLEN-1);
    t &= MASK;
    t += c;
    new_c += t >> (WORD_BITLEN-1);
    t &= MASK;
    *p = t;
    c = new_c;
  }
  if (c) {
    *w_above++ = c;
  }
}

/* WRONG carry can propagate ... */
int
by_nails_lazy (void)
{
  struct tms t; times(&t);
  unsigned runlen = 0;
  unsigned long run_number = 1;
  unsigned max_runlen = 0;
  WORD prev_wparity = 1;

  /*
    1011000110000101100111000110111[1],1000101100000000000000000000000,
    1011000110000101100111000110111    01001011

    0010101001001110011010101000110,0101110010000000000000000000000,
    0010101001001110011010101000110 001111001


  */
  memset(w,'\0',sizeof(w));
  w[0] = 1;
  w_nails = 1;
  w_above = w+1;
  for (; steps < 5200000000000;) {
    steps++;
    DEBUG(printf("steps %lu ",steps); w_print());
    /* propagate_nails(&w_above); */
    /* DEBUG(printf("  propag "); w_print()); */

    /* w <- ceil(3/2 * w) */
    WORD t = w[0];
    WORD c = t&1;
    WORD *p;
    for (p=w; p<w_above; p++) {
      WORD next_t = p[1];
      WORD lo = t>>1;
      WORD hi_nail = 0; /* (t>>1) & (1 << (WORD_BITLEN-2)); */
      WORD hi_next = (next_t & 1) << (WORD_BITLEN-2);
      WORD new_c = t >> (WORD_BITLEN-1);
      t &= MASK;
      WORD orig_t = t;
      t += c;
      t += lo;
      t += hi_nail;
      t += hi_next;
      p[0] = t;
      if (! (t >= orig_t)) abort();
      t = next_t;
      c = new_c;
    }
    if (UNLIKELY( c )) {
      *w_above++ = 1;
      DEBUG(fprintf(stderr, "\nextend w_above = %u\n", w_above-w));
      if (UNLIKELY( w_above >= w + W_LIMIT - 2 )) {
        fprintf(stderr, "\nstop at w_above = %u\n", w_above-w);
        break;
      }
    }

    WORD wparity = w[0] & 1;
    if (wparity == prev_wparity) {
      runlen++;
    } else {
      if (WANT_ALL_RUNLENS) {
        printf ("%5lu %2u\n", run_number, runlen);
      }
      if (UNLIKELY( runlen > max_runlen )) {
        if (WANT_RECORD_RUNLENS) {
          printf ("%d,", runlen);
        }
        if (WANT_RECORD_RUNLEN_INDICES) {
          printf ("%lu,", run_number);
        }
        max_runlen = runlen;
      }
      prev_wparity = wparity;
      runlen = 1;
      run_number++;
    }
  }
  struct tms t2;
  times(&t2);
  fprintf(stderr, "\nsteps %lu elapsed time %.1f seconds\n",
          steps, (t2.tms_utime - t.tms_utime)/(double)sysconf(_SC_CLK_TCK));

  printf("final "); DEBUG(w_print());
  return (0);
}

void
z_print (const mpz_ptr w)
{
  mpz_t z;
  mpz_init_set(z,w);
  while (mpz_sgn(z) != 0) {
    printf("%lu", mpz_get_ui(z) & 1);
    mpz_fdiv_q_2exp (z, z,1);
  }
  printf("\n");
}

int
by_mpz (void)
{
  mpz_t comp;
  mpz_init(comp);
  size_t w_len = w_above - w;
  mpz_import (comp,
              w_len,
              -1,  /* words endian */
              sizeof(WORD),
              -1,  /* bytes endian */
              w_nails,
              w);
  struct tms t;
  times(&t);
  mpz_t w;
  int runlen = 0;
  int max_runlen = 0;
  unsigned long prev_wparity;
  unsigned long zsteps = 0;

  mpz_init_set_ui(w,1);
  prev_wparity = mpz_get_ui(w) & 1;
  for ( ; zsteps < steps; ) {
    unsigned long wparity = mpz_get_ui(w) & 1;
    DEBUG(z_print(w));
    if (wparity == prev_wparity) {
      runlen++;
    } else {
      if (runlen > max_runlen) {
        printf ("%d,", runlen);
        max_runlen = runlen;
      }
      prev_wparity = wparity;
      runlen = 1;
    }

    /* ceil(3/2 *w) */
    mpz_mul_ui (w, w,3);
    mpz_cdiv_q_2exp (w, w,1);
    /* if (mpz_size(w) >= W_LIMIT) { */
    /*   fprintf(stderr, "\nstop at size = %lu\n", (unsigned long) mpz_size(w)); */
    /*   break; */
    /* } */
    zsteps++;
  }
  struct tms t2;
  times(&t2);
  fprintf(stderr, "\nzsteps %lu elapsed time %.1f seconds\n",
          zsteps, (t2.tms_utime - t.tms_utime)/(double)sysconf(_SC_CLK_TCK));
  DEBUG(z_print(w));

  if (mpz_cmp(w,comp)!=0) {
    fprintf(stderr, "different\n");
    gmp_printf("w    %Zx\n", w);
    gmp_printf("comp %Zx  w_len %d\n", comp, w_len);
  }

  return (0);
}

int
main (void)
{
  setbuf(stdout,0);
  by_full();
  /* by_nails(); */
  by_mpz();
  return (0);
}
