#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018, 2019, 2020, 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use warnings;
use 5.006;
use Carp 'croak';
use FLAT;
use Math::BaseCnv 'cnv';
use Math::PlanePath::DragonCurve;
use List::Util 'max','sum';
use Test::More;

use FindBin;
use lib $FindBin::Bin;
use MyFLAT ':all';
use MyDragon ':all';

use MyTestHelpers;
BEGIN { MyTestHelpers::nowarnings(); }

plan tests => 567;

# uncomment this to run the ### lines
# use Smart::Comments;


# check that the expand(0) and expand(1) of $conf have all of $conf as subsets
sub ok_conf_expand_within_self {
  my ($conf) = @_;
  foreach my $bit (0,1) {
    ok($conf->is_subset( $conf->expand($bit) ),
       "$conf->{'name'}, bit=$bit");
  }
}


#------------------------------------------------------------------------------
# FLAT_any_binary()

is (FLAT_any_binary()->num_states, 1);
is (FLAT_zeros()->num_states, 1);
is (FLAT_ones()->num_states, 1);


#------------------------------------------------------------------------------
# Dpred

{
  my $path = Math::PlanePath::DragonCurve->new;
  my $bad = 0;
  my $n_hi = 256;
 N: foreach my $n (0 .. $n_hi) {
    my $str = sprintf '%b', $n;
    my $str_n_minus_1 = sprintf '%b', $n-1;
    foreach my $zeros (0 .. 3) {
      my $str = ('0' x $zeros) . $str;
      my ($x,$y) = $path->n_to_xy($n);
      my @n_list = $path->xy_to_n_list($x,$y);
      my $turn = $n && turn($n);
      # {
      #   my $f = SpredInf_flat();
      #   my $f_pred = $f->contains($str) ? 1 : 0;
      #   my $p_pred = @n_list == 1 ? 1 : 0;
      #   if ($p_pred != $f_pred) {
      #     print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
      #     last N if $bad++ > 10;
      #   }
      # }
      {
        my $f = DpredFirstL_inf_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = @n_list == 2 && $n==$n_list[0] && $turn==1 ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred  n_list=",join(',',@n_list),"\n";
          last N if $bad++ > 10;
        }
      }
      {
        my $f = DpredFirstR_inf_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = @n_list == 2 && $n==$n_list[0] && $turn==-1 ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred  n_list=",join(',',@n_list),"\n";
          last N if $bad++ > 10;
        }
      }
      {
        my $f = DpredFirst_inf_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = @n_list == 2 && $n==$n_list[0] ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred  n_list=",join(',',@n_list),"\n";
          last N if $bad++ > 10;
        }
      }
      {
        my $f = DpredSecond_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = @n_list == 2 && $n==$n_list[1] ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred  n_list=",join(',',@n_list),"\n";
          last N if $bad++ > 10;
        }
      }
      {
        my $f = DpredSecond_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = DpredSecond($n);
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want DpredSecond=$p_pred  n_list=",join(',',@n_list),"\n";
          last N if $bad++ > 10;
        }
      }
      {
        my $f = Dpred_inf_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = @n_list == 2 ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
          last N if $bad++ > 10;
        }
      }
    }
  }
  is($bad, 0, "Dpred etc vs path");
}


#------------------------------------------------------------------------------
# LpredInf_flat()

isa_ok(LpredInf_flat(),    'FLAT::DFA');
isa_ok(NonLpredInf_flat(), 'FLAT::DFA');

# complements
ok(LpredInf_flat()->intersect(NonLpredInf_flat())->is_empty,
   'LpredInf_flat intersect NonLpredInf_flat is_empty');
ok(LpredInf_flat()->union(NonLpredInf_flat())->as_dfa
   ->equals(FLAT_any_binary()),
   'LpredInf_flat union NonLpredInf_flat = all');


#------------------------------------------------------------------------------
# Lpred_k_flat()

isa_ok(Lpred_k_flat(), 'FLAT::DFA');
isa_ok(NonLpred_k_flat(), 'FLAT::DFA');
isa_ok(Lpred_k_LtoH_flat(), 'FLAT::DFA');
isa_ok(NonLpred_k_LtoH_flat(), 'FLAT::DFA');

{
  my $max_k = 10;
  is_deeply([ FLAT_count_contains(Lpred_k_LtoH_flat(),$max_k) ],
            [ map {L($_)} 0 .. $max_k ]);

  is_deeply([ FLAT_count_contains(Lpred_k_flat(),$max_k) ],
            [ map {L($_)} 0 .. $max_k ]);
}

# complements
ok(Lpred_k_flat()->equals( NonLpred_k_flat()->complement ),
   'Lpred_k_flat complement NonLpred_k_flat');
ok(Lpred_k_flat()->intersect(NonLpred_k_flat())->is_empty,
   'Lpred_k_flat intersect NonLpred_k_flat is_empty');
ok(Lpred_k_flat()->union(NonLpred_k_flat())->as_dfa->equals(FLAT_any_binary()),
   'Lpred_k_flat union NonLpred_k_flat = all');

ok(NonLpred_k_LtoH_flat()->MyFLAT::reverse->equals(NonLpred_k_flat()));
ok(Lpred_k_LtoH_flat()   ->MyFLAT::reverse->equals(Lpred_k_flat()));

{
  my $path = Math::PlanePath::DragonCurve->new;
  my $bad = 0;
 K: foreach my $k (0 .. 10) {
    my $n_hi = 1 << $k;
    foreach my $n (0 .. 2**$k-1) {
      my $str = sprintf '%0*b', $k, $n;
      my $p_pred = path_n_is_left_boundary($path,$n,$n_hi) ?1:0;

      {
        my $f_pred = Lpred_k_flat()->contains($str) ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "Lpred_k_flat() at N=$n $str got $f_pred want path=$p_pred\n";
          last K if $bad++ > 10;
        }
      }
      {
        my $LtoH_str = reverse $str;
        my $f_LtoH_pred = Lpred_k_LtoH_flat()->contains($LtoH_str) ? 1 : 0;
        if ($p_pred != $f_LtoH_pred) {
          print "Lpred_k_LtoH_flat() at N=$n $LtoH_str got $f_LtoH_pred want path=$p_pred\n";
          last K if $bad++ > 10;
        }
      }
    }
  }
  is($bad, 0, 'Lpred_flat() vs path');
}
# for(n=0,31,if(Lpred(n),print(to_binary(n))))

# ok(Lpred_re_flat()   ->equals(Lpred_k_flat()), 'Lpred_re_flat() vs Lpred_k_flat');
# ok(NonLpred_re_flat()->equals(NonLpred_k_flat()));


#------------------------------------------------------------------------------
# BitAboveLowestOne_LtoH_flat()

{
  my $bad = 0;
  my $n_hi = 256;
 DIR: foreach my $bit (0,1) {
    my $f = BitAboveLowestOne_LtoH_flat($bit);
    foreach my $n (0 .. $n_hi) {
      my $str = sprintf '%b', $n;
      foreach my $zeros (0 .. 3) {
        my $str = reverse(('0' x $zeros) . $str);

        my $by_flat = ($f->contains($str) ? 1 : 0);
        my $by_func = (BitAboveLowestOne_allowing_zero($n) == $bit ? 1 : 0);
        if ($by_flat != $by_func) {
          print "$f->{name} at N=$n $str got flat=$by_flat want func=$by_func\n";
          last DIR if $bad++ > 10;
        }
      }
    }
  }
  is($bad, 0, "BitAboveLowestOne_LtoH_flat() vs BitAboveLowestOne()");
}


#------------------------------------------------------------------------------
# BitAboveLowestOne_HtoL_flat()

{
  my $bad = 0;
  my $n_hi = 256;
 DIR: foreach my $bit (0,1) {
    my $f = BitAboveLowestOne_HtoL_flat($bit);
    foreach my $n (0 .. $n_hi) {
      my $str = sprintf '%b', $n;
      foreach my $zeros (0 .. 3) {
        my $str = ('0' x $zeros) . $str;

        my $by_flat = ($f->contains($str) ? 1 : 0);
        my $by_func = (BitAboveLowestOne_allowing_zero($n) == $bit ? 1 : 0);
        if ($by_flat != $by_func) {
          print "$f->{name} at N=$n $str got flat=$by_flat want func=$by_func\n";
          last DIR if $bad++ > 10;
        }
      }
    }
  }
  is($bad, 0, "BitAboveLowestOne_HtoL_flat() vs BitAboveLowestOne()");
}


#------------------------------------------------------------------------------
# dir4_flat()

{
  my $bad = 0;
  my $n_hi = 256;
 DIR: foreach my $want_dir4 (0 .. 3) {
    my $f = dir4_flat($want_dir4);
    foreach my $n (0 .. $n_hi) {
      my $str = sprintf '%b', $n;
      foreach my $zeros (0 .. 3) {
        my $str = ('0' x $zeros) . $str;

        my $by_f   = $f->contains($str) ? 1 : 0;
        my $by_dir = (dir($n) % 4) == $want_dir4 ? 1 : 0;
        if ($by_f != $by_dir) {
          print "$f->{name} at N=$n $str got flat=$by_f want dir=$by_dir\n";
          last DIR if $bad++ > 10;
        }
      }
    }
  }
  is($bad, 0, "dir4_flat() vs dir()");
}


#------------------------------------------------------------------------------
# dTurnRight_flat()

{
  my $f1even_n     = FLAT::Regex->new('(0|1)* 110')->as_dfa;
  my $f1even_nseen = FLAT::Regex->new('(0|1)* 100')->as_dfa;
  ok ($f1even_n->intersect(TurnLpred_flat())->is_empty,
      'dTurnRight f1even_n no match lefts');
  my $got_f1even_nseen
    = $f1even_n->MyFLAT::transmute (func => \&transmute_lowest_1_to_0,
                                    direction => 'reverse');
  ok ($got_f1even_nseen->equals($f1even_nseen));

  my $f2even_n     = FLAT::Regex->new('(0|1)* 1100')->as_dfa;
  my $f2even_nseen = FLAT::Regex->new('(0|1)* 1000')->as_dfa
    ->MyFLAT::set_name('dTurnLeft f2even_nseen');
  ok ($f2even_n->intersect(TurnLpred_flat())->is_empty,
      'dTurnRight f2even_n no match lefts');
  my $got_f2even_nseen
    = $f2even_n->MyFLAT::transmute (func => \&transmute_lowest_1_to_0,
                                    direction => 'reverse');
  ok ($got_f2even_nseen->equals($f2even_nseen));

  my $f3even_n     = FLAT::Regex->new('(0|1)* 11 0* 000')->as_dfa;
  my $f3even_nseen = FLAT::Regex->new('(0|1)* 10 0* 000')->as_dfa
    ->MyFLAT::set_name('dTurnLeft f3even_nseen');
  ok ($f3even_n->intersect(TurnLpred_flat())->is_empty,
      'dTurnRight f3even_n no match lefts');
  my $got_f3even_nseen
    = $f3even_n->MyFLAT::transmute (func => \&transmute_lowest_1_to_0,
                                    direction => 'reverse');
  ok ($got_f3even_nseen->equals($f3even_nseen));

  my $f1odd_n     = FLAT::Regex->new('(0|1)* 10 1* 11')->as_dfa;
  ok ($f1odd_n->intersect(TurnLpred_flat())->is_empty,
      'dTurnRight f1odd_n no match lefts');
  # no odd d=2, they are all left turns

  my $f3odd_n     = FLAT::Regex->new('((0|1)* 00 | 0 | []) 11')->as_dfa;
  ok ($f3odd_n->contains('11'));
  ok ($f3odd_n->intersect(TurnLpred_flat())->is_empty,
      'dTurnRight f3odd_n no match lefts');

  my $f4odd_n     = FLAT::Regex->new('((0|1)* 00 | 0 | []) 1* 111')->as_dfa
    ->MyFLAT::set_name('dTurnLeft f4odd_n');
  ok ($f4odd_n->intersect(TurnLpred_flat())->is_empty,
      'dTurnRight f4odd_n no match lefts');

  my $got_f1 = $f1even_nseen->union($f1odd_n)->as_dfa
    ->MyFLAT::set_name('dTurnRight got_f1');

  # dTurnRight_flat(3) accepts "00...00" including empty "" which don't
  # occur from the procedure as such
  my $got_f3 = $f3even_nseen
    ->union($f3odd_n)
    ->union(FLAT::Regex->new('0*')->as_dfa)
    ->as_dfa
    ->MyFLAT::set_name('dTurnRight got_f3');

  ok (dTurnRight_flat(1)->equals($got_f1));
  ok (dTurnRight_flat(2)->equals( $f2even_nseen ));
  ok (dTurnRight_flat(3)->equals($got_f3));
  ok (dTurnRight_flat(4)->equals( $f4odd_n ));

  # FLAT_check_is_equal(dTurnRight_flat(3), $got_f3);

  my @nseens = ($f1even_nseen, $f2even_nseen, $f3even_nseen,
                $f1odd_n, $f3odd_n, $f4odd_n);
  foreach my $i (0 .. $#nseens) {
    foreach my $j ($i+1 .. $#nseens) {
      ok ($nseens[$i]->intersect($nseens[$j])->is_empty,
          'dTurnLeft nseens disjoint');
    }
  }
}

foreach my $d (1 .. 4) {
  my $f = dNextR_flat($d);
  $f = $f->intersect(TurnRpred_flat());

  my $f_even = $f->intersect(FLAT::Regex->new('(0|1)* 0 | []')->as_dfa);
  $f_even = $f_even->MyFLAT::transmute (func => \&transmute_lowest_1_to_0,
                                        direction => 'reverse')
    ->MyFLAT::optional_leading_0s;
  my $f_odd = $f->intersect(FLAT::Regex->new('(0|1)* 1')->as_dfa);
  $f = $f_even->union($f_odd)->as_dfa;
  if ($d == 3) {
    # dTurnRight state machine state 3a accepts "00..00" including empty
    # string, which does not occur as transmuted n as such
    $f = $f->union(FLAT::Regex->new('0*')->as_dfa)->as_dfa;
  }

  ok(dTurnRight_flat($d)->equals($f),
     "dTurnRight_flat() vs transmuted n, d=$d");

  # dTurnRight_flat($d)->difference($f)->MyFLAT::minimize->MyFLAT::view;
  # dTurnRight_flat($d)->MyFLAT::minimize->MyFLAT::view;
  # $f->MyFLAT::minimize->MyFLAT::view;
  # FLAT_check_is_equal(dTurnRight_flat($d), $f);
}


#------------------------------------------------------------------------------
# dTurnLeft_flat()

{
  my $f1even_n     = FLAT::Regex->new('((0|1)* 0 | []) 1 0* 00')->as_dfa;
  my $f1even_nseen = FLAT::Regex->new('((0|1)* 0 | []) 1 1* 11')->as_dfa
    ->MyFLAT::set_name('dTurnLeft f2even_nseen');
  my $got_f1even_nseen
    = $f1even_n->MyFLAT::transmute (func => \&transmute_initial_0s_to_1s,
                                    direction => 'reverse');
  ok ($got_f1even_nseen->equals($f1even_nseen));

  my $f2even_n     = FLAT::Regex->new('((0|1)* 00 | 0 | []) 10')->as_dfa;
  my $f2even_nseen = FLAT::Regex->new('((0|1)* 00 | 0 | []) 11')->as_dfa;
  my $got_f2even_nseen
    = $f2even_n->MyFLAT::transmute (func => \&transmute_initial_0s_to_1s,
                                    direction => 'reverse');
  ok ($got_f2even_nseen->equals($f2even_nseen));

  my $f3even_n     = FLAT::Regex->new('(0|1)* 1010')->as_dfa;
  my $f3even_nseen = FLAT::Regex->new('(0|1)* 1011')->as_dfa;
  my $got_f3even_nseen
    = $f3even_n->MyFLAT::transmute (func => \&transmute_initial_0s_to_1s,
                                    direction => 'reverse');
  ok ($got_f3even_nseen->equals($f3even_nseen));


  my $f1odd_n     = FLAT::Regex->new('((0|1)* 00 | 0 | []) 1')->as_dfa;
  my $f1odd_nseen = FLAT::Regex->new('((0|1)* 000 | 00 | 0 | [])')->as_dfa;
  my $got_f1odd_nseen = $f1odd_n->MyFLAT::digits_increment (add => -1)
    ->MyFLAT::optional_leading_0s;
  ok ($got_f1odd_nseen->equals($f1odd_nseen));

  # no odd d=2, they are all right turns
  my $f2odd_n     = FLAT::Regex->new('(0|1)* 10 1* 11')->as_dfa;
  ok ($f2odd_n->intersect(TurnLpred_flat())->is_empty);

  my $f3odd_n     = FLAT::Regex->new('((0|1)* 00 | 0 | []) 1* 101')->as_dfa;
  my $f3odd_nseen = FLAT::Regex->new('((0|1)* 00 | 0 | []) 1* 100')->as_dfa;
  my $got_f3odd_nseen = $f3odd_n->MyFLAT::digits_increment (add => -1)
    ->MyFLAT::optional_leading_0s;
  ok ($got_f3odd_nseen->equals($f3odd_nseen));

  my $f4odd_n     = FLAT::Regex->new('(0|1)* 10 1* 101')->as_dfa;
  my $f4odd_nseen = FLAT::Regex->new('(0|1)* 10 1* 100')->as_dfa
    ->MyFLAT::set_name('f4odd_nseen');
  my $got_f4odd_nseen = $f4odd_n->MyFLAT::digits_increment (add => -1)
    ->MyFLAT::optional_leading_0s;
  ok ($got_f4odd_nseen->equals($f4odd_nseen));

  ok (dTurnLeft_flat(1)->equals($f1even_nseen->union($f1odd_nseen)->as_dfa));
  ok (dTurnLeft_flat(2)->equals( $f2even_nseen ));
  ok (dTurnLeft_flat(3)->equals($f3even_nseen->union($f3odd_nseen)->as_dfa));
  ok (dTurnLeft_flat(4)->equals( $f4odd_nseen ));

  my @nseens = ($f1even_nseen, $f2even_nseen, $f3even_nseen,
                $f1odd_nseen, $f3odd_nseen, $f4odd_nseen);
  foreach my $i (0 .. $#nseens) {
    foreach my $j ($i+1 .. $#nseens) {
      ok ($nseens[$i]->intersect($nseens[$j])->is_empty,
         'dTurnLeft nseens disjoint');
    }
  }

  # dTurnLeft_flat(4)->MyFLAT::minimize->MyFLAT::view;
  # $f4odd_nseen->MyFLAT::minimize->MyFLAT::view;
  # FLAT_check_is_equal(dTurnLeft_flat(4), $f4odd_nseen);
}

foreach my $d (1 .. 4) {
  my $f = dNextL_flat($d);
  $f = $f->intersect(TurnLpred_flat());

  my $f_even = $f->intersect(FLAT::Regex->new('(0|1)* 0 | []')->as_dfa);
  $f_even = $f_even->MyFLAT::transmute (func => \&transmute_initial_0s_to_1s,
                                        direction => 'reverse');
  my $f_odd = $f->intersect(FLAT::Regex->new('(0|1)* 1')->as_dfa);
  $f_odd = $f_odd->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::optional_leading_0s;
  $f = $f_even->union($f_odd)->as_dfa;

  ok(dTurnLeft_flat($d)->equals($f),
     "dTurnLeft_flat() vs transmuted d=$d");
}


#------------------------------------------------------------------------------
# dNextLsub1_flat

foreach my $d (1 .. 4) {
  my $dNextL = dNextL_flat($d);
  my $dNextLsub1 = dNextLsub1_flat($d);
  my $by_incr = $dNextL->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::set_name("dNextL incr -1");
  ok ($dNextLsub1->equals($by_incr));
  # MyFLAT::FLAT_check_is_equal($dNextLsub1,$by_incr);
}


#------------------------------------------------------------------------------
# dNextL, dNextR


# high 0s are optional in dNextL_flat and dNextR_flat, so ensure optional in
# the Regex
#
ok(dNextL_flat(1)->equals(FLAT::Regex->new('((0|1)* 00 | 0 | []) 1*')
                          ->as_dfa));
ok(dNextL_flat(2)->equals(FLAT::Regex->new('((0|1)* 00 | 0 | []) 1* 10
                                          |  (0|1)* 10 1* 11')->as_dfa));
ok(dNextL_flat(3)->equals(FLAT::Regex->new('( (0|1)* 00 | 0 | []) 1* 101
                                            | (0|1)* 10 1* 10     ')->as_dfa));
ok(dNextL_flat(4)->equals(FLAT::Regex->new('(0|1)* 10 1* 101')->as_dfa));


ok(dNextR_flat(1)->equals(FLAT::Regex->new('(0|1)* 10 1*')->as_dfa));
ok(dNextR_flat(2)->equals(FLAT::Regex->new('((0|1)* 00 | 0 | []) 1
                                          |  (0|1)* 100')->as_dfa));
ok(dNextR_flat(3)->equals(FLAT::Regex->new('( (0|1)* 000 | 00 | 0 | [])
                                          | ( (0|1)* 00 | 0 | []) 11')
                          ->as_dfa));
ok(dNextR_flat(4)->equals(FLAT::Regex->new('( (0|1)* 00 | 0 | []) 1* 111')
                          ->as_dfa));


foreach my $func (\&dNextL_flat,
                  \&dNextR_flat
                 ) {
  foreach my $i (1 .. 4) {
    foreach my $j ($i+1 .. 4) {
      ok ($func->($i)->intersect($func->($j))->is_empty);
    }
  }

  my $u = $func->(1)
    ->union($func->(2))
    ->union($func->(3))
    ->union($func->(4))
    ->as_dfa
    ->MyFLAT::set_name("union");
  ok ($u->equals(FLAT_any_binary()), "union is all");
  # MyFLAT::FLAT_check_is_equal($u,FLAT_any_binary());
}
{
  my $f1 = TurnLpred_flat()->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::optional_leading_0s->as_dfa
    ->MyFLAT::set_name('dNextL=1 by incr');
  ok (dNextL_flat(1)->equals($f1),
      'dNextL_flat(1) vs next TurnLpred_flat()');
  # MyFLAT::FLAT_check_is_equal($f1,dNextL_flat(1));

  my $f_next_R = TurnRpred_flat()->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::set_name('TurnRpredNext');

  my $f2 = $f1->MyFLAT::digits_increment(add => -1)
    ->intersect($f_next_R)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::set_name('dNextL=2 by incr');
  ok (dNextL_flat(2)->equals($f2),
      'dNextL_flat(2) vs next TurnLpred_flat()');

  my $f3 = $f2->MyFLAT::digits_increment(add => -1)
    ->intersect($f_next_R)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::set_name('dNextL=3 by incr');
  ok (dNextL_flat(3)->equals($f3),
      'dNextL_flat(3) vs next TurnLpred_flat()');

  my $f4 = $f3->MyFLAT::digits_increment(add => -1)
    ->intersect($f_next_R)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::set_name('dNextL=4 by incr');
  ok (dNextL_flat(4)->equals($f4),
      'dNextL_flat(4) vs next TurnLpred_flat()');
}

foreach my $d (1 .. 4) {
  my $rev = dNextL_flat($d)->MyFLAT::reverse;
  ok (dNextL_LtoH_flat($d)->equals($rev),
      "dNextL_LtoH_flat($d) vs reverse dNextL_flat($d)");
}

{
  my $bad = 0;
  my @f = map {dNextL_flat($_)} 1..4;
  foreach my $n (0 .. 500) {
    my $by_search = dNextL_by_search($n);
    my $by_func   = dNextL($n);
    unless ($by_func == $by_search) {
      diag "n=$n dNextL()=$by_func search=$by_search";
      last if $bad++ > 10;
    }

    my $n2 = cnv($n,10,2);
    my @by_flat = map {$_->contains($n2) ? 1 : 0} @f;
    unless (sum(@by_flat) == 1) {
      diag "n=$n dNextL_flat() by_flat ",join(',',@by_flat);
      last if $bad++ > 10;
    }
    my $by_flat = 1 + sum(grep {$by_flat[$_]} 0 .. 3);
    unless ($by_flat == $by_search) {
      diag "n=$n flats ",join(',',@by_flat)," search=$by_search";
      last if $bad++ > 10;
    }
  }
  is($bad, 0);
}
sub dNextL_by_search {
  my ($n) = @_;
  foreach my $d (1 .. 4) {
    if (TurnLpred($n+$d)) { return $d; }
  }
  die "oops";
}

#------
# ok (dNextR_flat(1)->contains('1'));
# ok (dNextR_flat(3)->contains('2'));
#
# {
#   my $bad = 0;
#   my @f = map {dNextR_flat($_)} 1..4;
#   foreach my $n (0 .. 500) {
#     my $by_search = dNextR_by_search($n);
#     my $by_func   = dNextR($n);
#     my $n2 = cnv($n,10,2);
#     my @by_flat = map {$_->contains($n2)} @f;
#     unless (sum(@by_flat) == 1) {
#       last if $bad++ > 10;
#     }
#     unless ($by_func == $by_search) {
#       diag "n=$n dNextR()=$by_func search=$by_search";
#       last if $bad++ > 10;
#     }
#     unless ($by_flat == $by_search) {
#       diag "n=$n dNextR_flat()=$by_flat search=$by_search";
#       last if $bad++ > 10;
#     }
#   }
#   is($bad, 0);
# }
# sub dNextR_by_search {
#   my ($n) = @_;
#   foreach my $d (1 .. 3) {
#     if (TurnRpred($n+$d)) { return $d; }
#   }
#   die "oops";
# }


#------------------------------------------------------------------------------
# TurnLpred_flat()
# TurnRpred_flat()

ok(! TurnLpred_flat() ->contains(''));
ok(! TurnRpred_flat()->contains(''));
ok(TurnLpred_flat() ->contains('1'));
ok(TurnLpred_flat() ->contains('01'));
ok(TurnLpred_flat() ->contains('0001'));
ok(TurnLpred_flat() ->intersect(zero_digits_flat(2))->is_empty,
   "TurnLpred_flat() no match zero");
ok(TurnRpred_flat()->intersect(zero_digits_flat(2))->is_empty,
   "TurnLpred_flat() no match zero");
ok(TurnLpred_flat() ->intersect(TurnRpred_flat())->is_empty,
   "TurnLpred_flat() and TurnRpred_flat() disjoint");

{
  my $bad = 0;
  foreach my $n (1 .. 1000) {
    my $turn = turn($n);
    my $t_left  = $turn==1;
    my $t_right = $turn==-1;
    my $str = sprintf '%b', $n;
    foreach (0 .. 2) {  # leading zeros
      my $f_left  = TurnLpred_flat() ->contains($str) ? 1 : 0;
      my $f_right = TurnRpred_flat()->contains($str) ? 1 : 0;
      if ($t_left != $f_left)  { $bad++; }
      if ($t_right != $t_right) { $bad++; }
      $str = "0$str";
    }
  }
  is($bad, 0, "TurnLpred_flat() and TurnRpred_flat() vs turn()");
}


#------------------------------------------------------------------------------
# MinusUnrotPred_LtoH_flat()

# check accepted N against path
{
  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  my $bad = 0;
 K: foreach my $k (0 .. 12) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      my ($rx,$ry) = xy_rotate_minus90($x,$y);
      my $by_path = $path->xy_to_n($rx,$ry) <= $n_hi ? 1 : 0;
      my $n2 = sprintf '%0*s', $k, cnv($n,10,2);
      my $n2rev = reverse $n2; # scalar context
      my $by_flat = MinusUnrotPred_LtoH_flat()->contains($n2rev) ? 1 : 0;
      if ($by_path != $by_flat) {
        print "k=$k  n=$n2  xy=$x,$y  by_path=$by_path by_flat=$by_flat\n";
        last K if $bad++ > 10;
      }
    }
  }
  is($bad,0, 'MinusUnrotPred_LtoH_flat() vs path');
}

#------------------------------------------------------------------------------
# MinusRotPred_LtoH_flat()

# check accepted N against path
{
  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  my $bad = 0;
 K: foreach my $k (0 .. 12) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      my ($rx,$ry) = xy_rotate_plus90($x,$y);
      my $by_path = $path->xy_to_n($rx,$ry) <= $n_hi ? 1 : 0;
      my $n2 = sprintf '%0*s', $k, cnv($n,10,2);
      my $n2rev = reverse $n2; # scalar context
      my $by_flat = MinusRotPred_LtoH_flat()->contains($n2rev) ? 1 : 0;
      if ($by_path != $by_flat) {
        print "k=$k  n=$n2  xy=$x,$y  by_path=$by_path by_flat=$by_flat\n";
        last K if $bad++ > 10;
      }
    }
  }
  is($bad,0, 'MinusRotPred_LtoH_flat() vs path');
}

# MinusRotIncPred_LtoH_flat()
{
  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  foreach my $dir8 (0 .. 7) {
    my $flat = MinusRotIncPred_dir8_LtoH_flat($dir8);
    my $dx = $dir8_to_dx[$dir8];
    my $dy = $dir8_to_dy[$dir8];
    # $flat->MyFLAT::view;

    my $bad = 0;
  K: foreach my $k (0 .. 8) {
      my ($n_lo, $n_hi) = $path->level_to_n_range($k);
      foreach my $n ($n_lo .. $n_hi-1) {
        my ($x,$y) = $path->n_to_xy($n);
        my ($rx,$ry) = xy_rotate_plus90($x,$y);
        my $by_path = $path->xy_to_n($rx+$dx,$ry+$dy) <= $n_hi ? 1 : 0;
        my $n2 = sprintf '%0*s', $k, cnv($n,10,2);
        my $n2rev = reverse $n2; # scalar context
        my $by_flat = $flat->contains($n2rev) ? 1 : 0;
        if ($by_path != $by_flat) {
          print "k=$k  n=$n2  xy=$x,$y dxdy=$dx,$dy  by_path=$by_path by_flat=$by_flat\n";
          last K if $bad++ > 10;
        }
      }
    }
    is($bad,0, $flat->{'name'});
  }
}

#------------------------------------------------------------------------------
# $NonRpred_re

{
  my $str = MyFLAT::perl_regexp_to_flat_regex($NonRpred_re);
  # diag $str;

  my $fa = FLAT::Regex->new($str)->as_dfa
    ->MyFLAT::set_name('Perl NonRpred_re');
  ok ($fa->equals(NonRpred_flat()));

  # $fa = $fa->MyFLAT::reverse->MyFLAT::minimize(verbose=>1);
  #  $fa->MyFLAT::view;
  # MyFLAT::FLAT_check_is_equal($fa, NonRpred_flat());
}

{
  ok ('11010100' =~ $NonRpred_re);

  my $from_flat_re
    = MyFLAT::flat_regex_to_perl_regexp($NonRpred_flatregex_str);
  # diag $from_flat_re;

  my $bad = 0;
  foreach my $k (0 .. 16) {
    foreach my $n (0 .. 2**$k-1) {
      my $str = ($k==0 ? '' : sprintf '%0*b', $k, $n);
      my $by_func = (Rpred($n) ? 0 : 1);
      {
        my $by_fa = (NonRpred_flat()->contains($str) ? 1 : 0);
        unless ($by_func == $by_fa) {
          print "n=$n str $str  func $by_func NonRpred_flat $by_fa";
          die if $bad++ > 10;
        }
      }
      {
        my $by_re = ($str =~ $NonRpred_re ? 1 : 0);
        unless ($by_func == $by_re) {
          print "n=$n str $str  func $by_func re $by_re\n";
          die if $bad++ > 10;
        }
      }
      {
        my $by_re = ($str =~ $from_flat_re ? 1 : 0);
        unless ($by_func == $by_re) {
          print "n=$n str $str  func $by_func flat re $by_re\n";
          die if $bad++ > 10;
        }
      }
      {
        $str = reverse $str;
        my $by_re = ($str =~ $MyDragon::NonRpred_LtoH_re ? 1 : 0);
        unless ($by_func == $by_re) {
          print "n=$n reverse str $str  func $by_func LtoH re $by_re\n";
          die if $bad++ > 10;
        }
      }
    }
  }
  is($bad, 0);
}


#------------------------------------------------------------------------------
# Rpred_flat()

# ok (FLAT::Regex->new('(001)*')->as_dfa->is_subset_of(MyDragon::Rpred_flat()),
#    '7th (001)* is Rpred_flat');
# ok (FLAT::Regex->new('(010)*')->as_dfa->is_subset_of(MyDragon::Rpred_flat()),
#    '7th (010)* is Rpred_flat');
# ok (FLAT::Regex->new('(100)*')->as_dfa->is_subset_of(MyDragon::Rpred_flat()),
#    '7th (100)* is Rpred_flat');

my $NonRpred_eventually_flat   =MyFLAT::eventually_accepting(NonRpred_flat());
my $NonLpred_k_eventually_flat =MyFLAT::eventually_accepting(NonLpred_k_flat());
my $NonBpred_k_eventually_flat =MyFLAT::eventually_accepting(NonBpred_k_flat());

{
  my $max_k = 10;
  is_deeply([ FLAT_count_contains(Rpred_LtoH_flat(),$max_k) ],
            [ map {R($_)} 0 .. $max_k ]);

  is_deeply([ FLAT_count_contains(Rpred_flat(),$max_k) ],
            [ map {R($_)} 0 .. $max_k ]);

  # eventually NonR = 2^k-(R-RQ3o)
  is_deeply([ FLAT_count_contains($NonRpred_eventually_flat,$max_k) ],
            [ map {2**$_ - (R($_)-RQ3o($_))} 0 .. $max_k ]);
}

{
  my $f = NonRpred_flat();
  my @states = $f->get_states;
  is_deeply([grep {$f->MyFLAT::is_accepting_sink($_)} @states],
            [$states[-1]]);
}

ok(Rpred_flat()->equals( NonRpred_flat()->complement ));
ok(Rpred_LtoH_flat()   ->MyFLAT::reverse->equals(Rpred_flat()));
ok(NonRpred_LtoH_flat()->MyFLAT::reverse->equals(NonRpred_flat()));

{
  my $path = Math::PlanePath::DragonCurve->new;
  my $bad = 0;
 K: foreach my $k (0 .. 10) {
    foreach my $n (0 .. 2**$k-1) {
      my $p_pred = $path->_UNDOCUMENTED__n_segment_is_right_boundary($n) ?1:0;
      my $str = sprintf '%0*b', $k, $n;
      my $f_pred = Rpred_flat()->contains($str) ? 1 : 0;
      if ($p_pred != $f_pred) {
        diag "Rpred_flat() at N=$n $str got $f_pred want path=$p_pred";
        last K if $bad++ > 10;
      }
    }
  }
  is($bad, 0, "Rpred_flat() vs path");
}
# for(n=0,31,if(Rpred(n),print(to_binary(n))))


#------------------------------------------------------------------------------
# Bpred_k_flat()

ok(Bpred_k_flat()->equals( NonBpred_k_flat()->complement ),
   'Bpred_k_flat complement NonBpred_k_flat');

is(ref Bpred_k_flat(), 'FLAT::DFA',
   'Bpred_k_flat is FLAT::DFA');
{
  my @starting = Bpred_k_flat()->get_starting;
  is(scalar(@starting), 1,
   'Bpred_k_flat starting state');
}


#------------------------------------------------------------------------------
# fBpred cases extras vs digit increment

{
  # fRpred 2n + 11/15

  my $extra_plus11_left_flat
    = SpredLeft_k_flat()
    ->intersect(bits_N_even_flat())
    ->MyFLAT::concat(MyFLAT::fraction_digits(11,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_plus11_left');
  ok($extra_plus11_left_flat->equals(fBpred_cases_extra('left',0,11)),
    'fRpred 2n + 11/15');

  my @fRpred_cases_extra_list = fRpred_cases_extra_list();
  ok($extra_plus11_left_flat->equals($fRpred_cases_extra_list[0]));

  {
    my %info = $extra_plus11_left_flat->MyFLAT::get_prefix_accepting_info;
    is ($info{'depth'}, 0,          'no more accepting prefixes');
    is_deeply ($info{'states'}, []);
  }

  # state 163 0111 repeating
  # 10001 0111 01    17 + 7/15 no
  # 1000 1011 101     8 + 11/15     turn(8)==1 left
  # (8 + 11/15)/16  == (17 + 7/15)/32
  # $extra_plus11_left_flat->contains('01000 1101 1101') or die;
  # MyFLAT::view($extra_plus11_left_flat);
  ok($extra_plus11_left_flat->contains('1') );
  ok($extra_plus11_left_flat->contains('10') );
  ok($extra_plus11_left_flat->contains('1000') );
  ok($extra_plus11_left_flat->contains('10001') );
  ok($extra_plus11_left_flat->contains('100010') );
  ok($extra_plus11_left_flat->contains('1000101') );
  ok($extra_plus11_left_flat->contains('10001011') );
  ok($extra_plus11_left_flat->contains('100010111') );
  ok($extra_plus11_left_flat->contains('1000101110') );
  ok($extra_plus11_left_flat->contains('10001011101') );
}
{
  # fRpred 2n - 1/15
  #      = 2n-1 + 14/15 with SpredLeft 2n
  # so SpredLeft evens - 1 and +14/15

  my $extra_minus1_left_flat
    = SpredLeft_k_flat()
    ->intersect(bits_N_even_flat())
    ->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::concat(MyFLAT::fraction_digits(14,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_minus1_left');
  ok($extra_minus1_left_flat->equals(fBpred_cases_extra('left',0,-1)),
    'fRpred 2n - 1/15');

  my @fRpred_cases_extra_list = fRpred_cases_extra_list();
  ok($extra_minus1_left_flat->equals($fRpred_cases_extra_list[1]));
}
{
  # fRpred 2n + 7/15
  # where SpredLeft 2n+1

  my $extra_plus7_left
    = SpredLeft_k_flat()
    ->intersect(bits_N_odd_flat())
    ->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::concat(MyFLAT::fraction_digits(7,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_plus7_left');
  ok($extra_plus7_left->equals(fBpred_cases_extra('left',1,-8)),
    'fRpred 2n + 7/15');

  my @fRpred_cases_extra_list = fRpred_cases_extra_list();
  ok($extra_plus7_left->equals($fRpred_cases_extra_list[2]));
}

{
  # fLpred 2n - 11/15
  # is 2n-1 + 4/15

  my $extra_minus11_right_flat
    = SpredRight_k_flat()
    ->intersect(bits_N_even_flat())
    ->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::concat(MyFLAT::fraction_digits(4,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_minus11_right');
  ok($extra_minus11_right_flat->equals(fBpred_cases_extra('right',0,-11)),
     'fLpred 2n - 11/15');

  my @fLpred_cases_extra_list = fLpred_cases_extra_list();
  ok($extra_minus11_right_flat->equals($fLpred_cases_extra_list[0]));
}
{
  # fLpred 2n + 1/15

  my $extra_plus1_right_flat
    = SpredRight_k_flat()
    ->intersect(bits_N_even_flat())
    ->MyFLAT::concat(MyFLAT::fraction_digits(1,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_plus1_right');
  ok($extra_plus1_right_flat->equals(fBpred_cases_extra('right',0,1)),
    'fLpred 2n + 1/15');

  my @fLpred_cases_extra_list = fLpred_cases_extra_list();
  ok($extra_plus1_right_flat->equals($fLpred_cases_extra_list[1]));
}
{
  # fLpred 2n - 7/15
  # with SpredRight of 2n-1
  # so 2n-1 + 8/15

  my $extra_minus7_right_flat
    = SpredRight_k_flat()
    ->intersect(bits_N_odd_flat())
    ->MyFLAT::concat(MyFLAT::fraction_digits(8,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_plus1_right');
  ok($extra_minus7_right_flat->equals(fBpred_cases_extra('right',1,8)),
     'fLpred 2n - 7/15');

  my @fLpred_cases_extra_list = fLpred_cases_extra_list();
  ok($extra_minus7_right_flat->equals($fLpred_cases_extra_list[2]));
}


#------------------------------------------------------------------------------
# CentrelinePred_flat()

is(CentrelinePred_by_path(4093,12)?1:0, 1);

foreach my $type ('point','segment') {
  my $fa = CentrelinePred_flat($type);
  # CentrelinePred_flat()->MyFLAT::view;
  my $bad = 0;
 K: foreach my $k (0 .. 12) {
    my ($n_lo, $n_hi) = DragonCurve_path()->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my $by_path = CentrelinePred_by_path($n,$k, $type) ? 1 : 0;
      my $n2 = sprintf '%0*s', $k, cnv($n,10,2);
      my $by_flat = $fa->contains($n2) ? 1 : 0;
      if ($by_path != $by_flat) {
        print "$type k=$k  n=$n2  by_path=$by_path by_flat=$by_flat\n";
        last K if $bad++ > 10;
      }
    }
  }
  is($bad, 0, 'CentrelinePred vs path');
}

{
  # CentrelineHex

  my $f = MyDragon::CentrelinePred_flat('segments')
    ->MyFLAT::blocks(4);
  $f = $f->MyFLAT::FLAT_rename(symbols_func => sub {
                                 my ($binary) = @_;
                                 return cnv($binary,2,16);
                               });
  # $f->MyFLAT::view;
  my @reachables = $f->MyFLAT::descendants($f->get_starting);
  @reachables = sort {$a <=> $b} @reachables;
  my @reachable_names = map {$CentrelinePred_state_to_name[$_]} @reachables;
  is_deeply(\@reachable_names, ['h','a2','d2','non'],
            'CentrelineHex reachable states');
  # $f = $f->MyFLAT::minimize(verbose=>1);
}


#------------------------------------------------------------------------------
# fBpred hullstouch state counts, not minimized

fBpred_hullstouch_init();
is($fRpred_hullstouch_flat->num_states, 92);  # as stated in dragon.tex
is($fLpred_hullstouch_flat->num_states, 92);
is($fBpred_hullstouch_flat->num_states, 92);

is (scalar(@{$fBpred_hullstouch_state_enclosed_lists[0]}), 26); # as stated in dragon.tex
is (scalar(@{$fBpred_hullstouch_state_enclosed_lists[1]}), 26); # no eventually accepting
is (scalar(@{$fBpred_hullstouch_state_enclosed_lists[2]}), 2); 

is($fNonRpred_hullstouch_flat->MyFLAT::num_accepting, 26); # stated in dragon.tex
is($fNonLpred_hullstouch_flat->MyFLAT::num_accepting, 26); # incl eventually
is($fNonBpred_hullstouch_flat->MyFLAT::num_accepting, 2);

{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_hullstouch_state_table,
     name => 'fNonRpred_hullstouch',
     accepting_list => $fBpred_hullstouch_state_enclosed_lists[0],
    );
  is ($f->MyFLAT::num_accepting, 26, 'fNonRpred num accepting');
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 0);
  is (scalar(@{$info{'states'}}), 0, 'fRpred no eventually accepting');
}
{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_hullstouch_state_table,
     name => 'fNonLpred_hullstouch',
     accepting_list => $fBpred_hullstouch_state_enclosed_lists[1],
    );
  is ($f->MyFLAT::num_accepting, 26, 'fNonLpred num accepting');
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 0);
  is (scalar(@{$info{'states'}}), 0, 'fLpred no eventually accepting');
}
{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_hullstouch_state_table,
     name => 'fNonBpred_hullstouch',
     accepting_list => $fBpred_hullstouch_state_enclosed_lists[2],
    );
  is ($f->MyFLAT::num_accepting, 2, 'fNonBpred num accepting');
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 0);
  is (scalar(@{$info{'states'}}), 0, 'fBpred no eventually accepting');
}


#------------------------------------------------------------------------------
# fRpred vs Rpred states

fBpred_hullstouch_init();

# Rpred except 3eB are subsets of fRpred, meaning such Rpred sides always
# have some part on the boundary.
#
foreach my $state (0 .. $#Rpred_state_table) {
  my $state_name = $Rpred_state_to_name[$state];
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@Rpred_state_table,
     accepting => $state);
  is($f->is_subset_of($fRpred_hullstouch_flat) ? 1 : 0,
     $state_name eq '3eB' || $state_name eq 'non' ? 0 : 1,
     "R $state_name");
}
# {
#   my $f = MyFLAT::aref_to_FLAT_DFA
#     (\@Rpred_state_table,
#      accepting => $Rpred_name_to_state{'3oB'},
#      name => "R 3oB");
#   ok(! $f->is_subset_of($fNonRpred_hullstouch_flat));
# 
#   # my $int = $f->intersect($fRpred_hullstouch_flat->MyFLAT::minimize(verbose=>1))
#   #   ->MyFLAT::minimize(verbose=>1);
#   # $int->MyFLAT::view;
# }


#------------------------------------------------------------------------------
# fBpred hullstouch, fRpred_flat() minimized equality 

is(fRpred_flat()->num_states, 30, 'fRpred num states'); # in dragon.tex
is(fLpred_flat()->num_states, 32, 'fLpred num states');

ok(fRpred_flat()->equals($fRpred_hullstouch_flat));
ok(fLpred_flat()->equals($fLpred_hullstouch_flat));
ok(fNonRpred_flat()->equals($fNonRpred_hullstouch_flat));
ok(fNonLpred_flat()->equals($fNonLpred_hullstouch_flat));

ok( fNonRpred_flat()->equals($fNonRpred_hullstouch_flat->MyFLAT::minimize) );
ok( fNonLpred_flat()->equals($fNonLpred_hullstouch_flat->MyFLAT::minimize) );


#------------------------------------------------------------------------------
# fBpred cases extras are on the boundary so are subset of fRpred, etc

fBpred_hullstouch_init();
foreach my $extra (fRpred_cases_extra_list()) {
  ok($extra->is_subset_of($fRpred_hullstouch_flat),
     'fRpred_cases_extra_list are fRpred subsets');
}
foreach my $extra (fLpred_cases_extra_list()) {
  ok($extra->is_subset_of($fLpred_hullstouch_flat),
     'fLpred_cases_extra_list are fLpred subsets');
}
foreach my $extra (fRpred_cases_extra_list(), fLpred_cases_extra_list()) {
  ok($extra->is_subset_of($fBpred_hullstouch_flat),
     'fRpred_cases_extra_list and fLpred_cases_extra_list are fBpred subsets');
}

# Extras subtract from NonBpred_k_flat().
# Extra is bigger than NonBpred but has a non-empty intersection so that
# it does subtract.
foreach my $extra (fRpred_cases_extra_list(), fLpred_cases_extra_list()) {
  my $int = NonBpred_k_flat()->intersect($extra);
  ok(! $int->is_empty,
     'fBpred extra non-empty intersection so does add something');
}

# anything with Spred at one end of segment is fBpred,
# including prefixes of Spred as they are not yet proven enclosed
ok (Spred_k_flat()
    ->is_subset_of($fBpred_hullstouch_flat),
    'Spred start is subset of fBpred');
ok (Spred_k_flat()
    ->MyFLAT::prefix
    ->is_subset_of($fBpred_hullstouch_flat),
    'Spred start is subset of fBpred');


#------------------------------------------------------------------------------
# Rsides()

sub Rsides_by_path {
  my ($path, $n) = @_;
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ### Rsides_by_path(): "n=$n $x,$y dir $dx,$dy"
  my $ret = 1;
  foreach (1..3) {
    $x += $dx;
    $y += $dy;
    ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to right side
    ### try: "$x,$y dir $dx,$dy"
    if (path_xyxy_is_traversed_either($path, $x,$y, $x+$dx,$y+$dy)) { $ret++; }
  }
  return $ret;
}

{
  my @flats = (undef, map {Rsides_flat($_)} 1..4);
  my $bad = 0;
  N: foreach my $n (0 .. (1<<8)) {
    my $s = Rsides($n);
    foreach my $t (1..4) {
      my $n2 = cnv($n,10,2);
      unless (($flats[$t]->contains($n2) ? 1 : 0)
              == ($s==$t ? 1 : 0)) {
        print "Rsides_flat() wrong\n";
        last N if $bad++ > 10;
      }
    }
  }
  is($bad,0, "Rsides() vs Rsides_flat()");
}
{
  my $path = Math::PlanePath::DragonCurve->new;
  my $bad = 0;
  foreach my $n (0 .. (1<<10)) {
    unless (Rsides($n) == Rsides_by_path($path,$n)) {
      print "Rsides() wrong\n";
      last if $bad++ > 10;
    }
  }
  is($bad,0, "Rsides() vs Rsides_by_path()");
}


#------------------------------------------------------------------------------
# EpredNine

my $EpredNine = Epred_flat();
for (1 .. 8) {
  $EpredNine = Epred_flat()->intersect($EpredNine->MyFLAT::digits_increment(add => -1))
    ->MyFLAT::minimize(verbose=>1);
  ok(! $EpredNine->is_empty);
}
$EpredNine->MyFLAT::set_name('EpredNine');

{
  # one further for n+9 is 10 terms n to n+8 inclusive
  my $e = Epred_flat()->intersect($EpredNine->MyFLAT::digits_increment(add => -1));
  ok($e->is_empty);
}

{
  # turn sequence / enclosure side sequence is always
  my @turn = (undef,    # side LLLRR LRRL, turn opposite
              TurnRpred_flat(),TurnRpred_flat(),TurnRpred_flat(),,
              TurnLpred_flat(),TurnLpred_flat(),
              TurnRpred_flat(),
              TurnLpred_flat(),TurnLpred_flat(),
              TurnRpred_flat());
  foreach my $i (1 .. 9) {
    my $p = $EpredNine->MyFLAT::digits_increment(add => $i);
    ok($p->is_subset_of($turn[$i]), "EpredNine turn i=$i");
  }
}

my $DpredSecondNine = DpredSecond_flat();
{
  foreach (1 .. 8) {
    $DpredSecondNine = DpredSecond_flat()
      ->intersect($DpredSecondNine->MyFLAT::digits_increment(add=>-1))
      ->MyFLAT::minimize(verbose=>1);
    ok(! $DpredSecondNine->is_empty);
  }
  $DpredSecondNine->MyFLAT::set_name('DpredSecondNine');
  {
    # one further for n+9 is 10 terms n to n+8 inclusive
    my $e = DpredSecond_flat()->intersect($DpredSecondNine->MyFLAT::digits_increment(add=>-1));
    ok($e->is_empty);
  }
  ok($EpredNine->MyFLAT::digits_increment->equals($DpredSecondNine));
  ok($DpredSecondNine->contains('1010111110'));  # first n=702
}

{
  # MyFLAT::FLAT_show_breadth($DpredSecondNine, 12);
  my $max_k = 20;
  my $want = '0,0,0,0,0,0,0,0,0,0,1,3,7,17,40,90,197,424,900,1889,3929';
  is(join(',', MyFLAT::FLAT_count_contains($EpredNine,      $max_k)), $want);
  is(join(',', MyFLAT::FLAT_count_contains($DpredSecondNine,$max_k)), $want);
}

{
  my $n = 701;  # first nine
  ok($EpredNine->contains(sprintf '%b', $n));
  ok($DpredSecondNine->contains(sprintf '%b', $n+1));
  ok (! Epred($n-1));
  foreach my $i (0 .. 8) {
    ok (Epred($n+$i));
  }
  ok (! Epred($n+9));
}


#------------------------------------------------------------------------------
# EpredL_flat()
# EpredR_flat()

ok (DpredSecondR_flat()->equals(EpredL_flat()->MyFLAT::digits_increment));
ok (EpredL_flat()->equals(DpredSecondR_flat()
                          ->MyFLAT::digits_increment(add=>-1)));
ok (EpredL_flat()->MyFLAT::reverse->equals(EpredL_LtoH_flat()));

ok (DpredSecondL_flat()->equals(EpredR_flat()->MyFLAT::digits_increment));
ok (EpredR_flat()->equals(DpredSecondL_flat()
                          ->MyFLAT::digits_increment(add=>-1)));
ok (EpredR_flat()->MyFLAT::reverse->equals(EpredR_LtoH_flat()));

{
  my $bad = 0;
  my $n_hi = 256;
 N: foreach my $n (0 .. $n_hi) {
    my $str = sprintf '%b', $n;
    {
      my $f = EpredL_flat();
      my $f_pred = $f->contains($str) ? 1 : 0;
      my $p_pred = EpredL($n) ? 1 : 0;
      if ($p_pred != $f_pred) {
        print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
        last N if $bad++ > 10;
      }
    }
    {
      my $f = EpredR_flat();
      my $f_pred = $f->contains($str) ? 1 : 0;
      my $p_pred = EpredR($n) ? 1 : 0;
      if ($p_pred != $f_pred) {
        print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
        last N if $bad++ > 10;
      }
    }
    {
      my $f = Epred_flat();
      my $f_pred = $f->contains($str) ? 1 : 0;
      my $p_pred = Epred($n) ? 1 : 0;
      if ($p_pred != $f_pred) {
        print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
        last N if $bad++ > 10;
      }
    }
  }
  is($bad, 0, "Epred flats vs path");
}

#------------------------------------------------------------------------------
# EpredL3_flat()
# EpredR3_flat()

{
  my $f = EpredL_flat();
  for (1 .. 2) {
    $f = EpredL_flat()->intersect($f->MyFLAT::digits_increment(add => -1))
      ->MyFLAT::minimize(verbose=>1);
  }
  $f->MyFLAT::set_name('EpredL3');

  my $t = EpredL_flat()
    ->MyFLAT::concat(FLAT::Regex->new('101')->as_dfa)
    ->MyFLAT::set_name('concat');
  ok($t->equals($f));
}
{
  my $f = EpredR_flat();
  for (1 .. 2) {
    $f = EpredR_flat()->intersect($f->MyFLAT::digits_increment(add => -1))
      ->MyFLAT::minimize(verbose=>1);
  }
  $f->MyFLAT::set_name('EpredR3');

  my $even = EpredR_flat()->intersect(FLAT::Regex->new('(0|1)*0'))
      ->MyFLAT::minimize(verbose=>1);

  my $o = EpredR_flat()->MyFLAT::concat(FLAT::Regex->new('1111'));
  my $e = $even        ->MyFLAT::concat(FLAT::Regex->new('0111'));
  my $t = $e->union($o)
    ->MyFLAT::minimize(verbose=>1)
    ->MyFLAT::set_name('concat');

  ok($t->equals($f));

  # $f = $f->MyFLAT::reverse->MyFLAT::minimize(verbose=>1);
  # $t = $t->MyFLAT::reverse->MyFLAT::minimize(verbose=>1);
  # $f->MyFLAT::view;
  # $t->MyFLAT::view;
  # EpredR_LtoH_flat()->MyFLAT::view;
}



#------------------------------------------------------------------------------
# TAparentDir LtoH

{
  my @TAparentDir_LtoH_state_table = ([0,0 ],
                                      [  2,12 ],   # s1
                                      [  3, 4 ],   # s2
                                      [  3, 5 ],   # s3
                                      [  6, 4 ],   # s4
                                      [  7,11 ],   # s5
                                      [ 12, 7 ],   # s6
                                      [  8, 9 ],   # s7
                                      [  8, 1 ],   # s8
                                      [ 10, 9 ],   # s9
                                      [ 11, 2 ],   # s10
                                      [ 11,11 ],   # 11=yes
                                      [ 12,12 ]);  # 12=non
  my @TAparentDir_LtoH_accepting = (5,7, 8,9,10, 11);
  foreach my $elem ([0,  1, \&TAparentDir0_flat],
                    [1,  8, \&TAparentDir1_flat],
                    [2, 10, \&TAparentDir2_flat],
                    [3,  9, \&TAparentDir3_flat]) {
    my ($dir, $starting, $HtoL_func) = @$elem;
    my $HtoL = &$HtoL_func();

    my $LtoH = MyFLAT::aref_to_FLAT_DFA
      (\@TAparentDir_LtoH_state_table,
       name           => "TAparentDir$dir LtoH",
       starting       => $starting,
       accepting_list => \@TAparentDir_LtoH_accepting);
    if ($dir >= 2) { $LtoH = $LtoH->complement; }

    ok($LtoH->MyFLAT::reverse->equals($HtoL), "TAparentDir LtoH dir=$dir");
  }
}

#------------------------------------------------------------------------------
# MinusNeg hanging squares touch only one diagonal

#     +---+
#     | * |
#     +---+

foreach my $type ('MinusNeg','MinusRot','MinusDouble') {
  my $pred = Pred($type);
  my $f_no_nsew =
    $pred
    ->intersect(NeighbourPred($type,0)->complement)
    ->MyFLAT::minimize(verbose=>1)
    ->intersect(NeighbourPred($type,2)->complement)
    ->MyFLAT::minimize(verbose=>1)
    ->intersect(NeighbourPred($type,4)->complement)
    ->MyFLAT::minimize(verbose=>1)
    ->intersect(NeighbourPred($type,6)->complement)
    ->MyFLAT::minimize(verbose=>1);

  for (my $i = 1; $i <= 7; $i+=2) {
    for (my $j = $i+2; $j <= 7; $j+=2) {
      my $f = $f_no_nsew
        ->intersect(NeighbourPred($type,$i))
        ->intersect(NeighbourPred($type,$j));
      ok($f->is_empty, "$type diagonals $i $j");
    }
  }
}


#------------------------------------------------------------------------------
# MinusDoublePred_LtoH_flat()

# MinusDoubleIncPred_LtoH_flat()
{
  # to_binary(MinusUnpoint(3-2*I))     == 111001
  # to_binary(MinusUnpoint(3-2*I + I))     == 111010110
  # to_binary(MinusUnpoint(2*(3-2*I + I))) == 101000
  # MinusUnpoint(1+I) == 14
  # MinusUnpoint(2*(1+I)) == 8

  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  foreach my $dir8 (0 .. 7) {
    my $flat = MinusDoubleIncPred_dir8_LtoH_flat($dir8);
    my $dx = $dir8_to_dx[$dir8];
    my $dy = $dir8_to_dy[$dir8];
    # $flat->MyFLAT::view;

    my $bad = 0;
  K: foreach my $k (0 .. 12) {
      my ($n_lo, $n_hi) = $path->level_to_n_range($k);
      foreach my $n ($n_lo .. $n_hi-1) {
        my ($x,$y) = $path->n_to_xy($n);
        my $by_path = ($path->xy_to_n(2*($x+$dx),2*($y+$dy)) <= $n_hi
                       ? 1 : 0);
        my $n2 = sprintf '%0*s', $k, cnv($n,10,2);
        my $n2rev = reverse $n2; # scalar context
        my $by_flat = $flat->contains($n2rev) ? 1 : 0;
        if ($by_path != $by_flat) {
          print "k=$k  n=$n2  xy=$x,$y dxdy=$dx,$dy  by_path=$by_path by_flat=$by_flat\n";
          last K if $bad++ > 10;
        }
      }
    }
    is($bad,0, $flat->{'name'});
  }
}

# check accepted N against path
{
  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  my $bad = 0;
 K: foreach my $k (0 .. 12) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      my ($ex,$ey) = (2*$x,2*$y);
      my $by_path = $path->xy_to_n($ex,$ey) <= $n_hi ? 1 : 0;
      my $n2 = sprintf '%0*s', $k, cnv($n,10,2);
      my $n2rev = reverse $n2; # scalar context
      my $by_flat = MinusDoublePred_LtoH_flat()->contains($n2rev) ? 1 : 0;
      if ($by_path != $by_flat) {
        print "k=$k  n=$n2  xy=$x,$y  by_path=$by_path by_flat=$by_flat\n";
        last K if $bad++ > 10;
      }
    }
  }
  is($bad,0);
}

#------------------------------------------------------------------------------
# MinusNegB

{
  my @MinusNegB_flats = map {
    my $dir4 = $_;
    MinusNegPred_LtoH_flat()
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat(2*$dir4)->complement)
      ->MyFLAT::set_name("dir4=$dir4");
  } 0 .. 3;

  my $max_k = 20;
  my @count_arefs = map {
    my $dir4 = $_;
    [ MyFLAT::FLAT_count_contains ($MinusNegB_flats[$dir4], $max_k) ]
  } 0 .. 3;
  ### counts 0: join(', ',@{$count_arefs[0]})
  ### counts 1: join(', ',@{$count_arefs[1]})
  ### counts 2: join(', ',@{$count_arefs[2]})
  ### counts 3: join(', ',@{$count_arefs[3]})

  is_deeply($count_arefs[0],$count_arefs[2]);
  is_deeply($count_arefs[1],$count_arefs[3]);

  my @total = map {
    my $k = $_;
    sum(map {$count_arefs[$_]->[$k]} 0 .. 3)
  } 0 .. $max_k;
  ### total: join(', ',@total)

  my $want = [4, 4, 4, 8, 12, 36, 56, 80, 136, 228, 460, 784];
  is_deeply([@total[0 .. $#$want]], $want,
            'MinusNegB sample values');

  is($MinusNegB_flats[0]->num_states, 34);
  is($MinusNegB_flats[2]->num_states, 34);
  is($MinusNegB_flats[1]->num_states, 53);
  is($MinusNegB_flats[3]->num_states, 53);
}


#------------------------------------------------------------------------------
# MinusNegPred_LtoH_flat()

# check accepted N against path
{
  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  my $bad = 0;
 K: foreach my $k (0 .. 12) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      my $by_path = $path->xy_to_n(-$x,-$y) <= $n_hi ? 1 : 0;
      my $n2 = sprintf '%0*s', $k, cnv($n,10,2);
      my $n2rev = reverse $n2; # scalar context
      my $by_flat = MinusNegPred_LtoH_flat()->contains($n2rev) ? 1 : 0;
      if ($by_path != $by_flat) {
        print "k=$k  n=$n2  xy=$x,$y  by_path=$by_path by_flat=$by_flat\n";
        last K if $bad++ > 10;
      }
    }
  }
  is($bad,0);
}

# MinusNegIncPred_LtoH_flat()
{
  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  foreach my $dir8 (0 .. 7) {
    my $flat = MinusNegIncPred_dir8_LtoH_flat($dir8);
    my $dx = $dir8_to_dx[$dir8];
    my $dy = $dir8_to_dy[$dir8];
    # $flat->MyFLAT::view;

    my $bad = 0;
  K: foreach my $k (0 .. 8) {
      my ($n_lo, $n_hi) = $path->level_to_n_range($k);
      foreach my $n ($n_lo .. $n_hi-1) {
        my ($x,$y) = $path->n_to_xy($n);
        my $by_path = $path->xy_to_n(-$x+$dx,-$y+$dy) <= $n_hi ? 1 : 0;
        my $n2 = sprintf '%0*s', $k, cnv($n,10,2);
        my $n2rev = reverse $n2; # scalar context
        my $by_flat = $flat->contains($n2rev) ? 1 : 0;
        if ($by_path != $by_flat) {
          print "k=$k  n=$n2  xy=$x,$y dxdy=$dx,$dy  by_path=$by_path by_flat=$by_flat\n";
          last K if $bad++ > 10;
        }
      }
    }
    is($bad,0, $flat->{'name'});
  }
}

#------------------------------------------------------------------------------
# MinusIncPred_dir8_LtoH_flat()

{
  # all even N has a neighbour to its East,
  # so $initial_0 is a subset of IncPred(0)
  require FLAT::Regex;
  require MyFLAT;
  my $initial_0 = FLAT::Regex->new('0 (0|1)*')->as_dfa;
  ok ($initial_0->is_subset_of(MinusIncPred_dir8_LtoH_flat(0)));
}
{
  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  foreach my $dir8 (0 .. 7) {
    my $flat = MinusIncPred_dir8_LtoH_flat($dir8);
    my $dx = $dir8_to_dx[$dir8];
    my $dy = $dir8_to_dy[$dir8];
    # $flat->MyFLAT::view;

    my $bad = 0;
  K: foreach my $k (0 .. 8) {
      my ($n_lo, $n_hi) = $path->level_to_n_range($k);
      foreach my $n ($n_lo .. $n_hi-1) {
        my ($x,$y) = $path->n_to_xy($n);
        my $by_path = $path->xy_to_n($x+$dx,$y+$dy) <= $n_hi ? 1 : 0;
        my $n2 = sprintf '%0*s', $k, cnv($n,10,2);
        my $n2rev = reverse $n2; # scalar context
        my $by_flat = $flat->contains($n2rev) ? 1 : 0;
        if ($by_path != $by_flat) {
          print "k=$k  n=$n2  xy=$x,$y dir8=$dir8 dxdy=$dx,$dy  by_path=$by_path by_flat=$by_flat\n";
          last K if $bad++ > 10;
        }
      }
    }
    is($bad,0, $flat->{'name'});
  }
}

#------------------------------------------------------------------------------
# OXpred_flat()

# OXpred_flat()->MyFLAT::view;
{
  my $k = 8;
  my @a = grep {OXpred_flat()->contains(sprintf '%b',$_)} 0 .. 2**$k-1;
  is_deeply(\@a,[12,24,28,48,56,60,96,108,112,120,124,192,216,220,224,236,240,248,252]);

  # foreach (@a) { printf "%*b  %d\n", $k, $_, $_; }
  # foreach (12,24,28,48,56,60,96,108,112,120,124,192,216,220,224,236,240,248,252) { printf "%*b  %d\n", $k, $_, $_; }
}

{
  my $path = DragonCurve_path();
  my $bad = 0;
  my %xor;
  foreach my $k (0 .. 12) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n) or return;
      my @n_list = $path->xy_to_n_list($x,$y);
      @n_list = grep {$_<$n_hi} @n_list;
      next unless @n_list == 2;
      $xor{$n_list[0] ^ $n_list[1]} = 1;
    }
    foreach my $n ($n_lo .. $n_hi-1) {
      my $n2 = sprintf '%0*s', $k, cnv($n,10,2);
      my $by_flat = OXpred_flat()->contains($n2) ? 1 : 0;
      my $by_path = $xor{$n}  ? 1 : 0;
      if ($by_path != $by_flat) {
        print "k=$k  n=$n2  by_path=$by_path by_flat=$by_flat\n";
        last K if $bad++ > 10;
      }
    }
  }
  is($bad,0);
}


#------------------------------------------------------------------------------
# fNonRpred etc subsets of NonRpred eventually

# fNonRpred has its right side square enclosed, so subset of NonRpred,
# or rather of NonRpred eventually enclosed
ok (fNonRpred_flat()->is_subset_of($NonRpred_eventually_flat));
ok (fNonLpred_flat()->is_subset_of($NonLpred_k_eventually_flat));
ok (fNonBpred_flat()->is_subset_of($NonBpred_k_eventually_flat));


#------------------------------------------------------------------------------
# PaperConst not on boundary

{
  my $str = '';
  foreach my $n (1 .. 18) {
    $str .= Paper($n);
    {
      my $got = NonLpred_k_flat()->contains($str) ? 1 : 0;
      my $want = $n >= 10 ? 1 : 0;
      is($got,$want, "PaperConst Lpred_k_flat() n=$n  $str");
    }
    {
      my $got = $NonLpred_k_eventually_flat->contains($str) ? 1 : 0;
      my $want = $n >= 10 ? 1 : 0;
      is($got,$want, "PaperConst NonLpred_k_eventually_flat n=$n  $str");
    }
    {
      my $got = fNonLpred_flat()->contains($str) ? 1 : 0;
      my $want = $n >= 13 ? 1 : 0;
      is($got,$want, "PaperConst fLpred_k_flat() n=$n  $str");
    }

    {
      my $got = NonRpred_flat()->contains($str) ? 1 : 0;
      my $want = $n >= 6 ? 1 : 0;
      is($got,$want, "PaperConst Rpred_flat() n=$n  $str");
    }
    {
      my $got = $NonRpred_eventually_flat->contains($str) ? 1 : 0;
      my $want = $n >= 5 ? 1 : 0;
      is($got,$want, "PaperConst NonRpred_eventually_flat n=$n  $str");
    }
    {
      my $got = fNonRpred_flat()->contains($str) ? 1 : 0;
      my $want = $n >= 5 ? 1 : 0;
      is($got,$want, "PaperConst fRpred_flat() n=$n  $str");
    }
  }
}


#------------------------------------------------------------------------------
# fBpred extra cases

{
  # Spred +/- fractions vs formed strings

  my $extra_plus1_left_flat = fBpred_cases_extra('left',0,1);
  my $extra_plus11_left_flat = fBpred_cases_extra('left',0,11);
  my $extra_minus1_left_flat = fBpred_cases_extra('left',0,-1);
  my $extra_minus11_left_flat = fBpred_cases_extra('left',0,-11);

  my $extra_plus1_right_flat = fBpred_cases_extra('right',0,1);
  my $extra_plus11_right_flat = fBpred_cases_extra('right',0,11);
  my $extra_minus1_right_flat = fBpred_cases_extra('right',0,-1);
  my $extra_minus11_right_flat = fBpred_cases_extra('right',0,-11);

  my $extra_plus7_left_flat = fBpred_cases_extra('left',1,-8);
  my $extra_minus7_right_flat = fBpred_cases_extra('right',1,8);

  my $bad = 0;
  foreach my $elem (#                        parity frac turn
                    [$extra_plus1_left_flat,    0,    1,   1],
                    [$extra_plus1_right_flat,   0,    1,  -1],
                    [$extra_plus11_left_flat,   0,   11,   1],
                    [$extra_plus11_right_flat,  0,   11,  -1],

                    [$extra_minus1_left_flat,   0,   -1,   1],
                    [$extra_minus1_right_flat,  0,   -1,  -1],
                    [$extra_minus11_left_flat,  0,  -11,   1],
                    [$extra_minus11_right_flat, 0,  -11,  -1],

                    [$extra_minus7_right_flat,  1,    8,  -1],
                    [$extra_plus7_left_flat,    1,   -8,   1],
                   ) {
    my ($f, $parity,$frac,$turn) = @$elem;
    diag "$f->{name}  parity=$parity frac=$frac turn=$turn";

    my $count = 0;
    my $fraclen = 16;
  K: foreach my $k (8) {
      my $n_hi = 1 << $k;
      foreach my $n (1 .. $n_hi-1) {
        if ($n % 2 == $parity
            && Spred_k($k,$n)
            && ($n==0 || turn($n)==$turn)) {
          my $num = $n*15 + $frac;
          next unless $num >= 0;
          $num = int(($num << $fraclen) / 15);
          my $str = sprintf '%0*b', $k+$fraclen, $num;
          # print "$str  k=$k n=$n num=$num\n";
          $count++;

          foreach my $len (0 .. length($str)) {
            my $substr = substr($str,0,$len);
            if (! $f->contains($substr)) {
              diag "  missing $substr  k=$k n=$n len=$len\n";
              last K if $bad++ > 10;
            }
          }
        }
      }
    }
    diag "fractions count $count\n";
  }
  is ($bad, 0);
}


#------------------------------------------------------------------------------
# fBpred cases extras - none can be omitted

{
  # excluding any of the extras is not possible

  # 1/15 = 0001  4/15 = 0100 are repeating 100 for R
  # R no need +1/15, -11/15
  # L no need +11/15, -1/15

  my $bad = 0;
  {
    my $c = fNonRpred_flat();

    my @extras_R_array = fRpred_cases_extra_list();
    foreach my $exclude (0 .. $#extras_R_array) {
      my $f = NonRpred_flat()->MyFLAT::eventually_accepting;
      foreach my $i (0 .. $#extras_R_array) {
        next if $i == $exclude;
        $f = $f->difference($extras_R_array[$i]);
      }
      $f->MyFLAT::set_name("fNonRpred_cases sans-extra$exclude");
      $f = $f->MyFLAT::minimize;
      # diag("$f->{name} num states ", $f->num_states);

      ok(! $f->equals($c), "R excluding $exclude different, ok");

      # and with the omitted gives equality
      $f = $f->difference($extras_R_array[$exclude]);
      ok ($f->equals($c));
    }
  }
  {
    my $c = fNonLpred_flat();

    $c = $c->MyFLAT::minimize;
    my @extras_L_array = fLpred_cases_extra_list();
    foreach my $exclude (0 .. $#extras_L_array) {
      my $f = NonLpred_k_flat()->MyFLAT::eventually_accepting;
      foreach my $i (0 .. $#extras_L_array) {
        next if $i == $exclude;
        $f = $f->difference($extras_L_array[$i]);
      }
      $f->MyFLAT::set_name("fNonLpred_cases sans-extra$exclude");
      $f = $f->MyFLAT::minimize;
      # diag("$f->{name} num states ", $f->num_states);

      ok(! $f->equals($c), "L excluding $exclude different, ok");

      # and with the omitted gives equality
      $f = $f->difference($extras_L_array[$exclude]);
      ok ($f->equals($c));
    }
  }
  is ($bad, 0);
}

#------------------------------------------------------------------------------
# fBpred cases extras

{
  my $extra_plus1_left_flat
    = SpredLeft_k_flat()->intersect(bits_N_even_flat())
    ->MyFLAT::concat(MyFLAT::fraction_digits(1,15, radix => 2))  # +1/15
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_plus1_left');
  ok($extra_plus1_left_flat->equals(fBpred_cases_extra('left',0,1)));
}
{
  my $extra_plus11_left_flat
    = SpredLeft_k_flat()->intersect(bits_N_even_flat())
    ->MyFLAT::concat(MyFLAT::fraction_digits(11,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_plus11_left');
  ok($extra_plus11_left_flat->equals(fBpred_cases_extra('left',0,11)));

  {
    my %info = $extra_plus11_left_flat->MyFLAT::get_eventually_accepting_info;
    is ($info{'depth'}, 0,          'no more accepting prefixes');
    is_deeply ($info{'states'}, []);
  }


  # state 163 0111 repeating
  # 10001 0111 01    17 + 7/15 no
  # 1000 1011 101     8 + 11/15     turn(8)==1 left
  # (8 + 11/15)/16  == (17 + 7/15)/32
  # $extra_plus11_left_flat->contains('01000 1101 1101') or die;
  ok($extra_plus11_left_flat->contains('1') );
  ok($extra_plus11_left_flat->contains('10') );
  ok($extra_plus11_left_flat->contains('1000') );
  ok($extra_plus11_left_flat->contains('10001') );
  ok($extra_plus11_left_flat->contains('100010') );
  ok($extra_plus11_left_flat->contains('1000101') );
  ok($extra_plus11_left_flat->contains('10001011') );
  ok($extra_plus11_left_flat->contains('100010111') );
  ok($extra_plus11_left_flat->contains('1000101110') );
  ok($extra_plus11_left_flat->contains('10001011101') );
}

{
  my $extra_plus1_right_flat
    = SpredRight_k_flat()->intersect(bits_N_even_flat())
    ->MyFLAT::concat(MyFLAT::fraction_digits(1,15, radix => 2))  # +1/15
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_plus1_right');
  ok($extra_plus1_right_flat->equals(fBpred_cases_extra('right',0,1)));
}

{
  my $extra_plus11_right_flat
    = SpredRight_k_flat()->intersect(bits_N_even_flat())
    ->MyFLAT::concat(MyFLAT::fraction_digits(11,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_plus11_right');
  ok($extra_plus11_right_flat->equals(fBpred_cases_extra('right',0,11)));
}


# match "SpredLeft_runs 01" but with -1 borrow so 00
# empty runs also 01 is 1 and 1 -1 =  0
# this is an odd 2n+1 with borrow
my $SpredLeft_borrow_flat
  = SpredLeft_runs_flat()
  ->MyFLAT::concat(FLAT::Regex->new('00'))
  ->union( FLAT::Regex->new('0') )
  ->MyFLAT::set_name('SpredLeft borrow');
# match SpredRight_runs 11 but with -1 borrow so 10
# this is an odd 2n+1 with borrow
my $SpredRight_borrow_flat
  = SpredRight_runs_flat()
  ->MyFLAT::concat(FLAT::Regex->new('10'))
  ->MyFLAT::set_name('SpredRight borrow');

{
  my $extra_minus1_left_flat         # 2n - 1/15
    = $SpredLeft_borrow_flat
    ->MyFLAT::concat(FLAT::Regex->new('1 1*'),    # even 2n, borrowing
                     MyFLAT::fraction_digits(15-1,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_minus1_left');
  ok($extra_minus1_left_flat->equals(fBpred_cases_extra('left',0,-1)));
}
{
  my $extra_minus11_left_flat
    = $SpredLeft_borrow_flat
    ->MyFLAT::concat(FLAT::Regex->new('1 1*'),  # even 2n, borrowing
                     MyFLAT::fraction_digits(15-11,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_minus11_left');
  ok($extra_minus11_left_flat->equals(fBpred_cases_extra('left',0,-11)));
}

{
  my $extra_minus1_right_flat         # 2n - 1/15
    = $SpredRight_borrow_flat
    ->MyFLAT::concat(FLAT::Regex->new('1 1*'),    # even 2n, borrowing
                     MyFLAT::fraction_digits(15-1,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_minus1_right');
  ok($extra_minus1_right_flat->equals(fBpred_cases_extra('right',0,-1)));
}
{
  my $extra_minus11_right_flat
    = $SpredRight_borrow_flat
    ->MyFLAT::concat(FLAT::Regex->new('1 1*'),  # even 2n, borrowing
                     MyFLAT::fraction_digits(15-11,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_minus11_right');
  ok($extra_minus11_right_flat->equals(fBpred_cases_extra('right',0,-11)));
}

{
  my $extra_plus7_left_flat
    = $SpredLeft_borrow_flat
    ->MyFLAT::concat(MyFLAT::fraction_digits(15-8,15, radix => 2))
    ->MyFLAT::prefix
    ->MyFLAT::set_name('extra_plus7_left');
  # done as 2n+1 - 8/15
  ok($extra_plus7_left_flat->equals(fBpred_cases_extra('left',1,-8)));

  ok($extra_plus7_left_flat->contains('00000111011101'));
  ok($extra_plus7_left_flat->contains('0000111011101'));
  ok($extra_plus7_left_flat->contains('000111011101')); # 7/15 / 4
  ok($extra_plus7_left_flat->contains( '00111011101')); # 7/15 / 2
}
{
  my $extra_minus7_right_flat
    = do {
      my $f = SpredRight_k_flat()->intersect(bits_N_odd_flat())
        ->MyFLAT::concat(MyFLAT::fraction_digits(8,15, radix => 2));
      MyFLAT::prefix($f)->MyFLAT::set_name('extra_minus7_right');
    };
  # done as 2n+1 + 8/15
  ok($extra_minus7_right_flat->equals(fBpred_cases_extra('right',1,8)));

  # FIXME: what was this bit?
  # 2n+2 - 7/15
  # = 2n+1 + 8/15  with right at 2n+1 is 13 untraversed
  # ok($extra_minus7_right_flat->contains('11000100010'));  # (2-7/15)/2
  # (3+1/15)/4 == (2-7/15)/2
}

{
  my $extra_7_15_flat
    = MyFLAT::fraction_digits(7,15, radix => 2)
    ->MyFLAT::set_name('extra_7_15');
  ok($extra_7_15_flat->contains('0111011101')); # 7/15
}

#------------------------------------------------------------------------------
# fBpred by cases

{
  my $fNonRpred_by_cases_flat = do {
    diag "fNonRpred_cases_flat difference ...";
    my $f = $NonRpred_eventually_flat;

    foreach my $extra (fRpred_cases_extra_list()) {
      $f = $f->difference($extra);
    }
    $f->MyFLAT::set_name('fNonRpred_by_cases');
    # $f = $f->MyFLAT::minimize;
    $f;
  };
  my $fRpred_by_cases_flat
    = $fNonRpred_by_cases_flat
    ->complement
    ->MyFLAT::set_name('fRpred_by_cases');

  ok ($fNonRpred_by_cases_flat->equals(fNonRpred_flat()));
  ok ($fRpred_by_cases_flat->equals(fRpred_flat()));
}
{
  my $fNonLpred_by_cases_flat = do {
    diag "fNonLpred_cases_flat difference ...";
    my $f = $NonLpred_k_eventually_flat;

    foreach my $extra (fLpred_cases_extra_list()) {
      $f = $f->difference($extra);
    }
    $f->MyFLAT::set_name('fNonLpred_by_cases');
    # $f = $f->MyFLAT::minimize;
    $f;
  };
  my $fLpred_by_cases_flat
    = $fNonLpred_by_cases_flat
    ->complement
    ->MyFLAT::set_name('fLpred_by_cases');

  # my $fLpred_by_cases = do {
  #   my $f = Lpred_k_flat()->union(fLpred_cases_extra_list());
  #   $f = $f->as_dfa;
  #   $f->{'name'} = 'fLpred_by_cases';
  #   $f;
  # };

  ok ($fNonLpred_by_cases_flat->equals(fNonLpred_flat()));
  ok ($fLpred_by_cases_flat->equals(fLpred_flat()));
}

#------------------------------------------------------------------------------
# R,L,B squares

# ok_conf_expand_within_self($R_square_segs);
# ok_conf_expand_within_self($L_square_segs);
ok_conf_expand_within_self($B_squares_segs);


#------------------------------------------------------------------------------
# Spred_k_flat(), Dpred_k_flat()

{
  my $max_k = 10;
  ### list: FLAT_count_contains(Dpred_k_flat(),$max_k)
  ### D: map {D($_)} 0 .. $max_k

  # D counts locations x,y and there are 2 N's there so Dpred = 2*D
  is_deeply([ FLAT_count_contains(Dpred_k_flat(),$max_k) ],
            [ map {2*D($_)} 0 .. $max_k ]);

  # Spred_k_flat doesn't include N=2^k, so S-1
  is_deeply([ FLAT_count_contains(Spred_k_flat(),$max_k) ],
            [ map {S($_)-1} 0 .. $max_k ]);
}

ok(SpredLeft_runs_flat()->contains(''));

# anything with Spred at one end of segment is boundary
# (but only subset as boundary has some with doubles at both ends)
ok(Spred_k_flat()->is_subset_of( Bpred_k_flat() ));

ok(Spred_k_flat()->equals( Dpred_k_flat()->complement ));
ok(Spred_k_flat() ->intersect(Dpred_k_flat())->is_empty); # disjoint

ok(DpredRight_k_flat()->is_subset_of(TurnRpred_flat()));
ok(DpredLeft_k_flat() ->is_subset_of(TurnLpred_flat()));
ok(SpredRight_k_flat()->is_subset_of(TurnRpred_flat()));
ok(SpredLeft_k_flat() ->is_subset_of(TurnLpred_flat()));

# left, right disjoint
ok(SpredLeft_k_flat() ->intersect(SpredRight_k_flat())->is_empty);
ok(DpredLeft_k_flat() ->intersect(DpredRight_k_flat())->is_empty);

# whole = left + right
ok(Dpred_k_flat()->equals( DpredRight_k_flat()
                           ->union(DpredLeft_k_flat()) ->as_dfa ));
ok(Spred_k_flat()->equals( SpredRight_k_flat()
                           ->union(SpredLeft_k_flat())
                           ->union(zero_digits_flat(2))   ->as_dfa ));

ok(  Spred_k_flat()->contains('111'));  # N=7
ok(! Dpred_k_flat()->contains('111'));
ok(  Dpred_k_flat()->contains(  '0111'));  # in k=4
ok(  Dpred_k_flat()->contains( '00111'));  # in k=5
ok(  Dpred_k_flat()->contains('000111'));  # in k=6

# double-visiteds in k are a subset of those in k+1,
# with extra doubles at the join
foreach my $f (Dpred_k_flat(),
               DpredLeft_k_flat(),
               DpredRight_k_flat()) {
  my $c = FLAT::Regex->new('0')->as_dfa->MyFLAT::concat($f)->as_min_dfa;
  ok($c->is_subset_of($f),
     "$f->{name} extra leading zero, k is subset of k+1");
}

foreach my $len (0 .. 3) {          # N=0
  my $str = '0' x $len;
  ok(Spred_k_flat()     ->contains($str));
  ok(! SpredRight_k_flat()->contains($str));
  ok(! SpredLeft_k_flat() ->contains($str));
  ok(! Dpred_k_flat()     ->contains($str));
  ok(! DpredLeft_k_flat() ->contains($str));
  ok(! DpredRight_k_flat()->contains($str));
}

{
  my $path = Math::PlanePath::DragonCurve->new;
  my $bad = 0;
 K: foreach my $k (0 .. 10) {
    my $n_hi = 1 << $k;
    ### $n_hi
    foreach my $n (0 .. $n_hi) {
      my $str = sprintf '%0*b', $k, $n;
      my ($x,$y) = $path->n_to_xy($n);
      my @n_list = $path->xy_to_n_list($x,$y);
      my $turn = $n>=1 && turn($n);
      my $t_left  = $turn==1;
      my $t_right = $turn==-1;
      ### @n_list
      ### $turn

      # no DpredInf_flat() yet
      # {
      #   my $p_pred = @n_list == 2;
      #   my $f_pred = DpredInf_flat()->contains($str) ? 1 : 0;
      #   if ($p_pred != $f_pred) {
      #     print "Dpred_k_flat() at N=$n $str got $f_pred want path=$p_pred\n";
      #     last K if $bad++ > 10;
      #   }
      # }

      @n_list = grep {$_ <= $n_hi} @n_list;
      {
        my $f = Dpred_k_flat();
        my $p_pred = @n_list == 2 ? 1 : 0;
        my $f_pred = $f->contains($str) ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got $f_pred want path=$p_pred\n";
          last K if $bad++ > 10;
        }
      }
      {
        my $f = DpredLeft_k_flat();
        my $p_pred = @n_list == 2 && $t_left ? 1 : 0;
        my $f_pred = $f->contains($str) ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got $f_pred want path=$p_pred\n";
          last K if $bad++ > 10;
        }
      }
      {
        my $f = DpredRight_k_flat();
        my $p_pred = @n_list == 2 && $t_right ? 1 : 0;
        my $f_pred = $f->contains($str) ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got $f_pred want path=$p_pred\n";
          last K if $bad++ > 10;
        }
      }
    }
  }
  is($bad, 0, "Dpred_flat() vs path");
}
# for(n=0,31,if(Dpred(n),print(to_binary(n))))

# sub check_flat_vs_predicate {
#   my ($flat, $func, $name) = @_;
#   my $bad = 0;
#   foreach my $k (1 .. 10) {
#     foreach my $n (0 .. 2**$k-1) {
#       my $str = sprintf '%0*b', $k, $n;
#       my $got = $flat->contains($str) ? 1 : 0;
#       my $want = $func->($k,$n) ? 1 : 0;
#       if ($got != $want) {
#         print "$name k=$k n=$n $str got $got want $want\n";
#         exit if $bad++ > 10;
#       }
#     }
#   }
# }
#   check_flat_vs_predicate($DpredLeft_k_flat, \&DpredLeft_by_path, 'DpredLeft');
#   check_flat_vs_predicate($DpredRight_k_flat, \&DpredRight_by_path, 'DpredRight');
#   check_flat_vs_predicate($SpredLeft_flat, \&SpredLeft_by_path, 'SpredLeft');
#   check_flat_vs_predicate($SpredRight_flat, \&SpredRight_by_path, 'SpredRight');

# sub Spred_k_by_path {
#   my ($k,$n) = @_;
#   my ($x,$y) = $path->n_to_xy($n);
#   my @n_list = $path->xy_to_n_list($x,$y);
#   my ($n_lo, $n_hi) = $path->level_to_n_range($k);
#   @n_list = grep {$_ <= $n_hi} @n_list;
#   return @n_list==1;
# }
# sub SpredLeft_by_path {
#   my ($k,$n) = @_;
#   return $n>=1 && turn($n)==1 && Spred_k_by_path($k,$n);
# }
# sub SpredRight_by_path {
#   my ($k,$n) = @_;
#   return $n>=1 && turn($n)==-1 && Spred_k_by_path($k,$n);
# }
# sub DpredRight_by_path {
#   my ($k,$n) = @_;
#   return $n>=1 && turn($n)==-1 && ! Spred_k_by_path($k,$n);
# }
# sub DpredLeft_by_path {
#   my ($k,$n) = @_;
#   return $n>=1 && turn($n)==1 && ! Spred_k_by_path($k,$n);
# }

#------------------------------------------------------------------------------
# fFifthsPred_flat()

ok(zero_digits_flat(2)->is_subset_of(fFifthsPred_flat()));

{
  require Math::PlanePath::GcdRationals;
  my $gcd = \&Math::PlanePath::GcdRationals::_gcd;

  my $is_fifth = sub {
    my ($num, $den) = @_;  # with no common factor
    foreach (0, 1) {
      my $d = $den;
      until ($d & 1) { $d >>= 1; }
      if ($d==5 && ($num==1 || $num==2)) { return 1; }
      $num = $den-$num;
    }
    return 0;
  };

  my $len = 10;
  my $bad = 0;
 DEN: foreach my $den (1 .. 50) {
    foreach my $num (1 .. $den-1) {
      next unless $gcd->($num,$den) == 1;
      my $want = $is_fifth->($num, $den);
      my $n = int( (1 << $len) * $num/$den );
      my $full_str = sprintf '%0*b', $len, $n;
      foreach my $i (0 .. $len) {
        my $str = substr($full_str,0,$len);
        my $got = fFifthsPred_flat()->contains($str) ? 1 : 0;

        # at $i==$len must have got=want as have seen enough bits,
        # prior to then got might accept more than want
        unless ($i==$len ? $got == $want : $got >= $want) {
          print "fFifthsPred_flat() $num/$den N=$n $str got $got want $want\n";
          $bad++;
          last DEN if $bad++ > 10;
        }
      }
    }
  }
  is($bad, 0, "fFifthsPred_flat() on fractions");
}

{
  my $bad = 0;
  my $len = 30;
  foreach my $k (0 .. 10) {
    foreach my $rev (0, 1) {
      my $n = ($rev 
               ? int((1-2/(5*(1<<$k))) * (1<<$len))
               : int((  2/(5*(1<<$k))) * (1<<$len)));
      my $str = sprintf '%0*b', $len, $n;
      my $got = fFifthsPred_flat()->contains($str) ? 1 : 0;
      if (! $got) {
        print "fFifthsPred_flat() N=$n $str got $got\n";
        $bad++;
      }
    }
  }
  is($bad, 0, "fFifthsPred_flat() on 2/(5*2^k) and 1-2/(5*2^k)");
}

#------------------------------------------------------------------------------
# NonRpred_flatregex()

ok(NonRpred_flatregex() ->as_dfa->equals(NonRpred_flat()));
# my $NonRdiff = $NonRpred_flatregex->as_dfa->difference($NonRpred_flat->as_dfa);


#------------------------------------------------------------------------------
# NonBlobPredInf_flat()

{
  my $bad = 0;
  my $n = 0;
 K: foreach my $k (4 .. 12) {
    my $try = sub {
      my ($want) = @_;
      my $str = sprintf '%b', $n;
      {
        my $got = NonBlobPredInf_flat()->contains($str) ? 1 : 0;
        if ($got != $want) {
          print "NonBlobPredInf_flat() N=$n $str got $got want $want\n";
          $bad++;
        }
      }
    };
    {
      my $blob_n = BlobN($k);         # BlobNend to BlobN-1 inclusive are non-blob
      for ( ; $n < $blob_n; $n++) {
        $try->(1); last K if $bad > 10;
      }
    }
    {
      my $blob_n_end = BlobNend($k);  # BlobNend to BlobN-1 inclusive are blob
      for ( ; $n < $blob_n_end; $n++) {
        $try->(0); last K if $bad > 10;
      }
    }
  }
  is($bad, 0, "NonBlobPredInf_flat() vs BlobN()");
}

#------------------------------------------------------------------------------


sub path_n_is_right_boundary {
  my ($path, $n, $n_limit) = @_;
  ### path_n_is_right_boundary(): $n
  ### $n_limit
  if (defined $n_limit && $n >= $n_limit) {
    croak "path_n_is_right_boundary() $n >= limit $n_limit";
  }
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to right side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit));
}
sub path_n_is_left_boundary {
  my ($path, $n, $n_limit) = @_;
  ### path_n_is_left_boundary(): $n
  if ($n >= $n_limit) {
    croak "path_n_is_left_boundary() $n >= limit $n_limit";
  }
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to left side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  if (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)) {
    ### x1 y1 to s ...
    return 1;
  }
  if (! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)) {
    ### x2 y2 to s ...
    return 1;
  }
  if (! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit)) {
    ### s to s ...
    return 1;
  }
  ### not left boundary ...
  return 0;
}
sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  ### xyxy_is_traversed_either(): "$x1,$y1 to $x2,$y2"
  my $arms = $path->arms_count;
  foreach my $n1 ($path->xy_to_n_list($x1,$y1)) {
    next if defined $n_limit && $n1 > $n_limit;
    foreach my $n2 ($path->xy_to_n_list($x2,$y2)) {
      next if defined $n_limit && $n2 > $n_limit;
      if (abs($n1-$n2) == $arms) {
        ### yes: "$n1 to $n2"
        return 1;
      }
    }
  }
  return 0;
}




#------------------------------------------------------------------------------
exit 0;
