#!/usr/bin/perl -w

# Copyright 2020, 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use warnings;
use 5.010;
use File::Temp;
use FLAT;
use File::Slurp;
use Math::BaseCnv 'cnv';
use IPC::Run;
use Test::More;

use FindBin;
use lib $FindBin::Bin;
use MyFLAT ':all';
use MyDragon ':all';

use MyTestHelpers;
BEGIN { MyTestHelpers::nowarnings(); }

plan tests => 628;

# uncomment this to run the ### lines
# use Smart::Comments;


my $dragon_foma_filename = File::Spec->catfile($FindBin::Bin,
                                               'dragon.foma');

my $foma_tempfile_object = File::Temp->new (SUFFIX => '.foma');
my $foma_tempfile = $foma_tempfile_object->filename;
my $att_tempfile_object = File::Temp->new (SUFFIX => '.att');
my $att_tempfile = $att_tempfile_object->filename;

# uncomment to force filenames for debugging
# $foma_tempfile = '/tmp/dragon-foma-tempfile.tmp';
# $att_tempfile  = '/tmp/dragon-foma-tempfile.att';

my $devnull_filename = File::Spec->devnull;
diag "temp file $foma_tempfile";
diag "temp file $att_tempfile";
diag "File::Spec->devnull is $devnull_filename";

sub run_foma_to_flat {
  my ($regex) = @_;
  # Debian packaged foma 1:0.9.18+r243-6 got segfaults on the Rsides union.
  # Don't know why, but running the script from a second -l works whereas
  # piping it to stdin with '<',\$script doesn't work.
  # But then another Rsides related didn't like that, so '<',filename direct.

  File::Slurp::write_file($att_tempfile,'');
  -s $att_tempfile == 0 or die;
  my $script = <<"HERE";
regex $regex;
write att $att_tempfile
HERE
  ### $script
  File::Slurp::write_file($foma_tempfile,$script);

  my @command = ('foma',
                 # '-p', # pipe mode, medium quiet
                 '-q', # quiet mode, quieter
                 '-l', $dragon_foma_filename,
                );
  ### run: join(' ',@command)." < $devnull_filename"
  IPC::Run::run(\@command,
                '<', $foma_tempfile,
                '>', \my $outstr, '2>&1')
      or die "Error exit from foma: $?";
  if ($outstr =~ /\*\*\*/) {
    print "$outstr";
    print "Script\n$script";
    print "Foma error\n";
    exit 1;
  }

  # no, a couple are expected to be empty ...
  # -s $att_tempfile > 0 or warn "oops, empty $att_tempfile";

  return FLAT::DFA->MyFLAT::read_att_file($att_tempfile)
    ->MyFLAT::FLAT_rename(symbols_map => {z=>0})
    ->MyFLAT::set_name("foma:$regex");
}

{
  # [] no good for an empty regex, it seems
  my %to = (0 => ' z ',
            1 => ' 1 ',
            '(' => '[',
            ')' => ']',
            "\n" => ' ',
            "[]" => '0',
           );
  sub regex_FLAT_to_foma {
    my ($str) = @_;
    $str =~ s{[01()\n]}{$to{$&}}ge;
    $str =~ s/  */ /g;
    $str =~ s/ +\*/*/g;
    return $str;
  }
}

#------------------------------------------------------------------------------
# Various

my $flat_null  = FLAT::Regex->new('#')->as_dfa
  ->MyFLAT::set_name('null');
my $flat_empty = FLAT::Regex->new('[]')->as_dfa
  ->MyFLAT::set_name('empty string');
my $flat_any  = MyFLAT::bits_of_length_or_more_flat(0);
my $flat_any_nonzero = FLAT::Regex->new('(0|1)* 1 (0|1)*')->as_dfa
  ->MyFLAT::set_name('any non-zero');

# for foma regexp funcs which give empty string regexp for true and null
# language (no strings) for false
my $flat_true  = FLAT::Regex->new('[]')->as_dfa->MyFLAT::set_name('true');
my $flat_false = $flat_null->clone->MyFLAT::set_name('false');;

foreach my $elem
  (
   # flip1z example
   ['[ [z z 1] .o. flip1z ].l', FLAT::Regex->new('1 1 0')->as_dfa ],

   # StripLeading0s examples
   ['[ [z z 1 z 1] .o. StripLeading0s ].l',
    FLAT::Regex->new('1 0 1')->as_dfa ],
   ['[ [1 z z 1] .o. StripLeading0s ].l',
    FLAT::Regex->new('1 0 0 1')->as_dfa ],

   # OptionalLeading0s example
   ['[ [z z 1 z 1] .o. OptionalLeading0s ].l',
    FLAT::Regex->new('0* 1 0 1')->as_dfa ],

   ["[ [    1 1] .o. increment ].l", FLAT::Regex->new('100')->as_dfa ],
   ["[ [1 z 1 1] .o. increment ].l", FLAT::Regex->new('1100')->as_dfa ],
   ["[ [  1 z z] .o. increment ].l", FLAT::Regex->new('101')->as_dfa ],
   ["[        z  .o. increment ].l", FLAT::Regex->new('1')->as_dfa ],
   ["[        0  .o. increment ].l", FLAT::Regex->new('1')->as_dfa ],
   # inverse of increment is decrement
   ["[ increment .o. [1 z 1 1] ].u", FLAT::Regex->new('1010')->as_dfa ],
   ["_isunambiguous(increment)", $flat_true ],
   ["_isidentity(increment)",    $flat_false ],

   ["[ [1 z 1 | z] .o. decrement ].l", FLAT::Regex->new('0* 100')->as_dfa ],
   ["[ [  z z z] .o. decrement ].l", $flat_null ],
   ["[ [      z] .o. decrement ].l", $flat_null ],
   ["[ [      1] .o. decrement ].l", FLAT::Regex->new('0*')->as_dfa ],
   ["[ [    1 1] .o. decrement ].l", FLAT::Regex->new('0* 10')->as_dfa ],
   ["[ [1 1 z z] .o. decrement ].l", FLAT::Regex->new('0* 1011')->as_dfa ],
   # ["_isunambiguous(decrement)", $flat_true ],
   ["_isidentity(decrement)",               $flat_false ],
   # ["_isidentity(increment .o. decrement)", $flat_true ],

   # MinusNegXpred
   ['MinusNegXpred',
    FLAT::Regex->new('(0 | 11100 | 100)*')->as_dfa ],

   # MinusNegPred
   ['MinusNegPredLtoH', MinusNegPred_LtoH_flat() ],
   ['[1 z 1 1 1] & MinusNegPredLtoH', FLAT::Regex->new('11111')->as_dfa,
    alphabet => 1],
   ['[] & MinusNegPredLtoH', FLAT::Regex->new('[]')->as_dfa,
    alphabet => 1],

   # MinusNegLtoH transducer
   ['[[1 1 1 z z] .o. MinusNegLtoH].l', FLAT::Regex->new('11000')->as_dfa],
   ['[[1 1 z 1 z] .o. MinusNegLtoH].l', FLAT::Regex->new('11110')->as_dfa],
   ['[[1 z z 1 z] .o. MinusNegLtoH].l', FLAT::Regex->new('10101')->as_dfa],
   # low side equals MinusNegPredLtoH
   ['[MinusNegPredLtoH .o. MinusNegLtoH].l - MinusNegPredLtoH',
    $flat_null],
   ['MinusNegPredLtoH - [MinusNegPredLtoH .o. MinusNegLtoH].l',
    $flat_null],

   ['x^{3}', FLAT::Regex->new('x x x')->as_dfa,
    alphabet => 1],
   ['x^{3,4}', FLAT::Regex->new('x x x (x | [])')->as_dfa,
    alphabet => 1],

   ['w x y z', FLAT::Regex->new('w x y z')->as_dfa,
    alphabet => 1],
   # right quotient for prefix of a language,
   # A///B strips B from end of A
   ['[w x y z] /// ?*',
    FLAT::Regex->new('w x y z | w x y | w x | w | []')->as_dfa,
    alphabet => 1],

   ['DpredFirstRinf', DpredFirstR_inf_flat()],
   ['DpredFirstLinf', DpredFirstL_inf_flat()],
   ['DpredFirstRinf & TurnRpred', $flat_null ], # no rights
   ['DpredFirstLinf & TurnLpred', $flat_null ], # no lefts
   ['DpredFirstRinf & DpredFirstLinf', $flat_null ], # disjoint
   ['DpredFirstInf & DpredInf', DpredFirst_inf_flat()],
   ['DpredFirstInf & DpredInf', DpredFirst_inf_flat()],

   ['DpredSecondL', DpredSecondL_flat()],
   ['DpredSecondR', DpredSecondR_flat()],
   ['DpredSecond',  DpredSecond_flat()],
   ['DpredSecondL & TurnRpred', $flat_null ], # no rights
   ['DpredSecondR & TurnLpred', $flat_null ], # no lefts
   ['DpredSecondL & DpredSecondR', $flat_null ], # disjoint

   ['DpredInf',  Dpred_inf_flat()],

   # other(n)
   ['[DpredFirstInf .o. other].l', DpredSecond_flat()],
   ['[DpredSecond   .o. other].l', DpredFirst_inf_flat()],

   # 7 bits match
   ['[1 1    ] & otherLowFlipLtoH1', FLAT::Regex->new('11')->as_dfa ],
   ['[1 1 1  ] & otherLowFlipLtoH1', FLAT::Regex->new('111')->as_dfa ],
   ['[1 1 1 z] & otherLowFlipLtoH1', $flat_null ],

   # # 7 -> 11 in 4 bits
   # ['[[1 1 1 z] .o. otherLtoH].l', FLAT::Regex->new('1101')->as_dfa
   #  ->MyFLAT::set_name('1101 in 4 bits') ],
   # 7 -> 11
   ['[[1 1 1 z z*] .o. otherLtoH].l',
    FLAT::Regex->new('1 1 0 1 0*')->as_dfa ],

   ['Pagoda1.r', FLAT::Regex->new('0 1 1 1*
                                  |0 1 1 1* 0
                                  |0 1 1 1* 0 0 (0|1)*
                                  |0 0 1 1* 0 1 (0|1)*
                                  |1          1 (0|1)*')->as_dfa ],
   ['Pagoda2.r', FLAT::Regex->new('(1 | 0 0 0 1* 0) ([] | 0 (0|1)*)
                                   |[]
                                   |0
                                   |0 0
                                   |0 0 0 1*
                                   |0 0 0 1* 0
                                   |0 1 0 1* 0 1 (0|1)*')->as_dfa ],
   ['Pagoda0.r', FLAT::Regex->new('0 (00|11) 1* 0 1 (0|1)*
                                   |0 1
                                   |(0 0 1
                                    |0 1 0) 1* ([]
                                               |0
                                               |0 0
                                               |0 0 (0|1)*)')->as_dfa ],

   ['Pagoda0 | Pagoda1 | Pagoda2', $flat_any ], # union
   ['Pagoda0 & Pagoda1', $flat_null ], # disjoint
   ['Pagoda0 & Pagoda2', $flat_null ], # disjoint
   ['Pagoda1 & Pagoda2', $flat_null ], # disjoint
   ["Pagoda2 & []", $flat_empty ],  # empty string in 2

   ["OptionalLeading0sFunc(z)",    FLAT::Regex->new('0*')->as_dfa],
   ["[z .o. OptionalLeading0s].l", FLAT::Regex->new('0*')->as_dfa],

   ['Rsides1', Rsides_flat(1) ],
   ['Rsides2', Rsides_flat(2) ],
   ['Rsides3', Rsides_flat(3) ],
   ['Rsides1 & Rsides2', $flat_null ], # disjoint
   ['Rsides1 & Rsides3', $flat_null ], # disjoint
   ['Rsides2 & Rsides3', $flat_null ], # disjoint
   ['Rsides1 | Rsides2 | Rsides3', Rpred_flat() ], # union

   # ['[dirP .o. P0mod4].u',  dir4_flat(0)],
   # ['[dirP .o. P1mod4].u',  dir4_flat(1)],
   # ['[dirP .o. P2mod4].u',  dir4_flat(2)],
   # ['[dirP .o. P3mod4].u',  dir4_flat(3)],
   ['dir0mod4',  dir4_flat(0)],
   ['dir1mod4',  dir4_flat(1)],
   ['dir2mod4',  dir4_flat(2)],
   ['dir3mod4',  dir4_flat(3)],
   ['dir0mod4 & [dir1mod4 | dir2mod4 | dir3mod4]', $flat_null ],
   ['dir1mod4 & [dir0mod4 | dir2mod4 | dir3mod4]', $flat_null ],
   ['dir2mod4 & [dir0mod4 | dir1mod4 | dir3mod4]', $flat_null ],
   ['dir3mod4 & [dir0mod4 | dir1mod4 | dir2mod4]', $flat_null ],
   ['dir0mod4 & [ any 1 ]', $flat_null ], # no odds
   ['dir0mod4 & [ any z ]', $flat_null ], # no evens
   ['dir0mod4 & [ any 1 ]', $flat_null ], # no odds
   ['dir0mod4 & [ any z ]', $flat_null ], # no evens
   ["dir0mod4 | dir1mod4 | dir2mod4 | dir3mod4", $flat_any ],  # all n

   ['StateMachine("@P.s.1e@", RpredStateTable, "@D.s.non@")', Rpred_flat()],
   ['StateMachine("@P.s.1e@", RpredStateTable, "@R.s.non@")', NonRpred_flat()],

   ["Rpred",    Rpred_flat()],
   ["NonRpred", NonRpred_flat()],
   ["[ Rpred .o. flip1z ].l",
    Rpred_flat()->MyFLAT::FLAT_rename(symbols_map=>{0=>1,1=>0}) ],
   ['Rpred | NonRpred', $flat_any ],

   ["LpredInf",    LpredInf_flat()],
   ["NonLpredInf", NonLpredInf_flat()],
   ['LpredInf & NonLpredInf', $flat_null ], # disjoint
   ['LpredInf | NonLpredInf', $flat_any ],

   ["BpredInf",    BpredInf_flat()],
   # ["NonBpredInf", NonBpredInf_flat()],
   # ['BpredInf & NonBpredInf', $flat_null ], # disjoint
   # ['BpredInf | NonBpredInf', $flat_any ],
   ["NonBlobPredInf", NonBlobPredInf_flat()],

   ['EpredL', EpredL_flat()],
   ['EpredR', EpredR_flat()],
   ['Epred',  Epred_flat()],
   ['EpredL & TurnRpred', $flat_null ], # no rights
   ['EpredR & TurnLpred', $flat_null ], # no lefts
   ['EpredL & EpredR', $flat_null ], # disjoint
   ['Epred & [Epred9 .o. decrement].l', $flat_null ], # no run >=10

   ["dNextL1", dNextL_flat(1)],
   ["dNextL2", dNextL_flat(2)],
   ["dNextL3", dNextL_flat(3)],
   ["dNextL4", dNextL_flat(4)],
   ['dNextL1 & [ dNextL2|dNextL3|dNextL4] ', $flat_null ],
   ['dNextL2 & [ dNextL1|dNextL3|dNextL4] ', $flat_null ],
   ['dNextL3 & [ dNextL1|dNextL2|dNextL4] ', $flat_null ],
   ['dNextL4 & [ dNextL1|dNextL2|dNextL3] ', $flat_null ],

   ['dNextR1', dNextR_flat(1)],
   ['dNextR2', dNextR_flat(2)],
   ['dNextR3', dNextR_flat(3)],
   ['dNextR4', dNextR_flat(4)],
   ['dNextR1 & [ dNextR2|dNextR3|dNextR4]', $flat_null ],
   ['dNextR2 & [ dNextR1|dNextR3|dNextR4]', $flat_null ],
   ['dNextR3 & [ dNextR1|dNextR2|dNextR4]', $flat_null ],
   ['dNextR4 & [ dNextR1|dNextR2|dNextR3]', $flat_null ],

   ['[ [1 z 1 1] .o. flip1z ].l', FLAT::Regex->new('0100')->as_dfa ],

   ['[ [z z z 1 z] .o. StripLeading0s ].l',
    FLAT::Regex->new('10')->as_dfa ],

   ['[ 1 .o. OptionalLeading0s ].l',
    FLAT::Regex->new('0* 1')->as_dfa ],
   ['[ [z z z 1 z] .o. OptionalLeading0s ].l',
    FLAT::Regex->new('0* 10')->as_dfa ],

   ['TurnLpred', TurnLpred_flat()],
   ['TurnRpred', TurnRpred_flat()],
   ['TurnLpred & TurnRpred', $flat_null ], # disjoint
   ['TurnLpred | TurnRpred', $flat_any_nonzero ],

   # all n != 0
   ["TurnLpred | TurnRpred",
    FLAT::Regex->new('(0|1)* 1 (0|1)*')->as_dfa ],

   [ regex_FLAT_to_foma($NonRpred_flatregex_str),
     NonRpred_flat() ],

   ["fFifthsPred",   fFifthsPred_flat()],
  ) {
  my ($regex, $flat, %options) = @$elem;
  my $foma = run_foma_to_flat($regex);
  isa_ok($flat, 'FLAT::DFA', 'flat');
  isa_ok($foma, 'FLAT::DFA', 'foma');

  unless ($options{'alphabet'}) {
    foreach my $felem ([$flat, $flat->{'name'} // 'flat'],
                       [$foma, $foma->{'name'}]) {
      my ($f,$name) = @$felem;
      my @alphabet = sort $f->alphabet;
      my @bad = grep {! ($_ eq '0' || $_ eq '1')} @alphabet;
      is (scalar(@bad), 0,
          "$name alphabet check"
          . (@bad ? ", bad: ".join(',',@bad) : ""));
    }
  }

  if (0) {
    $foma->MyFLAT::view;
    # $flat->MyFLAT::view;
    exit;
  }

  my $equals = $flat->equals($foma);
  ok ($equals, "$regex");
  if (! $equals) {
    MyFLAT::FLAT_check_is_equal($foma,$flat);
  }
}


#------------------------------------------------------------------------------
exit 0;
