#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use Math::PlanePath::ComplexMinus 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'xy_rotate_plus90','xy_rotate_minus90';
use MyLineSegments;
use MyPlanar ':all';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 10;
if (@ARGV) {
  $level = shift @ARGV;
}
print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::ComplexMinus->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);


sub points_square {
  my ($x,$y) = @_;
  return [ [2*$x-1, 2*$y-1],
           [2*$x+1, 2*$y-1],
           [2*$x+1, 2*$y+1],
           [2*$x-1, 2*$y+1] ];
}

my $plain_segments  = MyLineSegments->new (name => 'n');
my $other_segments = MyLineSegments->new (name => 'anyrot');
my $rotatable_segments = MyLineSegments->new (name => 'rotatable');
my $rotatable_gpc = MyPlanar::points_to_gpc(points_square(0,0));

my $MinusAnyRotA = 0;
foreach my $n ($n_lo .. $n_hi) {
  my ($x,$y) = $path->n_to_xy($n);
  my $points = points_square($x,$y);
  $plain_segments->add_cycle(map {@$_} @$points);

  if ($path->xy_to_n(xy_rotate_plus90($x,$y)) <= $n_hi
      && $path->xy_to_n(xy_rotate_minus90($x,$y)) <= $n_hi
      && $path->xy_to_n(-$x,-$y) <= $n_hi) {
    $rotatable_segments->add_cycle(map {@$_} @$points);
    my $gpc = MyPlanar::points_to_gpc($points);
    $rotatable_gpc = Math::Geometry::Planar::GpcClip('UNION',$rotatable_gpc,$gpc);
    $MinusAnyRotA++;
  }

  {
    my ($x,$y) = (-$x,-$y);
    $points = points_square($x,$y);
    $other_segments->add_cycle(map {@$_} @$points);
  }
  {
    my ($x,$y) = xy_rotate_plus90($x,$y);
    $points = points_square($x,$y);
    $other_segments->add_cycle(map {@$_} @$points);
  }
  {
    my ($x,$y) = xy_rotate_minus90($x,$y);
    $points = points_square($x,$y);
    $other_segments->add_cycle(map {@$_} @$points);
  }
}

# Remove $negatable_segments, they will be a filled area.
# By symmetry, $other_segments also contains $rotatable_segments, which is
# not true for instance just in MinusRot.
#
$rotatable_segments->is_subset_of($plain_segments) or die;
$rotatable_segments->is_subset_of($other_segments) or die;
$plain_segments->remove_segments($rotatable_segments);
$other_segments->remove_segments($rotatable_segments);

# remove $plain_segments, so $other_segments is rest of negated copy
$other_segments->remove_segments($plain_segments);
$plain_segments->is_disjoint($other_segments) or die;

my $MinusAnyRotB = 0;
{
  my $A = 0;
  my @contours = Math::Geometry::Planar::Gpc2Polygons($rotatable_gpc);
  my $total_holes = 0;
  foreach my $contour (@contours) {  # Math::Geometry::Planar contour
    ### $contour
    # Second and subsequent polygon of the contour are holes.
    # Think their points already cycle the opposite way as required by pgf
    # default \fill[nonzero rule], but using \fill[even odd rule] works either
    # way.
    $Printwrap = 2;
    printwrap(" \\fill[my rotatable,even odd rule]");
    $Printwrap += 2;
    my $polygons = $contour->polygons;  # arrayref of points arrayrefs
    $total_holes += scalar(@$polygons)-1;
    my $separator = '';
    foreach my $points (@$polygons) {
      ### $points
      printwrap($separator);
      $separator = "\n";
      printwrap_poly($points);
    }
    printwrap(";\n");
    $MinusAnyRotB += $contour->perimeter / 2;
    $A += $contour->area / 4;
  }
  $total_holes == 0 or die;
  $A == $MinusAnyRotA or die;
}
### $plain_segments

$Printwrap = 2;
printwrap("\\draw[my plain line] ");
$Printwrap += 2;
$plain_segments->printwrap;
printwrap(";\n");

$Printwrap = 2;
printwrap("\\draw[my other line] ");
$Printwrap += 2;
$other_segments->printwrap;
printwrap(";\n");

print "% GP-Test  MinusAnyRotA_by_pred($level) == $MinusAnyRotA\n";
print "% GP-Test  MinusAnyRotB_by_pred($level) == $MinusAnyRotB\n";
print "  \\newcommand{\\MyMinusAnyRotA}{$MinusAnyRotA}\n";
print "  \\newcommand{\\MyMinusAnyRotB}{$MinusAnyRotB}\n";
print "  % end generated\n";
exit 0;
