#!/usr/bin/perl -w

# Copyright 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use List::Util 'max';
use Math::PlanePath::ComplexMinus 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'xy_rotate_plus90','@dir4_to_dx','@dir4_to_dy';
use MyLineSegments;
use MyPlanar;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 8;
if (@ARGV) {
  $level = shift @ARGV;
}
print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::ComplexMinus->new;

my $max_n = 0;
my $max_n_double = 0;
my $MinusDoubleA = 0;

{
  print "  \\begin{scope}[shift=(pos-1)]\n";

  my $boundary_level = $level + 0;

  my $boundary_segments = MyLineSegments->new;
  my ($boundary_n_lo, $boundary_n_hi) = $path->level_to_n_range($boundary_level);
  {
    my ($prev_n_lo, $prev_n_hi) = $path->level_to_n_range($boundary_level-1);
    foreach my $n ($boundary_n_lo .. $boundary_n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      foreach my $dir4 (0 .. 3) {
        my $dx = $dir4_to_dx[$dir4];
        my $dy = $dir4_to_dy[$dir4];
        my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
        my $n_inc = $path->xy_to_n($x+$dx,$y+$dy);
        if ($n_inc > $boundary_n_hi
            # || ($n <= $prev_n_hi && $n_inc > $prev_n_hi)
           ) {
          $boundary_segments->add(2*$x+$dx-$lx, 2*$y+$dy-$ly,
                                  2*$x+$dx+$lx, 2*$y+$dy+$ly);
        }
      }
    }
  }

  my $doublable_segments = MyLineSegments->new;
  $Printwrap = 2;
  printwrap("\\fill[my square] foreach \\p in {");
  $Printwrap += 2;
  my ($n_lo, $n_hi) = $path->level_to_n_range($level);
  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    $x *= 2;
    $y *= 2;
    my $n_double = $path->xy_to_n($x,$y);
    if ($n_double > $max_n_double) {
      $max_n_double = $n_double;
      $max_n = $n;
    }
    if ($n_double <= $n_hi) {
      $MinusDoubleA++;
    }

    $x *= 2;
    $y *= 2;
    printwrap("($x,$y),");
  }
  printwrap("}");
  printwrap(" { \\p +(-\\MySquare,-\\MySquare) rectangle +(\\MySquare,\\MySquare) };\n");

  $Printwrap = 2;
  printwrap("\\draw ");
  $Printwrap += 2;
  $boundary_segments->printwrap;
  printwrap(";\n");
  print "  \\end{scope}\n";
}

sub points_square {
  my ($x,$y) = @_;
  return [ [2*$x-1, 2*$y-1],
           [2*$x+1, 2*$y-1],
           [2*$x+1, 2*$y+1],
           [2*$x-1, 2*$y+1] ];
}

my $MinusDoubleB = 0;
{
  my $plain_segments  = MyLineSegments->new (name => 'n');
  my $doublable_segments = MyLineSegments->new (name => 'doublable');
  my $doublable_gpc = MyPlanar::points_to_gpc(points_square(0,0));

  {
    my $A = 0;
    my ($n_lo, $n_hi) = $path->level_to_n_range($level);
    foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      my $points = points_square($x,$y);
      $plain_segments->add_cycle(map {@$_} @$points);

      if ($path->xy_to_n(2*$x,2*$y) <= $n_hi) {
        $doublable_segments->add_cycle(map {@$_} @$points);
        my $gpc = MyPlanar::points_to_gpc($points);
        $doublable_gpc = Math::Geometry::Planar::GpcClip('UNION',
                                                         $doublable_gpc, $gpc);
        $A++;
      }
    }
    $A == $MinusDoubleA or die;
  }

  $plain_segments->remove_segments($doublable_segments);

  print "  \\begin{scope}[shift=(pos-2)]\n";
  $Printwrap = 2;
  printwrap("\\draw[my plain line] ");
  $Printwrap += 2;
  $plain_segments->printwrap;
  printwrap(";\n");

  my $A = 0;
  my @contours = Math::Geometry::Planar::Gpc2Polygons($doublable_gpc);
  my $total_holes = 0;
  foreach my $contour (@contours) {  # Math::Geometry::Planar contour
    ### $contour
    # Second and subsequent polygon of the contour are holes.
    # Think their points already cycle the opposite way as required by pgf
    # default \fill[nonzero rule], but using \fill[even odd rule] works either
    # way.
    $Printwrap = 2;
    printwrap(" \\fill[my both,even odd rule]");
    $Printwrap += 2;
    my $polygons = $contour->polygons;  # arrayref of points arrayrefs
    $total_holes += scalar(@$polygons)-1;
    my $separator = '';
    foreach my $points (@$polygons) {
      ### $points
      printwrap($separator);
      $separator = "\n";
      printwrap_poly($points);
      $A += $contour->area / 4;
      $MinusDoubleB += $contour->perimeter / 2;
    }
    printwrap(";" . ($total_holes ? " % including $total_holes holes" : "")
              . "\n");
  }
  $A == $MinusDoubleA or die;

  print "  \\end{scope}\n";
}

printf "  %% max N double %d = %#b from N=%d %#b\n",
  $max_n_double, $max_n_double, $max_n, $max_n;
print "  \\newcommand{\\MyMinusDoubleA}{$MinusDoubleA}\n";
print "  \\newcommand{\\MyMinusDoubleB}{$MinusDoubleB}\n";
print "% GP-Test-Last  MinusDoubleA($level) == $MinusDoubleA\n";
print "% GP-Test-Last  MinusDoubleB($level) == $MinusDoubleB\n";
print "  % end generated\n";
exit 0;
