#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use List::Util 'min','max';
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'chamfer','bk','xy_rotate_minus90','BlobN';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 8;
if (@ARGV) { $level = shift @ARGV; }

print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::DragonCurve->new;
my ($n_lo,$n_hi) = $path->level_to_n_range($level);

my $blob_n_start = BlobN($level);
my $blob_n_end = BlobN($level+1)-3;

my ($m_n_lo,$m_n_hi) = $path->level_to_n_range($level-2);

# (3-I)*(x+I*y) == (3*x+y) + (-x+3*y)*I
my ($m_x,$m_y) = bk($level);
($m_x,$m_y) = ((3*$m_x + $m_y)/4,  (-$m_x - 3*$m_y)/4);

sub transform_m {
  my ($x,$y) = @_;
  ($x,$y) = xy_rotate_minus90($x,$y);  # to right side, -I
  return ($x+$m_x, $y+$m_y);
}

{
  my ($x,$y) = transform_m(0,0);
  print "  \\coordinate (start-m) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($m_n_hi);
  ($x,$y) = transform_m($x,$y);
  print "  \\coordinate (end-m) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($n_hi);
  print "  \\coordinate (end) at ($x,$y);\n";
}

{
  my $blob_segments = MyLineSegments->new;
  my $other_segments = MyLineSegments->new;
  foreach my $n ($n_lo .. $n_hi-1) {
    my $segs = ($n >= $blob_n_start && $n < $blob_n_end
                ? $blob_segments : $other_segments);
    $segs->add($path->n_to_xy($n),
               $path->n_to_xy($n+1));
  }
  foreach my $n ($m_n_lo .. $m_n_hi-1) {
    my ($x1,$y1) = $path->n_to_xy($n);
    my ($x2,$y2) = $path->n_to_xy($n+1);
    ($x1,$y1) = transform_m($x1,$y1);
    ($x2,$y2) = transform_m($x2,$y2);
    $blob_segments->remove($x1,$y1, $x2,$y2);
    $other_segments->remove($x1,$y1, $x2,$y2);
  }

  $Printwrap = 2;
  printwrap("\\draw[my grey] ");
  $Printwrap += 2;
  $other_segments->printwrap;
  printwrap(";\n");

  $Printwrap = 2;
  printwrap("\\draw ");
  $Printwrap += 2;
  $blob_segments->printwrap;
  printwrap(";\n");
}

sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
  return defined $n && $n < $n_limit;
}

my $dashed = -1;
$Printwrap = 2;
my $join = '';
foreach my $n ($m_n_lo .. $m_n_hi) {
  ### $n
  my ($x,$y) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  ($x,$y) = transform_m($x,$y);
  ($x2,$y2) = transform_m($x2,$y2);

  my $new_dashed =
    path_xyxy_is_traversed_either($path, $x,$y, $x2,$y2, $n_hi)
    ? 0 : 1;
  if ($n == $m_n_lo) {
    if ($new_dashed) {
      printwrap("\\draw[my dashed] ");
      $Printwrap += 2;
    } else {
      printwrap("\\draw[my thick] ");
      $Printwrap += 2;
    }
  }

  my $sx = $x;
  my $sy = $y;
  if (my ($sdx,$sdy) = $path->n_to_dxdy($n-1)) {
    ($sdx,$sdy) = xy_rotate_minus90($sdx,$sdy);
    $sx .= chamfer(-$sdx);
    $sy .= chamfer(-$sdy);
  }

  my ($tdx,$tdy) = $path->n_to_dxdy($n);
  ($tdx,$tdy) = xy_rotate_minus90($tdx,$tdy);
  my $tx = $x . chamfer($tdx);
  my $ty = $y . chamfer($tdy);

  if ($new_dashed && ! $dashed) {
    printwrap("$join($sx,$sy);\n");
    $Printwrap = 2;
    printwrap("\\draw[my dashed] ");
    $Printwrap += 2;
    $join = '';
  }

  if ($n == $m_n_lo) {
    printwrap("($x,$y)");
    $join = '--';
    next;
  } elsif ($n == $m_n_hi) {
    printwrap("--($x,$y);\n");
    next;
  } else {
    printwrap("$join($sx,$sy)");
  }
  $join = '--';

  if (! $new_dashed && $dashed) {
    printwrap("$join($tx,$ty);\n");
    $Printwrap = 2;
    printwrap("\\draw[my thick] ");
    $Printwrap += 2;
    $join = '';
  }
  printwrap("$join($tx,$ty)");
  $join = '--';

  $dashed = $new_dashed;
}

print "  % end generated\n";
exit 0;
