#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use List::Util 'max';
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon ':all';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 11;
if (@ARGV) { $level = $ARGV[0]; }

print "  % generated by tools/$FindBin::Script $level\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::DragonCurve->new;

sub path_n_is_right_boundary {
  my ($path, $n, $n_lo,$n_hi) = @_;
  # path_n_is_right_boundary(): $n
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to right side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_lo,$n_hi)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_lo,$n_hi)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_lo,$n_hi));
}
sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_lo,$n_hi) = @_;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
  return defined $n && ($n >= $n_lo && $n < $n_hi);
}

my %twin_lo_offset = (4 => 0,
                      5 => 0,
                      6 => 0,
                      7 => 3,
                      8 => 2,
                      9 => 4,
                      10 => 11,
                      11 => 25,
                     );

my $blob_n_lo = BlobN($level);
my $blob_n_hi = BlobN($level+1)-3;
{
  my ($x,$y) = $path->n_to_xy($blob_n_lo);
  print "  \\coordinate (start) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($blob_n_hi);
  print "  \\coordinate (end) at ($x,$y);\n";
}

my $segments = MyLineSegments->new;
foreach my $n ($blob_n_lo .. $blob_n_hi-1) {
  $segments->add($path->n_to_xy($n), $path->n_to_xy($n+1));
}

my @twindragons;
my @twindragon_sizes;
my $BlobTA = 0;

sub draw_blob {
  my ($k, $at_x,$at_y, $dir) = @_;
  $dir %= 4;
  ### draw_blob(): "k=$k  $at_x,$at_y  dir=$dir"

  return if $k < 4;

  my ($dx,$dy) = BlobDelta_xy($k);
  foreach (1 .. $dir) { ($dx,$dy) = xy_rotate_plus90($dx,$dy); }
  my $at_xe = $at_x + $dx;
  my $at_ye = $at_y + $dy;
  ### $at_xe
  ### $at_ye

  ### k-4 at end ...
  if ($k-4 >= 4) {
    my ($dx4,$dy4) = BlobDelta_xy($k-4);
    ### unrotated ...
    ### $dx4
    ### $dy4
    foreach (1 .. $dir) { ($dx4,$dy4) = xy_rotate_plus90($dx4,$dy4); }
    draw_blob($k-4, $at_xe+$dx4,$at_ye+$dy4, $dir+2);
  }

  # k-4 at start
  if ($k-4 >= 4) {
    draw_blob($k-4, $at_x,$at_y, $dir+2);
  }

  # k-2 in middle, starting back from end
  if ($k-2 >= 4) {
    my ($dx2,$dy2) = (0,0);
    if ($k==6) { ($dx2,$dy2) = (0,1); }
    elsif ($k==7) { ($dx2,$dy2) = (-1,1); }
    elsif ($k-4 >= 4) {
      my ($x1,$y1) = $path->n_to_xy(BlobN($k-4)-3);
      my ($x2,$y2) = BlobEnd_xy($k-4);
      ($dx2,$dy2) = ($x2-$x1, $y2-$y1);
    }
    ### unrotated ...
    ### $dx2
    ### $dy2
    foreach (1 .. $dir+2) { ($dx2,$dy2) = xy_rotate_plus90($dx2,$dy2); }
    ### $dx2
    ### $dy2
    draw_blob($k-2, $at_xe-$dx2,$at_ye-$dy2, $dir+3);
  }

  my ($bx,$by) = BlobStart_xy($k);

  # twindragon k-4 positions
  $twindragon_sizes[$k-4]++;
  $BlobTA += TA($k-4);
  my $t_offset = $twin_lo_offset{$k} || 0;
  ### $t_offset
  my $b4 = BlobSegments($k-4);
  my $tlength = 4*2**($k-4);
  my $n_lo = BlobN($k) + ($b4 ? $b4+3 : 0) + $t_offset;
  my $n_hi = $n_lo + $tlength;
  ### $tlength
  ### $n_lo
  ### $n_hi
  {
    my ($x1,$y1) = $path->n_to_xy($n_lo);
    my ($x2,$y2) = $path->n_to_xy($n_hi);
    unless ($x1==$x2 && $y1==$y2) {
      die "oops twindragon ends n_lo != n_hi";
    }
  }

  my $transform = sub {
    my ($x,$y) = @_;
    $x -= $bx;
    $y -= $by;
    foreach (1 .. $dir) { ($x,$y) = xy_rotate_plus90($x,$y); }
    $x += $at_x;
    $y += $at_y;
    return ($x,$y);
  };

  # all twindragon segments
  foreach my $n ($n_lo .. $n_hi-1) {
    my ($x,$y) = $path->n_to_xy($n);
    my ($x2,$y2) = $path->n_to_xy($n+1);
    ($x,$y) = $transform->($x,$y);
    ($x2,$y2) = $transform->($x2,$y2);
    $segments->remove($x,$y, $x2,$y2);
  }

  my @points;

  # boundary twindragon segments for fill
  for (my $n = $n_lo; $n != $n_hi; ) {
    my ($x,$y) = $path->n_to_xy($n);
    push @points, [ ($x,$y) = $transform->($x,$y) ];

    $n++;
    if (! path_n_is_right_boundary($path, $n, $n_lo,$n_hi)) {
      ($x,$y) = $path->n_to_xy($n);
      my @n_list = $path->xy_to_n_list($x,$y);
      $n = max(@n_list);
    }
  }
  push @twindragons, \@points;
}

draw_blob($level, BlobStart_xy($level), 0);


my %seen;
foreach my $points (@twindragons) {
  foreach my $point (@$points) {
    my ($x,$y) = @$point;
    $seen{"$x,$y"}++;
    if ($seen{"$x,$y"} > 2) {
      die "oops, seen $x,$y  > 2";
    }
  }
}

sub xy_want_chamfer {
  my ($x,$y) = @_;
  return ($seen{"$x,$y"} >= 2
          || ($seen{"$x,$y"} && $segments->is_visited($x,$y)));
}

$Printwrap = 2;
printwrap("\\fill ");
$Printwrap = 4;
foreach my $i (0 .. $#twindragons) {
  my $points = $twindragons[$i];
  my $last = ($i == $#twindragons);

  my $join = '';
  foreach my $j (0 .. $#$points) {
    my $p1 = $points->[$j];
    my $p2 = $points->[($j+1) % scalar(@$points)];
    my ($x1,$y1) = @$p1;
    my ($x2,$y2) = @$p2;
    my $dx = $x2 - $x1;
    my $dy = $y2 - $y1;

    if (xy_want_chamfer($x1,$y1)) {
      $x1 .= chamfer($dx);
      $y1 .= chamfer($dy);
      printwrap("$join($x1,$y1)");
      $join = '--';
    }
    if (xy_want_chamfer($x2,$y2)) {
      $x2 .= chamfer(-$dx);
      $y2 .= chamfer(-$dy);
    }
    printwrap("$join($x2,$y2)");
    $join = '--';
  }
  printwrap("${join}cycle".($last?';':'')."\n");
}

if ($segments->length) {
  $Printwrap = 2;
  printwrap("\\draw ");
  $Printwrap += 2;
  $segments->printwrap;
  printwrap(";");

  $Printwrap = 2;
  my $draw_count = $segments->lines_joined_count_points;
  printwrap(" % drawn with $draw_count points\n");
}

print "  \\newcommand{\\MyBlobTA}{$BlobTA}\n";
print "  \\newcommand{\\MyBlobTAlist}{",
  join(',', map{$_||0} @twindragon_sizes), "}\n";
print "% GP-Test-Last  BlobTA($level) == $BlobTA\n";
print "% GP-Test-Last  BlobTAlist($level) == [",
  join(',', map{$_||0} @twindragon_sizes), "]\n";
print "  % end generated\n";
exit 0;



















__END__

# use 5.004;
# use strict;
# use FindBin;
# use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()
#
# use FindBin;
# use lib $FindBin::Bin;
# use MyDragon ':all';
# use MyLineSegments;
# use MyPrintwrap ':all';
#
# # uncomment this to run the ### lines
# use Smart::Comments;
#
#
# my $level = 8;
#
# print "  % generated by tools/$FindBin::Script\n";
# print "  \\newcommand{\\MyLevel}{$level}\n";
#
# my $path = Math::PlanePath::DragonCurve->new;
#
# my $n_lo = BlobN($level);
# my $n_hi = BlobN($level+1) - 3;
#
# {
#   my ($x,$y) = $path->n_to_xy($n_lo);
#   print "% GP-Test  BlobStart_xy($level) = $x + I*$y\n";
#   print "  \\coordinate (start) at ($x,$y);\n";
# }
# {
#   my ($x,$y) = $path->n_to_xy($n_hi);
#   print "% GP-Test  BlobEnd($level) = $x + I*$y\n";
#   print "  \\coordinate (end) at ($x,$y);\n";
# }
#
# my %twin_lo_offset = (4 => 0,
#                       5 => 0,
#                       6 => 0,
#                       7 => 3,
#                       8 => 2,
#                       9 => 4,
#                       10 => 11,
#                       11 => 25,
#                      );
#
# # arefs [ [0,3], ...] which are segment numbers of twindragon parts
# my @blob_twindragons = ([],        # 0
#                         [],
#                         [],
#                         [],
#                         [ [0, 4] ],    # 4
#                         [ [0, 8] ]);   # 5
# foreach my $k (6 .. $level) {
#   my $b0 = BlobSegments($k);
#   my $b2 = BlobSegments($k-2);
#   my $b4 = BlobSegments($k-4);
#   my $t4 = 4 * 2**($k-4);
#   my $end = $b4 +3 +$b2+$t4 +3 +$b4;
#   ### $k
#   ### $b0
#   ### $b2
#   ### $b4
#   ### $t4
#   ### $end
#
#   my @ret;
#
#   # twindragon k-4 positions
#   my $t_offset = $twin_lo_offset{$k} || 0;
#   ### $t_offset
#   my $tstart = ($b4 ? $b4+3 : 0) + $t_offset;
#   my $tlength = 4*2**($k-4);
#   my $tend = $tstart + $tlength;
#
#   # blob k-2, backwards from end
#   {
#     my $e = $b0 - ($b4 ? $b4 + 3 : 0);
#     if ($k==6) { $e--; }
#     if ($k==7) { $e-=2; }
#     foreach my $part (@{$blob_twindragons[$k-2]}) {
#       push @ret, [ $e-$part->[1], $e-$part->[0] ];
#     }
#   }
#   # back before twindragon if necessary
#   foreach my $part (@ret) {
#     if ($part->[0] < $tend && $part->[1] > $tend) {
#       die "oops, part across twindragon";
#     }
#     if ($part->[1] < $tend) {
#       $part->[0] -= $tlength;
#       $part->[1] -= $tlength;
#     }
#   }
#
#   # twindragon k-4
#   push @ret, [$tstart, $tend];
#
#   # first blob k-4
#   if ($b4) {
#     foreach my $part (@{$blob_twindragons[$k-4]}) {
#       push @ret, [ $part->[0], $part->[1] ];
#     }
#   }
#
#   # last blob k-4
#   if ($b4) {
#     my $n = $end - $b4;
#     foreach my $part (@{$blob_twindragons[$k-4]}) {
#       push @ret, [ $part->[0]+$n, $part->[1]+$n ];
#     }
#   }
#
#   @ret = sort {$a->[0] <=> $b->[0]} @ret;
#   $blob_twindragons[$k] = \@ret;
# }
# ### @blob_twindragons
#
# my $aref = $blob_twindragons[$level];
# print "  % ",
#   join(', ',
#        map {
#          my ($lo,$hi) = @$_;
#          "${lo}..$hi"
#        } @$aref),
#   "\n";
#
# $Printwrap = 2;
# printwrap("\\fill ");
# $Printwrap = 4;
# foreach my $i (0 .. $#$aref) {
#   my $part = $aref->[$i];
#   my $last = $i==$#$aref;
#   my ($lo,$hi) = @$part;
#
#   my $plo = $n_lo + $lo;
#   my $phi = $n_lo + $hi;
#
#   {
#     my ($x1,$y1) = $path->n_to_xy($plo);
#     my ($x2,$y2) = $path->n_to_xy($phi);
#     unless ($x1==$x2 && $y1==$y2) {
#     }
#   }
#
#   my $join = '';
#   foreach my $n ($plo .. $phi-1) {
#     my ($x,$y) = $path->n_to_xy($n);
#     my @n_list = $path->xy_to_n_list($x,$y);
#     @n_list = grep {$_ != $n} @n_list;
#     my $n_other = $n_list[0] || 0;
#     my $is_join = ! ($n_other >= $plo && $n_other <= $phi);
#     $is_join=1;
#     if ($is_join) {
#       my ($dx,$dy) = $path->n_to_dxdy($n);
#       my $x2 = $x + $dx;
#       my $y2 = $y + $dy;
#       my $sx = $x . chamfer($dx);
#       my $sy = $y . chamfer($dy);
#       my $tx = $x2 . chamfer(-$dx);
#       my $ty = $y2 . chamfer(-$dy);
#       printwrap("$join($sx,$sy)");
#       printwrap("--($tx,$ty)");
#     } else {
#       printwrap("$join($x,$y)");
#     }
#     $join = '--';
#   }
#   printwrap("${join}cycle".($last?';':'')."\n");
# }
#
# print "  % end generated\n";
# exit 0;
