#!/usr/bin/perl -w

# Copyright 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use Math::PlanePath::DragonCurve 124; # v.124 for n_to_n_list()

use FindBin;
use lib $FindBin::Bin;
use MyDragon ':all';
use MyLineSegments;
use MyPrintwrap ':all';


my $max_k = 12;
if (@ARGV) {
  $max_k = shift @ARGV;
}

print "% generated by tools/$FindBin::Script $max_k\n";
my $path = Math::PlanePath::DragonCurve->new;

sub n_is_right_boundary {
  my ($n) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to right side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! defined $path->xyxy_to_n_either($x1,$y1, $x1s,$y1s)
          || ! defined $path->xyxy_to_n_either($x2,$y2, $x2s,$y2s)
          || ! defined $path->xyxy_to_n_either($x1s,$y1s, $x2s,$y2s));
}

sub n_is_left_boundary {
  my ($n) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to left side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! defined $path->xyxy_to_n_either($x1,$y1, $x1s,$y1s)
          || ! defined $path->xyxy_to_n_either($x2,$y2, $x2s,$y2s)
          || ! defined $path->xyxy_to_n_either($x1s,$y1s, $x2s,$y2s));
}

sub n_to_n_list {
  my ($n) = @_;
  my ($x,$y) = $path->n_to_xy($n) or return;
  return $path->xy_to_n_list($x,$y);
}

my $prev_is_bridge_segment = 1;

my @BlobR = (0,0,0,0);
my @BlobL = (0,0,0,0);
my @BlobB = (0,0,0,0);
my @BlobSegments = (0,0,0,0);
my @BlobS = (0,0,0,0);
my @BlobD = (0,0,0,0);
my @BlobP = (0,0,0,0);
my @BlobAR = (0,0,0,0);
my @BlobAL = (0,0,0,0);
my @BlobStart = (0,0,0,0);
my @BlobEnd = (0,0,0,0);
my @BlobHorizontals = (0,0,0,0);
my @BlobVerticals = (0,0,0,0);
my @BlobLines = (0,0,0,0);
my @BlobRQ = (0,0,0,0);
my @BlobLQ = (0,0,0,0);
my @BlobBQ = (0,0,0,0);

my $BlobR = 0;
my $BlobL = 0;
my $BlobB = 0;
my $BlobSegments = 0;
my $BlobS = 0;
my $BlobD = 0;
my $BlobP = 0;
my $BlobAR = 0;
my $BlobAL = 0;
my $BlobRQ = 0;
my $BlobLQ = 0;
my $BlobBQ = 0;
my %seen_squares;
my $segments;

for (my $n = 0; @BlobR <= $max_k; $n++) {
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  my ($x2,$y2) = ($x+$dx, $y+$dy);

  my $left = n_is_left_boundary($n);
  my $right = n_is_right_boundary($n);
  my $is_bridge_segment = $left && $right;
  my $is_blob_segment = ! $is_bridge_segment;

  my $is_bridge_start = ($is_bridge_segment && ! $prev_is_bridge_segment);
  my $is_blob_start = ($prev_is_bridge_segment && ! $is_bridge_segment);
  my $is_blob_point = ! $is_bridge_segment || ! $prev_is_bridge_segment;

  if ($is_blob_start) {
    # start of blob
    $BlobR = 0;
    $BlobL = 0;
    $BlobB = 0;
    $BlobAR = 0;
    $BlobAL = 0;
    $BlobSegments = 0;
    $BlobD = 0;
    $BlobS = 0;
    $BlobP = 0;
    $BlobRQ = 0;
    $BlobLQ = 0;
    $BlobBQ = 0;
    $segments = MyLineSegments->new;
    push @BlobStart, "$x+I*$y";
  }

  if ($is_blob_point) {
    # a point within or start or end of a blob
    my @n_list = n_to_n_list($n);
    if (@n_list == 1) {
      $BlobS++;
      $BlobP++;
    } elsif ($n == $n_list[0]) { # count first of double only
      $BlobD++;
      $BlobP++;
    }
  }

  if ($is_blob_segment) {
    $segments->add($x,$y, $x2,$y2);

    if (! $right) {
      $BlobAR++;
    }
    if (! $left) {
      $BlobAL++;
    }

    if ($right) {
      my ($rdx,$rdy) = xy_rotate_minus90($dx,$dy);  # to right side
      my $xc = $x2 + $rdx;
      my $yc = $y2 + $rdy;
      ($xc,$yc) = rect_corners_sort ($x,$y, $xc,$yc);
      if (! $seen_squares{"$xc,$yc"}++) {
        $BlobRQ++;
        $BlobBQ++;
      }
    }
    if ($left) {
      my ($ldx,$ldy) = xy_rotate_plus90($dx,$dy);  # to left side
      my $xc = $x2 + $ldx;
      my $yc = $y2 + $ldy;
      ($xc,$yc) = rect_corners_sort ($x,$y, $xc,$yc);
      if (! $seen_squares{"$xc,$yc"}++) {
        $BlobLQ++;
        $BlobBQ++;
      }
    }
  }

  if (! $is_bridge_segment) {
    $BlobSegments++;
    if ($right) {
      $BlobR++;
      $BlobB++;
    } elsif ($left) {
      $BlobL++;
      $BlobB++;
    }
  }

  if ($is_bridge_start) {
    # end of blob
    push @BlobEnd, "$x+I*$y";
    push @BlobR, $BlobR;
    push @BlobL, $BlobL;
    push @BlobB, $BlobB;
    push @BlobSegments, $BlobSegments;
    push @BlobS, $BlobS;
    push @BlobD, $BlobD;
    push @BlobP, $BlobP;
    push @BlobAR, $BlobAR/4;
    push @BlobAL, $BlobAL/4;
    my ($horizontals, $verticals, $lines) = $segments->num_lines_hvt;
    push @BlobHorizontals, $horizontals;
    push @BlobVerticals, $verticals;
    push @BlobLines, $lines;
    push @BlobRQ, $BlobRQ;
    push @BlobLQ, $BlobLQ;
    push @BlobBQ, $BlobBQ;
  }

  $prev_is_bridge_segment = $is_bridge_segment;
}

print "% GP-DEFINE  BlobR_generated_samples = [ ",join(',',@BlobR), " ];\n";
print "% GP-DEFINE  BlobL_generated_samples = [ ",join(',',@BlobL), " ];\n";
print "% GP-DEFINE  BlobB_generated_samples = [ ",join(',',@BlobB), " ];\n";
print "% GP-DEFINE  BlobAR_generated_samples = [ ",join(',',@BlobAR), " ];\n";
print "% GP-DEFINE  BlobAL_generated_samples = [ ",join(',',@BlobAL), " ];\n";
print "% GP-DEFINE  BlobSegments_generated_samples = [ ",join(',',@BlobSegments), " ];\n";
print "% GP-DEFINE  BlobS_generated_samples = [ ",join(',',@BlobS), " ];\n";
print "% GP-DEFINE  BlobD_generated_samples = [ ",join(',',@BlobD), " ];\n";
print "% GP-DEFINE  BlobP_generated_samples = [ ",join(',',@BlobP), " ];\n";
# print "% GP-DEFINE  BlobStart_generated_samples = [ ",join(',',@BlobStart), " ];\n";
# print "% GP-DEFINE  BlobEnd_generated_samples = [ ",join(',',@BlobEnd), " ];\n";
print "% GP-DEFINE  BlobHorizontals_generated_samples = [ ",join(',',@BlobHorizontals), " ];\n";
print "% GP-DEFINE  BlobVerticals_generated_samples = [ ",join(',',@BlobVerticals), " ];\n";
print "% GP-DEFINE  BlobLines_generated_samples = [ ",join(',',@BlobLines), " ];\n";
print "% GP-DEFINE  BlobRQ_generated_samples = [ ",join(',',@BlobRQ), " ];\n";
print "% GP-DEFINE  BlobLQ_generated_samples = [ ",join(',',@BlobLQ), " ];\n";
print "% GP-DEFINE  BlobBQ_generated_samples = [ ",join(',',@BlobBQ), " ];\n";

print "% end generated\n";
exit 0;
