#!/usr/bin/perl -w

# Copyright 2013, 2014, 2015, 2016, 2017, 2019, 2020, 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Usage: perl test.pl
#
# Run some self-tests on ../dragon.tex
#

use 5.010;
use strict;
use warnings;
use File::Slurp;
use IPC::Run;
use List::Util 'max','min','sum';

use FindBin;
use lib::abs $FindBin::Bin;
use MyDragon ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $verbose = 1;

my $source_dir = File::Spec->catdir ($FindBin::Bin, File::Spec->updir);
$source_dir = File::Spec->rel2abs($source_dir);
print "source dir $source_dir\n";

chdir $source_dir or die $!;

my $dragon_tex_filename = 'dragon.tex';
# $dragon_tex_filename = 'devel/more.tex';
my $dragon_pdf_filename = 'dragon.pdf';
my $dragon_tex_str = File::Slurp::read_file($dragon_tex_filename);
print count_lines($dragon_tex_str)," lines\n";

my $dragon_tex_str_sans_comments = $dragon_tex_str;
$dragon_tex_str_sans_comments =~ s/((^|[^\\])(\\\\)*)%[^\n]*/$1/mg;
# print count_lines($dragon_tex_str_sans_comments)," lines\n";

my $bad = 0;


#------------------------------------------------------------------------------
# Misc

{
  # no \section{X} temporary break for whizzytex
  while ($dragon_tex_str_sans_comments =~ /\\section\{X/g) {
    my $linenum = pos_linenum($dragon_tex_str_sans_comments);
    print "$dragon_tex_filename:$linenum: $&\n";
    $bad = 1;
  }
}


#------------------------------------------------------------------------------
# OEIS refs

{
  my %anums;
  while ($dragon_tex_str_sans_comments
         =~ /My[a-zA-Z]*Anumber.*?(?<anum>A\d{6,})/sg) {
    $anums{$+{'anum'}}++;
  }
  my $total = sum(0, values %anums);
  my $distinct = keys %anums;

  if (0) {
    print "\n";
    foreach (sort keys %anums) { print $_,", " };
    print "\n";
    print "STOP\n";
    exit 0;
  }

  my $idx_str = File::Slurp::read_file
    (File::Spec->catfile($source_dir,'dragon.anums.idx'),
     err_mode=>'quiet') // '';
  my $idx_num_lines = str_num_lines($idx_str);

  my %ind_anums;
  my $ind_str = File::Slurp::read_file
    (File::Spec->catfile($source_dir,'dragon.anums.ind'),
     err_mode=>'quiet') // '';
  while ($ind_str =~ /\\item.*?(A\d{6,})\b/sg) { $ind_anums{$1}++; }
  my $ind_num_items = keys %ind_anums;
  $ind_num_items == sum(0, values %ind_anums)
    or die "oops, duplicate in dragon.anums.ind";
  foreach my $anum (keys %ind_anums) {
    if (! exists $anums{$anum}) {
      print "dragon.anums.ind extra over tags: $anum\n";
      # $bad++;
    }
  }

  print "A-number tags $total, distinct $distinct, cf idx $idx_num_lines ind $ind_num_items\n";
}

sub str_num_lines {
  my ($str) = @_;
  return ($str =~ tr/\n/\n/);
}

#------------------------------------------------------------------------------
# dragon.tex copyright years also at:
#    text part of dragon.tex
#    ../devel/dragon-defines.gp
#    ../devel/dragon-defines.header.gp

{
  # same copyright years as dragon.tex where defines are extracted from
  $dragon_tex_str =~ /^% (Copyright .*)/m or die;
  my $want_copyright = $1;
  ### $want_copyright

  {
    $dragon_tex_str =~ /^\\footnote.*?\n\Q$want_copyright\E\.$/m
      or die "dragon.tex: header comment copyright not found as footnote";
  }

  foreach my $filename ('devel/dragon-defines.gp',
                        'devel/dragon-defines.header.gp') {
    my $str = File::Slurp::slurp($filename);
    $str =~ /^\\\\ \Q$want_copyright\E$/m
      or die "$filename copyright different from dragon.tex";
  }
}

#------------------------------------------------------------------------------
# dragon.pdf Title and Author

chdir $source_dir or die $!;
{
  IPC::Run::run(['pdfinfo', '-isodates', $dragon_pdf_filename],
                '>', \my $pdfinfo_out)
      or die;
  $pdfinfo_out =~ s/^Jav.*\n//m;
  print $pdfinfo_out;
  {
    $pdfinfo_out =~ /^Title:\s*(.*)/m
      or die;
    my $title = $1;
    # print "title: $title\n";
    $title =~ /Dragon/
      or die;
  }
  {
    $pdfinfo_out =~ /^Author:\s*(.*)/m
      or die;
    my $author = $1;
    # print "author: $author\n";
    $author =~ /Kevin/
      or die;
  }
}

#------------------------------------------------------------------------------

# Return the line number of character position $pos in $str.
# If $pos is undef or not supplied, use pos($str).
sub pos_linenum {
  my ($str, $pos) = @_;
  $pos //= pos($_[0]);
  $str = substr($str, 0, $pos);
  return 1 + scalar($str =~ tr/\n//);
}

sub count_lines {
  my ($str) = @_;
  my $count = ($str =~ tr/\n/\n/);
  if ($str =~ /[^\n]$/) { $count++; }
  return $count;
}

#------------------------------------------------------------------------------
exit ($bad ? 1 : 0);
