#!/usr/bin/perl -w

# Copyright 2015, 2016, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use File::Basename 'dirname';
use File::Spec;
use File::stat;
use File::Slurp;
use IPC::Run;
use List::Util 'pairmap';
use POSIX ();
use Graph;
use Graph::Maker::TwindragonAreaTree;
use List::Util 'min','max';
use Math::PlanePath::ComplexPlus;

use FindBin;
BEGIN {
  unshift @INC, File::Spec->catdir($FindBin::Bin,
                                   File::Spec->updir,
                                   'tools');
}
use MyDragon 'mul';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 10;
if (@ARGV) { $level = shift @ARGV; }
my $dir = "/tmp";
chdir $dir or die;

my $pdf_filename = File::Spec->catfile($dir, "twindragon-area-tree-$level.pdf");
my $tex_filename = File::Spec->catfile($dir, "twindragon-area-tree-$level.tex");
my $program_filename = File::Spec->catfile($FindBin::Bin, $FindBin::Script);

sub transform {
  my ($x,$y) = @_;
  ($x,$y) = mul($x,$y);
  return ($x, $y);
}

my $graph = Graph::Maker->new('twindragon_area_tree',
                              level       => $level,
                              undirected  => 1,
                             );
my $path = Math::PlanePath::ComplexPlus->new;

my $tree_segments = MyLineSegments->new;
foreach my $edge ($graph->edges) {
  ### $edge
  my ($v1,$v2) = @$edge;
  my ($x1,$y1) = $path->n_to_xy($v1);
  my ($x2,$y2) = $path->n_to_xy($v2);
  $tree_segments->add($x1,$y1, $x2,$y2);
}

my $x_min = 0;
my $x_max = 0;
my $y_min = 0;
my $y_max = 0;
foreach my $v ($graph->vertices) {
  my ($x,$y) = $path->n_to_xy($v);
  ($x,$y) = transform($x,$y);
  $x_min = min($x_min,$x);
  $x_max = max($x_max,$x);
  $y_min = min($y_min,$y);
  $y_max = max($y_max,$y);
}
my $x_mid = ($x_min + $x_max) / 2;

my $stat = File::stat::stat($program_filename)
  or die "Oops, cannot stat $program_filename: $!";
my $creation_date = POSIX::strftime('D:%Y%m%d',  # date-only here
                                    localtime($stat->mtime));

open my $fh, '>', $tex_filename or die;
print $fh <<"HERE";
\\documentclass{standalone}
\\pagestyle{empty}
\\usepackage{tikz}
\\usetikzlibrary{calc}

\\pdfinfo{
  /Title    (Twindragon Area Tree $level)
  /Author   (Kevin Ryde)
  /Keywords (Twindragon;Area tree;Mandelbrot)
  /CreationDate ($creation_date)
  /ModDate      ($creation_date)
  /Lang     (en)
}

\\begin{document}
\\begin{tikzpicture}[scale=.3]
HERE

# # node dots
# {
#   foreach my $v (sort $graph->vertices) {
#     my ($x,$y) = split /,/, $v;
#     ($x,$y) = $transform->($x,$y);
#     print $fh "\\fill ($x,$y) circle (.2);\n";
#   }
# }

# tree lines
{
  local *STDOUT = $fh;
  $Printwrap = 2;
  printwrap("\\draw ");
  $Printwrap += 2;
  $tree_segments->printwrap;
  printwrap(";\n");
}

print $fh "  \\node at ($x_mid, $y_min-1.5) [below=2ex,font=\\normalsize,align=center]\n";
print $fh "  {Twindragon area tree level $level";
if ($level == 10) {
  print $fh ", \\kern.3em after Mandebrot\\\\\n   ``The Fractal Geometry of Nature'', 1983, page 67";
}
print $fh "};\n";

print $fh <<'HERE';
  \coordinate (bottom-left) at (current bounding box.south west);
  \coordinate (top-right)   at (current bounding box.north east);
  \pgfresetboundingbox
  \useasboundingbox ($(bottom-left) - (6,2)$)
          rectangle ($(top-right)   + (6,4)$);

\end{tikzpicture}
\end{document}
HERE

system("cat $tex_filename");
IPC::Run::run(['latexmk','-gg','-pdf',$tex_filename]);

#------------------------------------------------------------------------------


{
  my $qpdf_filename = 'tmp-qpdf.pdf';
  IPC::Run::run(['qpdf',
                 '--linearize',
                 '--stream-data=compress',
                 $pdf_filename, $qpdf_filename])
      or die;
  rename $qpdf_filename, $pdf_filename or die;
}

IPC::Run::run(['qpdf','--check',$pdf_filename]);
IPC::Run::run(['pdfinfo',$pdf_filename], '>',\my $out);
$out =~ s/^Ja.*\n//m;
print $out;
IPC::Run::run(['ls', '-l', $pdf_filename]);

#------------------------------------------------------------------------------

exit 0;









__END__

# pdfcrop loses pdfinfo author etc
#
# # $filename is a pdf or ps.
# # Return ($x1,$y1, $x2,$y2) which are bounding box corners maximum and
# # minimum of the bounding boxes of all the pages.  So the bounding box
# # returned suffices for all pages.
# #
# sub pdf_bbox_maximum {
#   my ($filename) = @_;
#
#   my $bbox_out;
#   IPC::Run::run(['gs','-dNOPAUSE','-sDEVICE=bbox',$filename],
#                 '<', File::Spec->devnull,
#                 '>', \$bbox_out,
#                 '2>&1')
#       or die "ghostscript error:\n$bbox_out\n";
#   ### $bbox_out
#
#   my @bboxes;
#   while ($bbox_out =~ /%%BoundingBox: (\d+) (\d+) (\d+) (\d+)/g) {
#     print "bbox $1,$2 $3,$4\n";
#     push @bboxes, [$1,$2,$3,$4];
#   }
#   return (min(map {$_->[0]} @bboxes),  # x1
#           min(map {$_->[1]} @bboxes),  # y1
#           max(map {$_->[2]} @bboxes),  # x2
#           max(map {$_->[3]} @bboxes)); # y2
# }
#
# # stretch bounding box for some margin
# sub bbox_stretch {
#   my ($x1,$y1, $x2,$y2) = @_;
#   my $hstretch = int(($x2-$x1) * .04);
#   my $vstretch = int(($y2-$y1) * .04);
#   $hstretch = min($hstretch,$x1);
#   $vstretch = min($vstretch,$y1);
#   return ($x1 - $hstretch,
#           $y1 - $vstretch,
#           $x2 + $hstretch,
#           $y2 + $vstretch);
# }
#
# my ($x1,$y1, $x2,$y2) = pdf_bbox_maximum($pdf_filename);
# print "max  $x1,$y1 $x2,$y2\n";
#
# ($x1,$y1, $x2,$y2) = bbox_stretch ($x1,$y1, $x2,$y2);
# print "stretched $x1,$y1 $x2,$y2\n";
#
# {
#   my $pdfcrop_filename = 'tmp-cropped.pdf';
#   IPC::Run::run(['pdfcrop',
#                  # ,'--verbose'
#                  '--pdftex',
#                  '--bbox',"$x1 $y1 $x2 $y2",
#                  $pdf_filename, $pdfcrop_filename])
#       or die;
#   rename $pdfcrop_filename, $pdf_filename or die;
# }
